@@ -34,28 +34,24 @@ namespace Clasp {
  */
 class CBConsequences : public Enumerator {
 public:
-	enum Consequences_t { brave_consequences, cautious_consequences };
+	enum Consequences_t { 
+		brave_consequences    = (Model::max_value << 1) | Model::model_cons,
+		cautious_consequences = (Model::max_value << 2) | Model::model_cons,
+	};
 	/*!
 	 * \param type Type of consequences to compute.
 	 */
 	explicit CBConsequences(Consequences_t type);
 	~CBConsequences();
-	Consequences_t  cbType()    const { return type_; }
-
-	class GlobalConstraint;
-	class LocalConstraint;
+	int  modelType() const { return type_; }
+	bool exhaustive()const { return true; }
 private:
-	EnumeratorConstraint* doInit(SharedContext& ctx, uint32 t, bool start);
-	bool   updateConstraint(Solver& s, bool disjoint);
-	bool   backtrack(Solver& s);
-	bool   updateModel(Solver& s);
-	bool   ignoreSymmetric() const; 
-	void   add(Solver& s, Literal p);
-	Constraint* addNewConstraint(Solver& s);
-	void updateBraveModel(Solver& s);
-	void updateCautiousModel(Solver& s);
-	LitVec            C_;
-	GlobalConstraint* current_;
+	class  CBFinder;
+	class  SharedConstraint;
+	ConPtr doInit(SharedContext& ctx, MinimizeConstraint* m, int numModels);
+	void   addCurrent(Solver& s, LitVec& con, ValueVec& m);
+	LitVec            cons_;
+	SharedConstraint* shared_;
 	Consequences_t    type_;
 };
 
