@@ -87,97 +87,30 @@ public:
 	 */
 	explicit ClauseCreator(Solver* s = 0);
 	//! Sets the solver in which created clauses are stored.
-	void setSolver(Solver& s) { solver_ = &s; }
+	void setSolver(Solver& s)         { solver_ = &s; }
+	//! Adds additional flags to be applied in end().
+	void addDefaultFlags(uint32 f)    { flags_ |= f; }
 	//! Reserve space for a clause of size s.
 	void reserve(LitVec::size_type s) { literals_.reserve(s); }
-	//! Sets the initial activity for the to be created clause.
-	void setActivity(uint32 a)        { extra_.setActivity(a);  }
-	//! Sets the initial literal block distance for the to be created clause.
-	void setLbd(uint32 lbd)           { extra_.setLbd(lbd); }
 	//! Discards the current clause.
-	void clear() { literals_.clear(); }
-	
-	//! Starts the creation of a new clause.
-	/*!
-	 * \pre s.decisionLevel() == 0 || t != Constraint_t::static_constraint
-	 */
-	ClauseCreator& start(ConstraintType t = Constraint_t::static_constraint);
-	
-	//! Start the creation of an asserting or unit clause.
-	/*!
-   * All literals except the one passed to startAsserting() must be
-	 * false.
-	 * \param t The type of the learnt clause.
-	 * \param assertingLit The literal to be asserted.
-	 * \note Position 0 is always reserved for the asserting literal.
-	 */
-	ClauseCreator& startAsserting(ConstraintType t, const Literal& assertingLit = Literal());
-	
-	//! Adds the literal p to the clause.
-	/*!
-	 * \note p is only added if p is free or the current DL is > 0.
-	 * \pre Clause neither contains p nor ~p.
-	 * \pre If clause was started with startAsserting() p is false!
-	 */
-	ClauseCreator& add(const Literal& p);
+	void clear()                      { literals_.clear(); }
 
-	//! Returns the literal at position i.
+	//! Status of a clause.
 	/*!
-	 * \pre i < size()
+	 * For a clause with literals [l1,...,ln], status is one of:
 	 */
-	Literal& operator[](LitVec::size_type i)       { return literals_[i]; }
-	Literal  operator[](LitVec::size_type i) const { return literals_[i]; }
-	//! Returns the literals of the clause.
-	const LitVec&     lits()    const { return literals_; }
-	LitVec&           lits()          { return literals_; }
-	//! Returns the current size of the clause.
-	LitVec::size_type size()    const { return literals_.size(); }
-	bool              empty()   const { return literals_.empty();}
-	//! Returns the clause's type.
-	ConstraintType    type()    const { return extra_.type(); }
-	//! Returns the current position of the second watch.
-	uint32            sw()      const { return 1; }
-	ClauseInfo        info()    const { return extra_; }
-		
-	//! Possible status values of a clause.
 	enum Status {
-		status_open        = 0,  /**< More than one literal is free */
-		status_sat         = 1,  /**< At least one literal is true  */
-		status_unit        = 2,  /**< All but one literal false     */
-		status_asserting   = 3,  /**< Unit after backtracking to second highest dl in clause */
-		status_conflicting = 4,  /**< All literals are false        */
-		status_empty       = 12, /**< All literals are false on level 0 */
-		status_subsumed    = 17  /**< At least one literal is true on level 0 */
+		// BASE STATUS
+		status_open         = 0,  /**< Clause is neither sat, unsat, or unit.       */
+		status_sat          = 1,  /**< At least one literal is true.                */
+		status_unsat        = 2,  /**< All literals are false.                      */
+		status_unit         = 4,  /**< All but one literal false.                   */
+		// COMPLEX STATUS
+		status_sat_asserting= 5,  /**< Sat but literal is implied on lower dl.      */
+		status_asserting    = 6,  /**< Unsat but literal is implied on second highest dl. */
+		status_subsumed     = 9,  /**< Sat and one literal is true on level 0.      */
+		status_empty        = 10, /**< Unsat and all literals are false on level 0. */
 	};
-
-	//! Returns the current status of the clause.
-	/*!
-	 * 
-	 * \return Given a clause with literals [l1,...,ln] returns
-	 *  - status_empty      : if n == 0 or for all li, (s.isFalse(li) && s.level(li) == 0)
-	 *  - status_subsumed   : if exists li, s.isTrue(li) && s.level(li) == 0
-	 *  - status_sat        : if exists li, s.isTrue(li)
-	 *  - status_conflicting: if for all li, s.isFalse(li) and for some lj, s.level(lj) > 0
-	 *  - status_unit       : if s.isFree(li) and for all j != i, s.isFalse(lj)
-	 *  - status_asserting  : if exists li, s.th. for all lj != li, s.isFalse(lj) and s.level(lj) != s.level(li).
-	 *  - status_open       : else
-	 */
-	Status status() const;
-
-	//! Returns the implication level of the unit-resulting literal.
-	/*! 
-	 * \pre  status() == status_unit || status() == status_asserting
-	 */
-	uint32 implicationLevel() const { return swLev_; }
-	//! Returns the (numerical) highest DL on which the clause is conflicting.
-	/*!
-	 * \pre status() == status_conflicting || status_empty
-	 */
-	uint32 conflictLevel()    const { return fwLev_; }
-
-	//! Removes duplicate lits. Marks the clause as sat if it contains p and ~p.
-	void simplify();
-	
 	//! A type for storing the result of a clause insertion operation.
 	struct Result {
 		explicit Result(ClauseHead* loc = 0, Status st = status_open)
@@ -186,46 +119,97 @@ public:
 		ClauseHead*     local;
 		Status          status;
 		//! Returns false is clause is conflicting w.r.t current assignment.
-		bool     ok()   const { return status != status_conflicting; }
+		bool     ok()   const { return (status & status_unsat) == 0; }
+		//! Returns true if the clause implies a literal (possibly after backtracking).
 		bool     unit() const { return (status & status_unit) != 0; }
 		operator bool() const { return ok(); }
 	};
-
-	//! Adds the clause to the solver if possible.
+	//! Starts the creation of a new clause.
 	/*!
-	 * Adds the clause only if it is not conflicting.
+	 * \pre s.decisionLevel() == 0 || t != Constraint_t::static_constraint
+	 */
+	ClauseCreator& start(ConstraintType t = Constraint_t::static_constraint);
+	//! Sets the initial activity of the clause under construction.
+	ClauseCreator& setActivity(uint32 a)      { extra_.setActivity(a);  return *this; }
+	//! Sets the initial literal block distance of the clause under construction.
+	ClauseCreator& setLbd(uint32 lbd)         { extra_.setLbd(lbd); return *this; }
+	//! Adds the literal p to the clause under construction.
+	ClauseCreator& add(const Literal& p)      { literals_.push_back(p); return *this; }
+	//! Removes subsumed lits and orders first lits w.r.t watch order.
+	ClauseRep      prepare(bool fullSimplify);
+	//! Returns the current size of the clause under construction.
+	uint32         size()    const { return (uint32)literals_.size(); }
+	//! Returns the literal at the given idx.
+	Literal&       operator[](uint32 i)       { return literals_[i]; }
+	Literal        operator[](uint32 i) const { return literals_[i]; }
+	//! Returns the literals of the clause under construction.
+	const LitVec&  lits()    const { return literals_; }
+	LitVec&        lits()          { return literals_; }
+	//! Returns the clause's type.
+	ConstraintType type()    const { return extra_.type(); }
+	//! Returns the aux info of the clause under construction.
+	ClauseInfo     info()    const { return extra_; }
+	//! Creates a new clause object for the clause under construction.
+	/*!
+	 * \pre The clause does not contain duplicate/complementary literals or 
+	 *      flags contains clause_force_simplify.
+	 * 
 	 * \note If the clause to be added is empty, end() fails and s.hasConflict() is set to true.
+	 * \see Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& info);
 	 */
-	Result end();
-
-	//! Reorders lits such that the first two literals are valid watches.
-	/*!
-	 * \note 
-	 *   If allFree is true, all literals in lits shall be free w.r.t
-	 *   to s. In that case, the watches are picked according to
-	 *   the strategy stored in s.
-	 */
-	static Status initWatches(Solver& s, LitVec& lits, bool allFree);
-
-	//! Returns the status of the given clause w.r.t s.
-	static Status status(const Solver& s, const Literal* clause_begin, const Literal* clause_end);
+	Result end(uint32 flags = clause_not_sat | clause_not_conflict);
 
 	/*!
 	 * \name factory functions
 	 * Functions for creating and integrating clauses
 	 */
-		
+	//@{
 	//! Flags controlling clause creation and integration.
 	enum CreateFlag {
-		clause_no_add        = 1,  /**< do not add clause to solver db */
-		clause_explicit      = 2,  /**< force creation of explicit clause even if size <= 3 */
-		clause_not_sat       = 4,  /**< integrate only if clause is not satisfied w.r.t current assignment */
-		clause_not_root_sat  = 8,  /**< integrate only if clause is not satisfied w.r.t root level */
-		clause_not_conflict  = 16, /**< integrate only if clause is not conflicting */
-		clause_no_release    = 32, /**< do not call release on shared literals */
-		clause_int_lbd       = 64, /**< compute lbd when integrating asserting clauses */
-		clause_known_order   = 128,/**< assume clause is already ordered w.r.t watches */
-	};
+		// REPRESENTATION
+		clause_no_add        = 1,  /**< Do not add clause to solver db. */
+		clause_explicit      = 2,  /**< Force creation of explicit clause even if size <= 3. */
+		// STATUS
+		clause_not_sat       = 4,  /**< Do not add clause if it is satisfied (but not asserting) w.r.t current assignment. */
+		clause_not_root_sat  = 8,  /**< Do not add clause if it is satisfied w.r.t the root assignment.      */
+		clause_not_conflict  = 16, /**< Do not add clause if it is conflicting w.r.t the current assignment. */
+		// INTEGRATE
+		clause_no_release    = 32, /**< Do not call release on shared literals.         */
+		clause_int_lbd       = 64, /**< Compute lbd when integrating asserting clauses. */
+		// PREPARE
+		clause_no_prepare    = 128,/**< Assume clause is already ordered w.r.t watches. */
+		clause_force_simplify= 256,/**< Call simplify() on create.                      */
+		clause_no_heuristic  = 512 /**< Do not notify heuristic about new clause.       */
+	};	
+	//! Returns the status of the given clause w.r.t s.
+	static Status    status(const Solver& s, const Literal* clause_begin, const Literal* clause_end);
+	static Status    status(const Solver& s, const ClauseRep& c);
+	
+	//! Returns an abstraction of p's decision level that can be used to order literals.
+	/*!
+	 * The function returns a value, s.th 
+	 * order(any true literal) > order(any free literal) > order(any false literal).
+	 * Furthermore, for equally assigned literals p and q, order(p) > order(q), iff
+	 * level(p) > level(q).
+	 */
+	static uint32    watchOrder(const Solver& s, Literal p);
+	
+	//! Prepares the clause given in lits.
+	/*!
+	 * A prepared clause [l1...ln] with n >= 2 is a clause that,
+	 *  - does not contain any duplicate or complementary literals and, 
+	 *  - does not contain any subsumed literals (i.e. literals assigned on decision level 0) and,
+	 *  - is partially ordered w.r.t watchOrder(), i.e., watchOrder(l1) >= watchOrder(l2) and there
+	 *    is no lj, j > 2, s.th. watchOrder(lj) > watchOrder(l2)
+	 *  . 
+	 *
+	 * Removes subsumed literals from lits and reorders lits s.th. 
+	 * the first literals are valid watches. Furthermore, 
+	 * if flags contains clause_force_simplify, 
+	 * duplicate literals are removed from lits and tautologies are 
+	 * replaced with the single literal True. 
+	 */
+	static ClauseRep prepare(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& info = ClauseInfo());
 
 	//! Creates a clause from the literals given in lits.
 	/*!
@@ -235,54 +219,42 @@ public:
 	 * \param info  Initial information (e.g. type) for the new clause.
 	 *
 	 * \pre !s.hasConflict() and s.decisionLevel() == 0 or extra.learnt()
-	 * \pre lits does not contain duplicate/complementary literals
+	 * \pre lits is fully prepared or flags contains suitable prepare flags.
 	 *
 	 * \note 
 	 *   If the given clause is unit (or asserting), the unit-resulting literal is
 	 *   asserted on the (numerical) lowest level possible but the new information
 	 *   is not immediately propagated, i.e. on return queueSize() may be greater than 0.
-	 *   If the given clause is conflicting, the clause is stored as a conflict and 
-	 *   resolveConflict() must be called in order to proceed.
 	 *
 	 * \note 
 	 *   The local representation of the clause is always attached to the solver 
-	 *   but only added to the solver if clause_no_add is not contained in modeFlags.
-	 *   Otherwise, the returned clauses is owned by the caller
+	 *   but only added to the solver if clause_no_add is not contained in flags.
+	 *   Otherwise, the returned clause is owned by the caller
 	 *   and it is the caller's responsibility to manage it. Furthermore, 
 	 *   learnt statistics are *not* updated automatically in that case.
 	 *
+	 * \see prepare()
 	 */
-	static Result create(Solver& s, LitVec& lits, uint32 flags = 0, const ClauseInfo& info = ClauseInfo());
-	/*!
-	 * \overload Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& info);
-	 * \pre lits is empty or lits[0] is either free or true
-	 * \pre if lits[1] is false all lits[i], i > 1, are false and no literal was assigned on a higher dl than lits[1]
-	 * \note 
-	 *   The function always assumes clause_known_order, i.e. it is the caller's job
-	 *   to guarantee the first two literals in lits are valid watches.
-	 */
-	static Result create(Solver& s, LitVec& lits, const ClauseInfo& info = ClauseInfo()) {
-		return create(s, lits, clause_known_order, info);
-	}
-	
+	static Result create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& info = ClauseInfo());
+	static Result create(Solver& s, const ClauseRep& rep, uint32 flags);
 	//! Integrates the given clause into the current search of s.
 	/*!
 	 * \pre the assignment in s is not conflicting
-	 * \param s           The solver in which the clause should be integrated.
-	 * \param clause      The clause to be integrated.
-	 * \param modeFlags   A set of flags controlling integration (see CreateFlag).
-	 * \param t           Constraint type to use for the local representation.
+	 * \param s      The solver in which the clause should be integrated.
+	 * \param clause The clause to be integrated.
+	 * \param flags  A set of flags controlling integration (see CreateFlag).
+	 * \param t      Constraint type to use for the local representation.
 	 * 
 	 * \note 
 	 *   The function behaves similar to ClauseCreator::create() with the exception that
 	 *   it does not add local representations for implicit clauses (i.e. size <= 3) 
-	 *   unless modeFlags contains clause_explicit. 
+	 *   unless flags contains clause_explicit. 
 	 *   In that case, an explicit representation is created. 
-	 *   Implicit representations can only be created via ClauseCreator::createClause().
+	 *   Implicit representations can only be created via ClauseCreator::create().
 	 *
 	 * \note
 	 *   The function acts as a sink for the given clause (i.e. it decreases its reference count)
-	 *   unless modeFlags contains clause_no_release.
+	 *   unless flags contains clause_no_release.
 	 *   
 	 * \note integrate() is intended to be called in a post propagator. 
 	 *   To integrate a set of clauses F, one would use a loop like this:
@@ -296,104 +268,74 @@ public:
 	 *     return r;
 	 *   \endcode
 	 */
-	static Result integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags, ConstraintType t);
+	static Result integrate(Solver& s, SharedLiterals* clause, uint32 flags, ConstraintType t);
 	/*!
-	 * \overload Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags)
+	 * \overload Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 flags, ConstraintType t)
 	 */
-	static Result integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags);
+	static Result integrate(Solver& s, SharedLiterals* clause, uint32 flags);
 	//@}
-private:
-	void init(ConstraintType t);
-	struct Watches;
-	static inline Status status(uint32 DL, uint32 fw, uint32 sw);
-	static inline Status status(const Solver& s, const Watches& w, uint32 modeFlags);
-	static ClauseHead* newProblemClause(Solver& s, LitVec& lits, const ClauseInfo& e, uint32 flags);
-	static ClauseHead* newLearntClause(Solver& s,  LitVec& lits, const ClauseInfo& e, uint32 flags);
-	static ClauseHead* newUnshared(Solver& s, SharedLiterals* clause, const Watches& w, const ClauseInfo& e);
-	Solver*    solver_;    // solver in which new clauses are stored
-	LitVec     literals_;  // literals of the new clause
-	ClauseInfo extra_;     // extra info 
-	uint32     fwLev_;     // decision level of the first watched literal
-	uint32     swLev_;     // decision level of the second watched literal
+private:	
+	static ClauseRep   prepare(Solver& s, const Literal* in, uint32 inSize, const ClauseInfo& e, uint32 flags, Literal* out, uint32 outMax = UINT32_MAX);
+	static Result      create_prepared(Solver& s, const ClauseRep& pc, uint32 flags);
+	static ClauseHead* newProblemClause(Solver& s, const ClauseRep& clause, uint32 flags);
+	static ClauseHead* newLearntClause(Solver& s, const ClauseRep& clause, uint32 flags);
+	static ClauseHead* newUnshared(Solver& s, SharedLiterals* clause, const Literal* w, const ClauseInfo& e);
+	static bool        ignoreClause(const Solver& s, const ClauseRep& cl, Status st, uint32 modeFlags);
+	Solver*    solver_;   // solver in which new clauses are stored
+	LitVec     literals_; // literals of the new clause
+	ClauseInfo extra_;    // extra info 
+	uint32     flags_;    // default flags to be used in end()
 };
 
 //! Class for representing a clause in a solver.
 class Clause : public ClauseHead {
 public:
-	typedef Constraint::PropResult PropResult;
+	typedef Constraint::PropResult   PropResult;
 	enum { MAX_SHORT_LEN = 5 };
-	static void* alloc(Solver& s, uint32 lits, bool learnt);
 	
+	//! Allocates memory for storing a (learnt) clause with nLits literals.
+	static void* alloc(Solver& s, uint32 mLits, bool learnt);
+	
+	//! Creates a new clause from the clause given in rep.
+	/*!
+	 * \param s   Solver in which the new clause is to be used.
+	 * \param rep The raw representation of the clause.
+	 *
+	 * \pre The clause given in lits is prepared and contains at least two literals.
+	 * \note The clause must be destroyed using Clause::destroy.
+	 * \see ClauseCreator::prepare()
+	 */ 
+	static ClauseHead*  newClause(Solver& s, const ClauseRep& rep) {
+		return newClause(alloc(s, rep.size, rep.info.learnt()), s, rep);
+	}
 	//! Creates a new clause object in mem.
 	/*!
-	 * \pre mem points to a memory block that was allocated via Clause::alloc(s, size)
+	 * \pre mem points to a memory block that was allocated via Clause::alloc().
 	 */
-	static ClauseHead*  newClause(void* mem, Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra);
-
-	static ClauseHead*  newClause(Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra) {
-		return newClause(alloc(s, size, extra.learnt()), s, lits, size, extra);
-	}
+	static ClauseHead*  newClause(void* mem, Solver& s, const ClauseRep& rep);
 
+	//! Creates a new contracted clause from the clause given in rep.
 	/*!
-	 * Creates a new problem clause from the literals in lits.
-	 * 
-	 * \param s Solver in which the new clause is to be used.
-	 * \param lits The literals of the clause.
-	 *
-	 * \pre lits contains at least two literals and:
-	 *  - all literals are currently free.
-	 *  - lits does not contain any duplicate literals.
-	 *  - if lits contains literal p it must not contain ~p.
-	 *  . 
-	 * \note The clause must be destroyed using Clause::destroy.
-	 */ 
-	static ClauseHead*  newClause(Solver& s, const LitVec& lits) {
-		return newClause(s, &lits[0], static_cast<uint32>(lits.size()), ClauseInfo());
-	}
-	
-	/*!
-	 * Creates a new learnt clause from the literals in lits.
-	 * 
-	 * \param s     Solver in which the new clause is to be used.
-	 * \param lits  The literals of the clause.
-	 * \param extra Initial extra data of the learnt constraint.
-	 *
-	 * \pre lits contains at least two literals and:
-	 *  - lits does not contain any duplicate/complementary literals.
-	 *  - lits[0] and lits[1] are valid watches, i.e. 
-	 *    - they are either both not false or
-	 *    - lits[1] is false AND
-	 *      - for each l != lits[0] in lits, isFalse(l) && level(l) <= level(lits[1])
-	 *      - !isFalse(lits[0]) OR level(lits[0]) >= level(lits[1])
-	 *  . 
-	 * \note The clause must be destroyed using Clause::destroy
-	 */ 
-	static ClauseHead*  newLearntClause(Solver& s, const LitVec& lits, const ClauseInfo& extra) {
-		return newClause(s, &lits[0], static_cast<uint32>(lits.size()), extra);
-	}
-
-	/*!
-	 * Creates a new contracted clause from the literals contained in lits. A contracted clause
-	 * consists of an active head and a (false) tail. Propagation is restricted to the head. 
+	 * A contracted clause consists of an active head and a tail of false literals.
+	 * Propagation is restricted to the head. 
 	 * The tail is only needed to compute reasons from assignments.
 	 * 
 	 * \param s     Solver in which the new clause is to be used.
-	 * \param lits  The literals of the clause.
-	 * \param extra Initial extra data of the clause.
+	 * \param rep   The raw representation of the clause.
 	 * \param tail  The starting index of the tail (first literal that should be temporarily removed from the clause).
 	 * \param exten Extend head part of clause as tail literals become free?
 	 */ 
-	static ClauseHead*  newContractedClause(Solver& s, const LitVec& lits, const ClauseInfo& extra, LitVec::size_type tail, bool extend);
-
+	static ClauseHead*  newContractedClause(Solver& s, const ClauseRep& rep, uint32 tailPos, bool extend);
+	
 	//! Creates a new local surrogate for shared_lits to be used in the given solver.
 	/*!
-	 * \param s The solver in which this clause will be used.
-	 * \param shared_lits The shared literals of this clause.
-	 * \param e Initial data for the new (local) clause.
-	 * \param lits Watches and cache literal for the new (local) clause.
-	 * \param addRef Increment ref count of shared_lits.
+	 * \param s      The solver in which this clause will be used.
+	 * \param lits   The shared literals of this clause.
+	 * \param e      Initial data for the new (local) clause.
+	 * \param head   Watches and cache literal for the new (local) clause.
+	 * \param addRef Increment ref count of lits.
 	 */
-	static ClauseHead* newShared(Solver& s, SharedLiterals* shared_lits, const ClauseInfo& e, const Literal* lits, bool addRef = true);
+	static ClauseHead* newShared(Solver& s, SharedLiterals* lits, const ClauseInfo& e, const Literal head[3], bool addRef = true);
 
 	// Constraint-Interface
 	
@@ -435,7 +377,7 @@ public:
 	bool     strengthened()         const { return data_.local.strengthened(); }
 	uint32   computeAllocSize()     const;
 private:
-	Clause(Solver& s, const ClauseInfo& extra, const Literal* lits, uint32 size, uint32 tail, bool extend);
+	Clause(Solver& s, const ClauseRep& rep, uint32 tail = UINT32_MAX, bool extend = false);
 	Clause(Solver& s, const Clause& other);
 	typedef std::pair<Literal*, Literal*> LitRange;
 	void         undoLevel(Solver& s);
@@ -524,7 +466,8 @@ public:
 	Activity activity() const { return act_; }
 	
 	//! Halves the loop-formula's activity.
-	void decreaseActivity() { act_ = Activity(act_.activity()>>1, act_.lbd()); }
+	void decreaseActivity()          { act_ = Activity(act_.activity()>>1, act_.lbd()); }
+	void resetActivity(Activity hint){ act_ = hint; }
 	
 	//! Returns Constraint_t::learnt_loop.
 	ConstraintType type() const { return Constraint_t::learnt_loop; }
