@@ -168,6 +168,15 @@ public:
 	 * The default implementation returns 0.
 	 */
 	virtual ClauseHead* clause();
+
+	//! Shall return whether the constraint is valid (i.e. not conflicting) w.r.t the current assignment in s.
+	/*!
+	 * \pre The assignment in s is not conflicting and fully propagated.
+	 * \post A changed assignment if the assignment was not valid.
+	 * \note The default implementation always returns true and assumes
+	 *       that conflicts are detected by Constraint::propagate().
+	 */
+	virtual bool valid(Solver& s);
 protected:
 	virtual ~Constraint();
 private:
@@ -216,6 +225,8 @@ public:
 	
 	//! Asks the constraint to decrease its activity.
 	virtual void decreaseActivity();
+	//! Asks the constraint to reset its activity.
+	virtual void resetActivity(Activity hint);
 	
 	//! Shall return 0 if either !t.inSet(type()) or this constraint is satisfied w.r.t the current assignment.
 	/*!
@@ -227,27 +238,28 @@ protected:
 	~LearntConstraint();
 };
 
-//! Base class for all post propagators.
+//! Base class for post propagators.
 /*!
- * Post propagators are called after a solver finished unit-propagation and 
+ * Post propagators are called during propagation on each decision level and
  * once after a total assignment is found.
- * They may add more elaborate propagation mechanisms.
- * The typical asp example would be an unfounded set check.
+ * 
+ * They may extend a solver's unit-propagation with more elaborate propagation mechanisms.
+ * The typical asp example is an unfounded set check.
  *
- * \note Post propagators are called in priority order. For this,
- *       the function PostPropagator::priority() is used. 
- *       Currently, the solver distinguishes *three* classes of 
- *       post propagators:
- *       - det_single: deterministic post propagators that only work on 
- *         the current decision level. During propagation, these post 
- *         propagators must neither backtrack nor create new levels. 
- *         They shall have a priority in the range: [priority_highest, priority_lookahead)
- *       - det_multi: post propagators that use some form of lookahead. They are only
- *         called once all det_single post propagators finished. Furthermore,
- *         they are disabled during lookahead operations. They shall have
- *         a priority in the range [priority_lookahead, priority_general).
- *       - multi: post propagators that are non-deterministic or work on many
- *         decision levels shall have a priority in the range [priority_general, priority_lowest).
+ * \note Currently, the solver distinguishes *two* classes of post propagators:
+ *       - class_simple: deterministic post propagators shall only extend
+ *         the current decision level. That is, given DL, the decision level on which propagation started,
+ *         these post propagators shall neither backtrack below DL nor permanently add new decision levels. 
+ *         Deterministic post propagators are called in priority order. For this,
+ *         the function PostPropagator::priority() is used and shall return a priority in the range: [priority_class_simple, priority_class_general)
+ *       - class_general: post propagators that are non-deterministic or those that are not limited to extending
+ *         the current decision level shall have a priority of priority_class_general. They are called in LIFO order but
+ *         only after *all* simple post propagators have reached a fixpoint. 
+ *
+ * \note There are currently three reserved priority values, namely
+ *       priority_reserved_msg for message and termination handler (if any),
+ *       priority_reserved_ufs for the default unfounded set checker (if any),
+ *       and priority_reserved_look for the default lookahead operator (if any).
  */
 class PostPropagator : public Constraint {
 public:
@@ -255,88 +267,84 @@ public:
 	virtual ~PostPropagator();
 	using Constraint::propagate;  // Enable overloading!
 
-	PostPropagator* next; // for supporting lists of post propagators
-	//! Default priorities for post propagators
+	PostPropagator* next; // main propagation lists of post propagators
+	//! Default priorities for post propagators.
 	enum Priority {
-		priority_highest     = 0,   /**< highest possible priority  */
-		priority_single_high = 50,  /**< high det_single priority   */
-		priority_single      = 100, /**< default priority           */
-		priority_single_low  = 150, /**< low det_single priority    */
-		priority_lookahead   = 200, /**< default lookahead priority */
-		priority_general     = 300, /**< default general priority   */
-		priority_lowest      = 1024 /**< lowest possible priority   */
+		priority_class_simple    = 0,    /**< Starting priority of simple post propagators. */
+		priority_reserved_msg    = 0,    /**< Reserved priority for message/termination handlers (if any). */
+		priority_reserved_ufs    = 10,   /**< Reserved priority for the default unfounded set checker (if any). */
+		priority_reserved_look   = 1023, /**< Reserved priority for the default lookahead operator (if any). */
+		priority_class_general   = 1024, /**< Priortiy of extended post propagators. */
 	};
-	//! Returns a value representing the priority of this propagator.
+
+	//! Shall return a value representing the priority of this propagator.
 	/*!
 	 * The priority is used to order sequences of post propagators and to
-	 * classify post propagators w.r.t the three classes: det_single, det_multi, and
-	 * multi.
-	 * \note The default implementation returns priority_single.
-	 * \note See class description for an overview of the three priority classes.
+	 * classify post propagators w.r.t the classes: class_simple and class_general.
+	 * \note See class description for an overview of the two priority classes.
 	 */
-	virtual uint32 priority() const;
+	virtual uint32 priority() const = 0;
 
-	//! Called once after all constraints were added to solver.
+	//! Called during initialization of s.
+	/*!
+	 * \note During initialization a post propagator may assign variables
+	 *       but it must not yet propagate them.
+	 */
 	virtual bool   init(Solver& s);
-
-	//! Shall enqueue new assignments to be propagated.
+	
+	//! Shall enqueue and propagate new assignments implied by this propagator.
 	/*!
-	 * \pre    the assignment is fully propagated.
-	 * \param  s The solver in which this object is used.
-	 * \return 
-	 *   - false if propagation led to conflict
-	 *   - true otherwise
-	 *
+	 * This function shall enqueue and propagate all assignments currently implied by 
+	 * this propagator until a fixpoint is reached w.r.t this post propagator or 
+	 * a conflict is detected.
+	 * 
+	 * \pre   The assignment is fully propagated w.r.t any previous post propagator.
+	 * \param s    The solver in which this post propagator is used.
+	 * \param ctx  The post propagator from which this post propagator is called or
+	 *             0 if no other post propagator is currently active.
+	 * \post  s.queueSize() == 0 || s.hasConflict()
+	 * \return false if propagation led to conflict, true otherwise
+	 * 
 	 * \note 
-	 *   The function is called in the context of
-	 *   PostPropagator::propagateFixpoint(Solver& s).
-	 *   It must not call s.propagate() or any other function
-	 *   that would result in a recursive chain of propagate()
-	 *   calls.
-	 *
-	 * \see PostPropagator::propagateFixpoint(Solver& s)
+	 *   The function shall not call Solver::propagate()
+	 *   or any other function that would result in a recursive chain 
+	 *   of propagate() calls. On the other hand, it shall call
+	 *   Solver::propagateUntil(this) after enqueuing new assignments 
+	 *   to initiate propagation up to this propagator.
+	 * 
+	 * Typically, propagateFixpoint() should implemet a loop like this:
+	 * \code
+	 * for (;;) {
+	 *   if (!assign_newly_implied_literals(s)){ return false; }
+	 *   if (s.queueSize() == 0)               { return true;  }
+	 *   if (!s.propagateUntil(this))          { return false; }
+	 * }
+	 * \endcode
 	 */
-	virtual bool propagate(Solver& s) = 0;
+	virtual bool   propagateFixpoint(Solver& s, PostPropagator* ctx) = 0;
 
-	//! Clears internal state of this object and aborts any active propagation.
+	//! Aborts an active propagation operation.
 	/*!
-	 * The solver containing this object calls reset() whenever a conflict is
-	 * detected during propagation.
+	 * The function reset() is called whenever propagation on the
+	 * current decision level is stopped before a fixpoint is reached.
+	 * In particular, a solver calls reset() when a conflict is detected 
+	 * during propagation.
+	 *
 	 * \note The default implementation is a noop.
 	 */
-	virtual void reset();
+	virtual void   reset();
 
 	//! Is the current total assignment a model?
 	/*!
-	 * \pre the assignment is total and not conflicting
+	 * \pre The assignment is total and not conflicting.
 	 * \return
 	 *  - true if the assignment is a model w.r.t this post propagator
 	 *  - false otherwise
-	 * \post If the function returned true, s.numFreeVars() == 0 && !s.hasConflict().
-	 *       Otherwise, s.numFreeVars() > 0 || s.hasConflict().
+	 * \post If the function returned true:  s.numFreeVars() == 0 && !s.hasConflict().
+	 *       If the function returned false: s.numFreeVars() > 0 || s.hasConflict().
+	 * \note The default implementation returns Constraint::valid(s);
 	 */
-	virtual bool isModel(Solver& s);
-
-	//! Propagates the current assignment until a fixpoint is reached w.r.t this post propagator.
-	/*!
-	 * This function implements a basic propagation loop that calls
-	 * PostPropagator::propagate(Solver& s) until a fixpoint is reached or
-	 * a conflict is detected.
-	 * \note
-	 *   Normally, subclasses only need to implement 
-	 *   PostPropagator::propagate(Solver& s).
-	 *
-	 * \param  s The solver in which this object is used.
-	 * \return false if propagation led to conflict, true otherwise
-	 * \pre    The assignment is fully propagated.
-	 * \post   A fixpoint was reached or propagation led to a conflict.
-	 * \note   The default implementation implements the following loop:
-	 *   REPEAT:
-	 *     if (!propagate())            return false;
-	 *     if (s.queueSize() == 0)      return true;
-	 *     if (!s.propagateUntil(this)) return false;
-	 */
-	virtual bool propagateFixpoint(Solver& s);
+	virtual bool   isModel(Solver& s);
 protected:
 	//! PostPropagators are not clonable by default.
 	Constraint* cloneAttach(Solver&) { return 0; }
@@ -347,6 +355,15 @@ private:
 	PostPropagator(const PostPropagator&);
 	PostPropagator& operator=(const PostPropagator&);
 };
+
+class MessageHandler : public PostPropagator {
+public:
+	MessageHandler();
+	virtual bool   handleMessages() = 0;
+	virtual uint32 priority()const { return PostPropagator::priority_reserved_msg; }
+	virtual bool   propagateFixpoint(Solver&, PostPropagator*) { return handleMessages(); }
+};
+
 //@}
 
 //! Stores a reference to the constraint that implied a literal.
