@@ -103,6 +103,7 @@ struct ScoreLook {
 	enum Mode { score_max, score_max_min };
 	typedef PodVector<VarScore>::type VarScores; /**< A vector of variable-scores */
 	ScoreLook() : best(0), mode(score_max), addDeps(true), nant(false) {}
+	bool    validVar(Var v) const { return v < score.size(); }
 	void    scoreLits(const Solver& s, const Literal* b, const Literal *e);
 	void    clearDeps();
 	uint32  countNant(const Solver& s, const Literal* b, const Literal *e) const;
@@ -110,7 +111,7 @@ struct ScoreLook {
 	bool    greaterMax(Var x, uint32 max) const { 
 		return score[x].nVal() > max || score[x].pVal() > max; 
 	}
-	bool     greaterMaxMin(Var lhs, uint32 max, uint32 min) const {
+	bool    greaterMaxMin(Var lhs, uint32 max, uint32 min) const {
 		uint32 lhsMin, lhsMax;
 		score[lhs].score(lhsMax, lhsMin);
 		return lhsMin > min || (lhsMin == min && lhsMax > max);
@@ -143,10 +144,20 @@ public:
 		body_lookahead,   /**< Test only bodies in both phases */
 		hybrid_lookahead, /**< Test atoms and bodies but only their preferred decision literal */
 	};
+	struct Params {
+		Params(Type t = atom_lookahead) : type(t), topLevelImps(true), restrictNant(false) {}
+		Params& lookahead(Type t){ type         = t; return *this; }
+		Params& addImps(bool b)  { topLevelImps = b; return *this; }
+		Params& nant(bool b)     { restrictNant = b; return *this; }
+		Type type;
+		bool topLevelImps;
+		bool restrictNant;
+	};
+	static bool isType(uint32 t) { return t != 0 && t <= hybrid_lookahead; }
 	/*!
 	 * \param t Lookahead-type to use.
 	 */
-	explicit Lookahead(Type t, bool topLevelImps = true);
+	explicit Lookahead(const Params& p);
 	~Lookahead();
 	
 	bool    init(Solver& s);
@@ -157,17 +168,18 @@ public:
 	//! Adds literal p to the lookahead list.
 	void    append(Literal p, bool testBoth);
 	//! Single-step lookahead on all vars in the loookahead list.
-	bool    propagateFixpoint(Solver& s);
-	//! Returns PostPropagator::priority_lookahead.
+	bool    propagateFixpoint(Solver& s, PostPropagator*);
+	//! Returns PostPropagator::priority_reserved_look.
 	uint32  priority() const;
 	void    destroy(Solver* s, bool detach);
 	//! Updates state with lookahead result.
 	ScoreLook score;
 	//! Returns "best" literal w.r.t scoring of last lookahead or posLit(0) if no such literal exists.
 	Literal heuristic(Solver& s);
+
+	void    setLimit(UnitHeuristic* n);
 protected:
-	friend class UnitHeuristic;
-	bool propagate(Solver& s); // called by propagateFixpoint
+	bool propagateLevel(Solver& s); // called by propagate
 	void undoLevel(Solver& s);
 	bool test(Solver& s, Literal p);
 private:
@@ -180,6 +192,7 @@ private:
 	};
 	typedef PodVector<NodeId>::type  UndoStack;
 	typedef PodVector<LitNode>::type LookList;
+	typedef UnitHeuristic*           HeuPtr;
 	void splice(NodeId n);
 	LitNode* node(NodeId n) { return &nodes_[n]; }
 	LitNode* head()         { return &nodes_[head_id]; } // head of circular candidate list
@@ -192,6 +205,7 @@ private:
 	NodeId     last_;  // last candidate in list; invariant: node(last_)->next == head_id;
 	NodeId     pos_;   // current lookahead start position
 	uint32     top_;   // size of top-level
+	HeuPtr     limit_; // 
 };
 
 //! Heuristic that uses the results of lookahead.
@@ -212,39 +226,21 @@ private:
 class UnitHeuristic : public DecisionHeuristic {
 public:
 	/*!
-	 * \param t Lookahead-type to use.
+	 * \param p Lookahead parameters to apply during lookahead.
 	 */
-	explicit UnitHeuristic(Lookahead::Type t);
+	explicit UnitHeuristic(const Lookahead::Params& p);
 	~UnitHeuristic();
+	//! Decorates the heuristic given in other with temporary lookahead of type t.
+	static UnitHeuristic* restricted(const Lookahead::Params& p, uint32 numOps, DecisionHeuristic* other);
 	void endInit(Solver& s);
-	void resurrect(const Solver&, Var v);
+	void updateVar(const Solver& s, Var v, uint32 n);
 	Literal doSelect(Solver& s);
-private:
-	Lookahead*      look_; // lookahead propagator
-	Lookahead::Type type_; // type of lookahead & heuristic
+	virtual bool notify(Solver&) { return true; }
+protected:
+	typedef SingleOwnerPtr<Lookahead> LookPtr;
+	LookPtr look_; // lookahead propagator
+	Solver* solver_;
 };
 
-class RestrictedUnit : public DecisionHeuristic {
-public:
-	~RestrictedUnit();
-	//! Extends the decision heuristic of s so that the next k choices are based on lookahead.
-	static  void decorate(Solver& s, uint32 k, Lookahead* look);
-private:
-	explicit RestrictedUnit(uint32 k, Lookahead* look, DecisionHeuristic* def);
-	void startInit(const Solver& s)        { default_->startInit(s); }
-	void resurrect(const Solver& s, Var v) { default_->resurrect(s, v); }
-	void simplify(const Solver&  s, LitVec::size_type st) { default_->simplify(s, st); }
-	void undoUntil(const Solver& s, LitVec::size_type st) { default_->undoUntil(s, st);}
-	void newConstraint(const Solver& s, const Literal* f, LitVec::size_type sz, ConstraintType t) { default_->newConstraint(s, f, sz, t); }
-	void updateReason(const Solver& s, const LitVec& lits, Literal resolveLit) { default_->updateReason(s, lits, resolveLit); }
-	bool bump(const Solver& s, const WeightLitVec& lits, double adj)           { return default_->bump(s, lits, adj); }
-	Literal selectRange(Solver& s, const Literal* first, const Literal* last)  { return default_->selectRange(s, first, last); }
-	void    endInit(Solver& s);
-	Literal doSelect(Solver& s);
-	void    destroy(Solver& s);
-	Lookahead*         look_;
-	DecisionHeuristic* default_;
-	uint32             numChoices_;
-};
 }
 #endif
