@@ -43,35 +43,41 @@ namespace Clasp { namespace mt {
 class ParallelHandler;
 class ParallelSolve;
 
-struct ParallelSolveOptions {
-	enum SearchMode      { mode_split = 0, mode_compete  = 1 };
-	ParallelSolveOptions() : mode(mode_compete) {}
+struct ParallelSolveOptions : BasicSolveOptions {
+	typedef Distributor::Policy Distribution;
+	ParallelSolveOptions() {}
+	struct Algorithm {
+		enum SearchMode { mode_split = 0, mode_compete  = 1 };
+		Algorithm() : threads(1), mode(mode_compete) {}
+		uint32     threads;
+		SearchMode mode;
+	};
 	struct Integration { /**< Nogood integration options. */
+		static const uint32 GRACE_MAX = (1u<<28)-1;
 		Integration() :  grace(1024), filter(filter_gp), topo(topo_all) {}
 		enum Filter   { filter_no = 0, filter_gp = 1, filter_sat = 2, filter_heuristic = 3 };
 		enum Topology { topo_all  = 0, topo_ring = 1, topo_cube  = 2, topo_cubex = 3       };
 		uint32 grace : 28; /**< Lower bound on number of shared nogoods to keep. */
 		uint32 filter: 2;  /**< Filter for integrating shared nogoods (one of Filter). */   
 		uint32 topo  : 2;  /**< Integration topology */
-	};          
+	};
 	struct GRestarts {   /**< Options for configuring global restarts. */
 		GRestarts():maxR(0) {}
 		uint32           maxR;
 		ScheduleStrategy sched;
 	};
-	SolveLimits  limit;     /**< Solve limit (disabled by default). */
-	Integration  integrate; /**< Nogood integration parameters. */
-	GRestarts    restarts;  /**< Global restart strategy. */
-	SearchMode   mode;
-	//! Returns a newly allocated solve object in out.
-	/*!
-	 * \pre sc points to an array of size >= ctx.numSolvers().
-	 */
-	void createSolveObject(SolveAlgorithm*& out, SharedContext& ctx, SolverConfig** sc) const;
+	Integration  integrate; /**< Nogood integration parameters.     */
+	Distribution distribute;/**< Nogood distribution parameters.    */
+	GRestarts    restarts;  /**< Global restart strategy.           */
+	Algorithm    algorithm; /**< Parallel algorithm to use.         */
+	//! Allocates a new solve object.
+	SolveAlgorithm* createSolveObject() const;
 	//! Returns the number of threads that can run concurrently on the current hardware.
-	static uint32   recommendedSolvers() { return Clasp::thread::hardware_concurrency(); }
+	static uint32   recommendedSolvers()     { return Clasp::thread::hardware_concurrency(); }
 	//! Returns number of maximal number of supported threads.
-	static uint32   supportedSolvers()   { return 64; }
+	static uint32   supportedSolvers()       { return 64; }
+	uint32          numSolver()        const { return algorithm.threads; }
+	bool            defaultPortfolio() const { return algorithm.mode == Algorithm::mode_compete; }
 };
 
 //! A parallel algorithm for multi-threaded solving with and without search-space splitting.
@@ -84,33 +90,26 @@ struct ParallelSolveOptions {
  */
 class ParallelSolve : public SolveAlgorithm {
 public:
-	//! Initializes this object for solving with at most ctx.numSolvers() threads.
-	/*!
-	 * \pre ctx.numSolvers() > 1
-	 */
-	ParallelSolve(SharedContext& ctx, const ParallelSolveOptions& opts);
+	explicit ParallelSolve(Enumerator* e, const ParallelSolveOptions& opts);
 	~ParallelSolve();
+	// base interface
+	virtual bool interrupted() const;
+	virtual void resetSolve();
+	virtual void enableInterrupts();
 	// own interface
-
-	//! Attaches a new thread to s and runs it in this algorithm.
-	/*!
-	 * Shall be called once for each solver except the master.
-	 * \pre s.id() != Solver::invalidId
-	 */
-	void   addSolver(Solver& s, const SolveParams& p);
-	
 	//! Returns the number of active threads.
 	uint32 numThreads()            const;
 	bool   integrateUseHeuristic() const { return test_bit(intFlags_, 31); }
 	uint32 integrateGrace()        const { return intGrace_; }
 	uint32 integrateFlags()        const { return intFlags_; }
-	bool   hasErrors()             const { return error_ != 0; }
-	//! Terminates current solving process and all client threads.
-	bool   terminate();
+	uint64 hasErrors()             const;
 	//! Requests a global restart.
 	void   requestRestart();
 	bool   handleMessages(Solver& s);
-	void   pushWork(LitVec& gp);
+	bool   integrateModels(Solver& s, uint32& mCount);
+	void   pushWork(LitVec* gp);
+	bool   commitModel(Solver& s);
+	bool   commitUnsat(Solver& s);
 	enum GpType { gp_none = 0, gp_split = 1, gp_fixed = 2 };
 private:
 	ParallelSolve(const ParallelSolve&);
@@ -128,37 +127,44 @@ private:
 	// Algorithm steps
 	void   setIntegrate(uint32 grace, uint8 filter);
 	void   setRestarts(uint32 maxR, const ScheduleStrategy& rs);
-	bool   doSolve(Solver& s, const SolveParams& p);
+	bool   beginSolve(SharedContext& ctx);
+	bool   doSolve(SharedContext& ctx, const LitVec& assume);
+	bool   doInterrupt();
 	void   solveParallel(uint32 id);
-	bool   initOpt(Solver& s, ValueRep last);
 	void   initQueue();
 	bool   requestWork(Solver& s, PathPtr& out);
-	bool   backtrackFromModel(Solver& s);
 	void   terminate(Solver& s, bool complete);
 	bool   waitOnSync(Solver& s);
 	void   exception(uint32 id, PathPtr& path, ErrorCode e, const char* what);
-	void   reportProgress(const SolveEvent& ev) const;
+	void   reportProgress(const Event& ev) const;
 	// -------------------------------------------------------------------------------------------
+	typedef ParallelSolveOptions::Distribution Distribution;
 	struct SharedData;
 	// SHARED DATA
 	SharedData*       shared_;       // Shared control data
 	ParallelHandler** thread_;       // Thread-locl control data
 	// READ ONLY
+	Distribution      distribution_; // distribution options
 	uint32            maxRestarts_;  // disable global restarts once reached 
-	uint32            intGrace_;     // grace period for clauses to integrate
+	uint32            intGrace_ : 30;// grace period for clauses to integrate
+	uint32            intTopo_  :  2;// integration topology
 	uint32            intFlags_;     // bitset controlling clause integration
-	int               error_;
 	GpType            initialGp_;
 };
 
+struct MessageEvent : SolveEvent<MessageEvent> {
+	enum Action { sent, received, completed };
+	MessageEvent(const Solver& s, const char* message, Action a, double t = 0.0) 
+		: SolveEvent<MessageEvent>(s, verbosity_high), msg(message), time(t) { op = (uint32)a; }
+	const char* msg;    // name of message
+	double      time;   // only for action completed
+};
 //! A per-solver (i.e. thread) class that implements message handling and knowledge integration.
 /*!
- * The class adds itself as a post propagator to the given solver. Each time
- * propagateFixpoint() is called (i.e. on each new decision level), it checks
- * for new lemmas to integrate and synchronizes the search with any new models.
- * Furthermore, it adds a second (high-priority) post propagator for message handling.
+ * The class adds itself as a post propagator to the given solver. During propagation
+ * it checks for new messages and lemmas to integrate.
  */
-class ParallelHandler : public PostPropagator {
+class ParallelHandler : public MessageHandler {
 public:
 	typedef ParallelSolve::GpType GpType;
 
@@ -187,14 +193,12 @@ public:
 	int join() { if (joinable()) { thread_.join(); } return error(); }
 	
 	// overridden methods
-	
-	//! Returns a priority suited for a post propagators that is non-deterministic.
-	uint32 priority() const { return priority_general + 100; }
 
 	//! Integrates new information.
-	bool propagateFixpoint(Solver& s);
-	bool propagate(Solver& s) { return ParallelHandler::propagateFixpoint(s); }
-	void reset()              { up_ = 1; }
+	bool propagateFixpoint(Solver& s, PostPropagator*);
+	bool handleMessages() { return ctrl_->handleMessages(solver()); }
+	void reset() { up_ = 1; }
+	bool simplify(Solver& s, bool re);
 	//! Checks whether new information has invalidated current model.
 	bool isModel(Solver& s);
 
@@ -208,13 +212,13 @@ public:
 	bool hasPath()      const { return gp_.type != ParallelSolve::gp_none; }
 	void setGpType(GpType t)  { gp_.type = t; }
 	
-	//! Called before solver starts to solve given guiding path.
+	//! Entry point for solving the given guiding path.
 	/*!
-	 * \param gp      The new guiding path.
+	 * \param solve   The object used for solving.
 	 * \param type    The guiding path's type.
 	 * \param restart Request restart after restart number of conflicts.
 	 */
-	void prepareForGP(const LitVec& gp, GpType type, uint64 restart);
+	ValueRep solveGP(BasicSolve& solve, GpType type, uint64 restart);
 
 	/*!
 	 * \name Message handlers
@@ -243,52 +247,44 @@ public:
 	 */
 	bool handleRestartMessage();
 
-	SolveStats aggStats;  // aggregated statistics over all gps
 	Solver&            solver()      { return *solver_; }
 	const SolveParams& params() const{ return *params_; }
 	//@}  
 private:
-	bool simplify(Solver& s, bool re);
-	bool integrateClauses(Solver& s);
 	void add(ClauseHead* h);
 	void clearDB(Solver* s);
-	ParallelSolve* ctrl() const { return messageHandler_.ctrl; }
+	bool integrate(Solver& s);
 	typedef LitVec::size_type size_type;
 	typedef PodVector<Constraint*>::type ClauseDB;
-	Clasp::thread      thread_;     // active thread or empty for master
-	ClauseDB           integrated_; // my integrated clauses
+	typedef SharedLiterals**             RecBuffer;
+	enum { RECEIVE_BUFFER_SIZE = 32 };
+	ParallelSolve*     ctrl_;       // my message source
 	Solver*            solver_;     // my solver
 	const SolveParams* params_;     // my solving params
-	size_type          intTail_;    // where to put next clause
+	Clasp::thread      thread_;     // active thread or empty for master
+	RecBuffer          received_;   // received clauses not yet integrated
+	ClauseDB           integrated_; // my integrated clauses
+	size_type          recEnd_;     // where to put next received clause
+	size_type          intEnd_;     // where to put next clause
 	uint32             error_:30;   // error code or 0 if ok
-	uint32             win_  : 1;
-	uint32             up_   : 1;
+	uint32             win_  : 1;   // 1 if thread was the first to terminate the search
+	uint32             up_   : 1;   // 1 if next propagate should check for new lemmas/models
+	uint32             act_  : 1;   // 1 if gp is active
 	struct GP {
-		LitVec      path;     // current guiding path
-		uint64      restart;  // don't give up before restart number of conflicts
-		size_type   pos;      // pos in trail
-		uint32      impl;     // number of additional implied literals
-		GpType      type ;    // type of gp
+		uint64 restart;  // don't give up before restart number of conflicts
+		uint32 modCount; // integration counter for synchronizing models
+		GpType type ;    // type of gp
 		void reset(uint64 r = UINT64_MAX, GpType t = ParallelSolve::gp_none) {
-			path.clear();
 			restart = r;
-			pos     = 0;
-			impl    = 0;
+			modCount= 0;
 			type    = t;
 		}
 	} gp_;
-	struct MessageHandler : PostPropagator {
-		explicit MessageHandler(ParallelSolve* c) : ctrl(c) {}
-		uint32 priority() const { return PostPropagator::priority_highest; }
-		bool   propagateFixpoint(Solver& s) { return ctrl->handleMessages(s); }
-		bool   propagate(Solver& s)         { return MessageHandler::propagateFixpoint(s); }
-		ParallelSolve* ctrl; // get messages from here
-	}    messageHandler_;
 };
 
 class GlobalQueue : public Distributor {
 public:
-	explicit GlobalQueue(uint32 maxShare, uint32 topo);
+	explicit GlobalQueue(const Policy& p, uint32 maxShare, uint32 topo);
 	~GlobalQueue();
 	uint32  receive(const Solver& in, SharedLiterals** out, uint32 maxOut);
 	void    publish(const Solver& source, SharedLiterals* n);
