@@ -30,27 +30,23 @@
 
 namespace Clasp { 
 
-//! Parameter-Object for controlling search limits.
+//! Parameter-Object for managing search limits.
 struct SearchLimits {
 	explicit SearchLimits(uint64 conf = UINT64_MAX) 
 		: conflicts(conf)
 		, local(UINT64_MAX)
-		, learnts(UINT32_MAX)
-		, xLbd(0.0f)
-		, xCfl(0.0f) {
+		, memLimit(UINT64_MAX)
+		, dynamic(0)
+		, learnts(UINT32_MAX) {
 	}
-	bool   reached(const SolveStats& st) const { 
-		return conflicts == 0 || local == 0 || dynamicRestart(st);
-	}
-	bool   dynamicRestart(const SolveStats& st) const {
-		return st.queue && st.queue->full() 
-			&&   (((st.queue->avgLbd()*xLbd) > st.avgLbd()) || ((st.queue->avgCfl()*xCfl) > st.avgCfl()));
-	}
-	uint64 conflicts; /*!< Search for at least number of conflicts.             */
-	uint64 local;     /*!< Search for at least number of conflicts in branch.   */
-	uint32 learnts;   /*!< Limit on number of learnt lemmas.                    */
-	float  xLbd;      /*!< Adjustment factor for dynamic restarts based on LBD. */
-	float  xCfl;      /*!< Adjustment factor for dynamic restarts based on CFL. */
+	bool reached()           const { return conflicts == 0 || local == 0 || hasDynamicRestart(); }
+	bool hasDynamicRestart() const { return dynamic && dynamic->isRestart(); } 
+	void setMemLimit(uint32 limitInMB) { memLimit = limitInMB ? static_cast<uint64>(limitInMB)<<20 : UINT64_MAX; }
+	uint64    conflicts; /*!< Search for at least number of conflicts.             */
+	uint64    local;     /*!< Search for at least number of conflicts in branch.   */
+	uint64    memLimit;  /*!< (Soft) memory limit for learnt lemmas (in bytes).    */
+	SumQueue* dynamic;   /*!< Use dynamic restarts based on lbd or conflict level. */
+	uint32    learnts;   /*!< Limit on number of learnt lemmas.                    */
 };
 
 /**
@@ -84,100 +80,78 @@ struct SearchLimits {
  *
  */
 class Solver {
-	friend class SharedContext;
-	void startInit(SharedContext* ctx, uint32 constraintGuess);
-	bool endInit();
-	void setEnumerationConstraint(Constraint* c) {
-		if (enum_) enum_->destroy(this, true);
-		enum_ = c;
-	}
 public:
 	typedef PodVector<Constraint*>::type      ConstraintDB;
-	typedef std::auto_ptr<DecisionHeuristic>  Heuristic;
+	typedef const ConstraintDB&               DBRef;
+	typedef SingleOwnerPtr<DecisionHeuristic> Heuristic;
+private:
 	/*!
 	 * \name Construction/Destruction/Setup
 	 */
 	//@{
+	friend class SharedContext;
 	//! Creates an empty solver object with all strategies set to their default value.
-	Solver();
-	
+	Solver(SharedContext* ctx, uint32 id);
 	//! Destroys the solver object and all contained constraints.
 	~Solver();
-	
-	//! Sets the solver id - default is 0.
-	/*!
-	 * Setting an id is mainly useful in parallel solving.
-	 */
-	void   setId(uint32 id) { id_ = id; }
-	uint32 id() const       { return id_; }
-	//! Used to signal an error in this solver.
-	static const   uint32 invalidId = UINT32_MAX;
-	
-	//! Resets a solver object to the state it had after default-construction.
+	//! Resets a solver object to the state it had after construction.
 	void reset();
-
-	//! Shuffle constraints upon next simplification.
-	void shuffleOnNextSimplify() { shuffle_ = true; }
-	
-	//! Sets the decision heuristic to use in this object.
-	void setHeuristic(uint32 heuId, DecisionHeuristic* h) {
-		strategies().heuId = heuId;
-		heuristic_.reset(h);
-	}
-	
-	//! Returns the strategies used by this solver-object. 
-	SolverStrategies& strategies() { return strategy_; }
-	Heuristic&        heuristic()  { return heuristic_;}
-
-	/*!
-	 * \overload SolverStrategies& Solver::strategies()
-	 */
-	const SolverStrategies& strategies() const { return strategy_; }
-	const Heuristic&        heuristic()  const { return heuristic_;}
-
+	void resetId(uint32 id) { strategy.id = id; }
+	void startInit(uint32 constraintGuess);
+	bool cloneDB(const ConstraintDB& db);
+	bool preparePost();
+	bool endInit();
+	//@}
+public:
 	//! Returns a pointer to the shared context object of this solver.
-	const SharedContext* sharedContext() const { return shared_; }
-
+	const SharedContext*    sharedContext() const { return shared_; }
 	//! Returns a pointer to the sat-preprocessor used by this solver.
-	SatPreprocessor* satPrepro() const;
-	
-	//! Adds the unary constraint p to the solver.
-	/*!
-	 * \note Unary constraints are immediately asserted.
-	 * \return false if asserting p leads to a conflict.
-	 */
-	bool addUnary(Literal p, ConstraintType);	
-	//! Adds the short clause cl to the solver.
-	/*! 
-	 * \pre  cl is either binary or ternary (sz > 1 && sz < 4)
-	 * \note cl is stored in the short implication graph unless
-	 *       updates of this graph are not permitted.
-	 * \see  SharedContext::allowImplicit(ConstraintType)
-	 */
-	void addShort(const Literal* cl, uint32 sz, const ClauseInfo& extra);
-	//! Adds the binary clause [p q] by calling addShort.
-	void addBinary(Literal p, Literal q, const ClauseInfo& extra) { Literal x[2] = {p,q}; addShort(x, 2, extra); }
-	//! Adds the ternary clause [p q r] by calling addShort.
-	void addTernary(Literal p, Literal q, Literal r, const ClauseInfo& extra) { Literal x[3] = {p,q,r}; addShort(x, 3, extra); }
-	
+	SatPreprocessor*        satPrepro()     const;
+	//! Returns the configuration for this object.
+	const SolverParams&     configuration() const;
+	//! Returns the solve parameters for this object.
+	const SolveParams&      searchConfig()  const;
+	const Heuristic&        heuristic()     const { return heuristic_;}
+	uint32                  id()            const { return strategy.id; }
+	Heuristic&              heuristic()           { return heuristic_;}
+	VarInfo                 varInfo(Var v)  const { return shared_->validVar(v) ? shared_->varInfo(v) : VarInfo(); }
+	const SymbolTable&      symbolTable()   const { return shared_->symbolTable(); }
+	Literal                 tagLiteral()    const { return tag_; }
 	//! Adds the problem constraint c to the solver.
-	void add(Constraint* c);
-
+	/*!
+	 * Problem constraints shall only be added to the master solver of 
+	 * a SharedContext object and only during the setup phase.
+	 * \pre this == sharedContext()->master() && !sharedContext()->frozen().
+	 */
+	void                    add(Constraint* c);
+	//! Adds a suitable representation of the given clause to the solver.
+	/*!
+	 * Depending on the type and size of the given clause, the function
+	 * either adds a (learnt) constraint to this solver or an implication
+	 * to the shared implication graph.
+	 * \note If c is a problem clause, the precondition of add(Constraint* c) applies.
+	 */
+	bool                    add(const ClauseRep& c, bool isNew = true);
+	//! Returns whether c can be stored in the shared short implication graph.
+	bool                    allowImplicit(const ClauseRep& c) const {
+		return c.isImp() 
+		  ? shared_->allowImplicit(c.info.type()) && !c.info.aux() && (c.prep == 1 || (!auxVar(c.lits[0].var()) && !auxVar(c.lits[1].var()) && (c.size == 2 || !auxVar(c.lits[2].var()))))
+		  : c.size <= 1;
+	}
+	//! Returns the post propagator with the given priority or 0 if no such post propagator exists.
+	PostPropagator*         getPost(uint32 prio) const;
 	//! Adds p as post propagator to this solver.
 	/*!
 	 * \pre p was not added previously and is not part of any other solver.
 	 * \note Post propagators are stored and called in priority order.
 	 * \see PostPropagator::priority()
 	 */
-	void addPost(PostPropagator* p)    { post_.add(p); }
-
+	bool                    addPost(PostPropagator* p) { return addPost(p, initPost_ != 0); }
 	//! Removes p from the solver's list of post propagators.
 	/*!
-	 * \note removePost(p) shall only be called during propagation
-	 *       of p or if no propagation is currently active.
+	 * \note The function shall not be called during propagation of any other post propagator.
 	 */
-	void removePost(PostPropagator* p) { post_.remove(p); }
-	
+	void                    removePost(PostPropagator* p){ post_.remove(p); }
 	/*!
 	 * \name CDNL-functions.
 	 */
@@ -207,6 +181,12 @@ public:
 		return search(limit, rp);
 	}
 
+	//! Adds path to the current root-path and adjusts the root-level accordingly.
+	bool pushRoot(const LitVec& path);
+	bool pushRoot(Literal p);
+	
+	void setEnumerationConstraint(Constraint* c);
+
 	//! Moves the root-level i levels down (i.e. away from the top-level).
 	/*!
 	 * The root-level is similar to the top-level in that it cannot be
@@ -214,27 +194,30 @@ public:
 	 * above the root-level. 
 	 */
 	void pushRootLevel(uint32 i = 1) { 
-		rootLevel_  = std::min(decisionLevel(), rootLevel_+i); 
-		btLevel_    = std::max(btLevel_, rootLevel_);
+		levels_.root      = std::min(decisionLevel(), levels_.root+i);
+		levels_.backtrack = std::max(levels_.backtrack, levels_.root);
 	}
-	//! Returns the current root level.
-	uint32 rootLevel() const { return rootLevel_; }
-
+	
 	//! Moves the root-level i levels up (i.e. towards the top-level).
 	/*!
-	 * The function undos all levels between the new root level and the current decision level,
+	 * The function removes all levels between the new root level and the current decision level,
 	 * resets the current backtrack-level, and re-assigns any implied literals.
-	 * \param i Number of levels to pop.
+	 * \param      i      Number of root decisions to pop.
+	 * \param[out] popped Optional storage for popped root decisions.
+	 * \param      filter Whether or not aux variables should be added to popped.
 	 * \post decisionLevel() == rootLevel()
 	 * \note The function first calls clearStopConflict() to remove any stop conflicts.
 	 * \note The function *does not* propagate any asserted literals. It is
 	 *       the caller's responsibility to call propagate() after the function returned.
 	 */
-	bool popRootLevel(uint32 i = 1);
+	bool popRootLevel(uint32 i = 1, LitVec* popped = 0, bool aux = true);
 
 	//! Removes a previously set stop conflict and restores the root level.
 	void clearStopConflict();
 
+	//! Returns the current root level.
+	uint32 rootLevel() const { return levels_.root; }
+
 	//! Removes any implications made between the top-level and the root-level.
 	/*!
 	 * The function also resets the current backtrack-level and re-assigns learnt facts.
@@ -243,15 +226,60 @@ public:
 	 */
 	bool clearAssumptions();
 
+	//! Adds the learnt constraint c to the solver.
+	void addLearnt(LearntConstraint* c, uint32 size, ConstraintType type) {
+		learnts_.push_back(c);
+		stats.addLearnt(size, type); 
+	}
+	void addLearnt(LearntConstraint* c, uint32 size) { addLearnt(c, size, c->type()); }
+	//! Tries to receive at most maxOut clauses.
+	/*!
+	 * The function queries the distributor object for new clauses to be delivered to
+	 * this solver. Clauses are stored in out.
+	 * \return The number of clauses received.
+	 */
+	uint32          receive(SharedLiterals** out, uint32 maxOut) const;
+	//! Distributes the clause in lits via the distributor.
+	/*!
+	 * The function first calls the distribution strategy 
+	 * to decides whether the clause is a valid candidate for distribution.
+	 * If so and a distributor was set, it distributes the clause and returns a handle to the
+	 * now shared literals of the clause. Otherwise, it returns 0.
+	 *
+	 * \param owner The solver that created the clause.
+	 * \param lits  The literals of the clause.
+	 * \param size  The number of literals in the clause.
+	 * \param extra Additional information about the clause.
+	 * \note 
+	 *   If the return value is not null, it is the caller's 
+	 *   responsibility to release the returned handle (i.e. by calling release()).
+	 * \note If the clause contains aux vars, it is not distributed.
+	 */
+	SharedLiterals* distribute(const Literal* lits, uint32 size, const ClauseInfo& extra);
+	
+
+	//! Returns to the maximum of rootLevel() and backtrackLevel() and increases the number of restarts.
+	void restart() {
+		undoUntil(0);
+		++stats.restarts;
+		ccInfo_.setActivity(ccInfo_.activity() + 1);
+	}
+
 	//! Sets the backtracking level to dl.
 	void setBacktrackLevel(uint32 dl) {
-		btLevel_ = std::max(std::min(dl, decisionLevel()), rootLevel_);
+		levels_.backtrack = std::max(std::min(dl, decisionLevel()), rootLevel());
 	}
 	//! Returns the current backtracking level.
-	uint32 backtrackLevel() const { return btLevel_; }
+	uint32 backtrackLevel() const { return levels_.backtrack; }
 
 	//! Returns whether the solver can split-off work.
-	bool splittable() const { return decisionLevel() > rootLevel() && !frozenLevel(rootLevel()+1); }
+	bool splittable() const;
+
+	//! Tries to split-off disjoint work from the solver's currrent guiding path and returns it in out.
+	/*!
+	 * \return splittable()
+	 */
+	bool split(LitVec& out);
 
 	//! Copies the solver's currrent guiding path to gp.
 	/*!
@@ -260,11 +288,9 @@ public:
 	 *   - any literals that are implied on a level <= rootLevel() because of newly learnt
 	 *     information. This particularly includes literals that were flipped during model enumeration.
 	 * 
-	 * \param[in,out] gp          where to store the guiding path
-	 * \param[in,out] startPos    position in the trail from which copying should start
-	 * \param[in,out] numImplied  number of newly implied literals in gp
+	 * \param[out] gp where to store the guiding path
 	 */
-	void updateGuidingPath(LitVec& gp, LitVec::size_type& startPos, uint32& numImplied);
+	void copyGuidingPath(LitVec& out);
 
 	//! If called on top-level, removes SAT-clauses + Constraints for which Constraint::simplify returned true.
 	/*!
@@ -273,17 +299,30 @@ public:
 	 * \return false, if a top-level conflict is detected. Otherwise, true.
 	 */
 	bool simplify();
-	void simplifyDB(ConstraintDB& db);
+	//! Shuffle constraints upon next simplification.
+	void shuffleOnNextSimplify(){ shufSimp_ = 1; }
 
+	//! Requests a special aux variable for tagging conditional knowledge.
+	/*!
+	 * Once a tag variable t is set, learnt clauses containing ~t are
+	 * tagged as "conditional". Conditional clauses are removed once t becomes
+	 * unassigned or Solver::removeConditional() is called. Furthermore, calling 
+	 * Solver::strengthenConditional() removes ~t from conditional clauses and
+	 * transforms them to unconditional knowledge.
+	 *
+	 * \note Typically, the tag variable is a root assumption and hence true during 
+	 *       the whole search.
+	 */
+	Var  pushTagVar(bool pushToRoot);
 	//! Removes all conditional knowledge, i.e. all previously tagged learnt clauses.
 	/*!
-	 * \see SharedContext::requestTagLiteral()
+	 * \see Solver::pushTagVar()
 	 */
 	void removeConditional();
-
+	
 	//! Resolves all tagged clauses with the tag literal and thereby strengthens the learnt db.
 	/*!
-	 * \see SharedContext::requestTagLiteral()
+	 * \see Solver::pushTagVar()
 	 */
 	void strengthenConditional();
 
@@ -308,14 +347,14 @@ public:
 	 * conflict can be requested using the conflict() function.
 	 */
 	bool force(const Literal& p, const Antecedent& a) {
-		assert((!hasConflict() || isTrue(p)) && !shared_->eliminated(p.var()));
+		assert(!hasConflict() || isTrue(p));
 		if (assign_.assign(p, decisionLevel(), a)) return true;
 		setConflict(p, a, UINT32_MAX);
 		return false;
 	}
 	/*!
 	 * \overload bool Solver::force(const Literal&, const Antecedent&)
-	 * \pre data == UINT32_MAX || SharedContext::requestData(p.var()) was called during setup
+	 * \pre data == UINT32_MAX || requestData(p.var()) was called during setup
 	 */
 	bool force(const Literal& p, const Antecedent& a, uint32 data) {
 		return data != UINT32_MAX 
@@ -333,23 +372,23 @@ public:
 	 */
 	bool force(Literal p, uint32 dl, const Antecedent& r, uint32 d = UINT32_MAX) {
 		// Correct level?
-		if (dl == decisionLevel())      { return force(p, r, d); } 
-		ImpliedLiteral* x = 0;
+		if (dl == decisionLevel()) { return force(p, r, d); } 
 		// Already implied?
-		if (isTrue(p) && (level(p.var()) <= dl || (x = impliedLits_.find(p)) != 0)) {
-			if (level(p.var()) <= dl)     { return true; }
-			if (x->level       <= dl)     { return true; }
-			*x = ImpliedLiteral(p, dl, r, d);
-			return setReason(p, r, d); // r is stronger
+		if (isTrue(p)) {
+			if (level(p.var()) <= dl) { return true; }
+			if (ImpliedLiteral* x = impliedLits_.find(p)) {
+				return x->level <= dl || (*x = ImpliedLiteral(p, dl, r, d), setReason(p, r, d));
+			}
 		}
 		// Can we return to correct level?
 		if (undoUntil(dl, false) == dl) { return force(p, r, d); }
 		// Logically the implication is on level dl.
 		// Store enough information so that p can be re-assigned once we backtrack.
-		assert(!x);
 		impliedLits_.add(decisionLevel(), ImpliedLiteral(p, dl, r, d));
 		return (isTrue(p) && setReason(p, r, d)) || force(p, r, d);
 	}
+	//! Assigns p as a fact at decision level 0.
+	bool force(Literal p) { return force(p, 0, Antecedent(posLit(0))); }
 
 	//! Assumes the literal p if possible.
 	/*!
@@ -398,10 +437,12 @@ public:
 
 	/*!
 	 * Does unit propagation and calls x->propagateFixpoint(*this)
-	 * for all post propagators up to but not including p.
+	 * for all post propagators x up to but not including p.
 	 * \note The function is meant to be called only in the context of p.
+	 * \pre  p is a post propagator of this solver, i.e. was previously added via addPost().
+	 * \pre  Post propagators are active, i.e. the solver is fully initialized.
 	 */
-	bool propagateUntil(PostPropagator* p) { return unitPropagate() && (p == post_.head || post_.propagate(*this, p)); }
+	bool propagateUntil(PostPropagator* p) { assert((!p || post_.active()) && "OP not allowed during init!"); return unitPropagate() && (p == post_.active() || post_.propagate(*this, p)); }
 
 	//! Executes a one-step lookahead on p.
 	/*!
@@ -415,8 +456,7 @@ public:
 	 *       c->undoLevel() is called *before* p is undone. Hence, the
 	 *       range [s.levelStart(s.decisionLevel()), s.assignment().size())
 	 *       contains p followed by all literals that were forced because of p.
-	 * \note During propagation of p, only post propagators with priority
-	 *       higher than that of c are called.
+	 * \note propagateUntil(c) is used to propagate p.
 	 */
 	bool test(Literal p, PostPropagator* c);
 
@@ -435,16 +475,7 @@ public:
 	 */
 	uint32 inDegree(WeightLitVec& out);
 	
-	//! Adds the learnt constraint c to the solver.
-	/*!
-	 * \pre endInit() was called.
-	 */
-	void addLearnt(LearntConstraint* c, uint32 size, ConstraintType type) {
-		learnts_.push_back(c);
-		stats.addLearnt(size, type); 
-	}
-	void addLearnt(LearntConstraint* c, uint32 size) { addLearnt(c, size, c->type()); }
-	
+	struct DBInfo { uint32 size; uint32 locked; uint32 pinned; };
 	//! Removes upto remMax percent of the learnt nogoods.
 	/*!
 	 * \param remMax  Fraction of nogoods to remove ([0.0,1.0]).
@@ -454,7 +485,6 @@ public:
 	 *   Nogoods that are the reason for a literal to be in the assignment
 	 *   are said to be locked and won't be removed.
 	 */
-	struct DBInfo { uint32 size; uint32 locked; uint32 pinned; };
 	DBInfo reduceLearnts(float remMax, const ReduceStrategy& rs = ReduceStrategy());
 
 	//! Resolves the active conflict using the selected strategy.
@@ -485,22 +515,23 @@ public:
 	 * \post decisionLevel == max(min(decisionLevel(), dl), max(rootLevel(), btLevel))
 	 */
 	void undoUntil(uint32 dl);
-
+	
 	/*!
 	 * Similar to undoUntil but optionally also pops the backtrack-level
 	 * to dl if possible.
 	 * \return The current decision level.
 	 */
 	uint32 undoUntil(uint32 dl, bool popBtLevel);
-
-	//! Checks whether there is a model that is symmetric to the current model.
+	
+	//! Adds a new auxiliary variable to this solver.
 	/*!
-	 * The function checks for symmetric models, i.e. models that differ only in the 
-	 * assignment of variables outside of the solver's assignment. 
-	 * Typical example: vars eliminated by the SAT-preprocessor
-	 * \pre the current assignment is a model
-	 */ 
-	bool nextSymModel();
+	 * Auxiliary variables are local to one solver and are not considered
+	 * as part of the problem. They shall be added/used only during solving, i.e.
+	 * after problem setup is completed.
+	 */
+	Var  pushAuxVar();
+	//! Pops the num most recently added auxiliary variables from this solver.
+	void popAuxVar(uint32 num = UINT32_MAX);
 	//@}  
 	
 	/*!
@@ -510,18 +541,17 @@ public:
 	 * parameter.
 	 */
 	//@{
-	
 	//! Returns the number of problem variables.
-	/*!
-	 * \see SharedContext::numVars()
-	 */
+	uint32   numProblemVars()       const { return shared_->numVars(); }
+	//! Returns the number of active solver-local aux variables.
+	uint32   numAuxVars()           const { return numVars() - numProblemVars(); }
+	//! Returns the number of solver variables, i.e. numProblemVars() + numAuxVars()
 	uint32   numVars()              const { return assign_.numVars() - 1; }
 	//! Returns true if var represents a valid variable in this solver.
-	/*!
-	 * \see SharedContext::validVar(Var v)
-	 */
 	bool     validVar(Var var)      const { return var <= numVars(); }
-	//! Returns the number of assigned problem variables.
+	//! Returns true if var is a solver-local aux var.
+	bool     auxVar(Var var)        const { return shared_->numVars() < var; }
+	//! Returns the number of assigned variables.
 	uint32   numAssignedVars()      const { return assign_.assigned(); }
 	//! Returns the number of free variables.
 	/*!
@@ -531,10 +561,10 @@ public:
 	uint32   numFreeVars()          const { return assign_.free()-1; }
 	//! Returns the value of v w.r.t the current assignment.
 	ValueRep value(Var v)           const { assert(validVar(v)); return assign_.value(v); }
-	//! Returns the previous value of v.
-	ValueRep savedValue(Var v)      const { assert(validVar(v)); return assign_.saved(v); }
-	//! Returns the preferred value of v.
-	ValueRep prefValue(Var v)       const { assert(validVar(v)); return assign_.pref(v); }
+	//! Returns the value of v w.r.t the top level.
+	ValueRep topValue(Var v)        const { return level(v) == 0 ? value(v) : value_free; }
+	//! Returns the set of preferred values of v.
+	ValueSet pref(Var v)            const { assert(validVar(v)); return assign_.pref(v);}
 	//! Returns true if p is true w.r.t the current assignment.
 	bool     isTrue(Literal p)      const { assert(validVar(p.var())); return assign_.value(p.var()) == trueValue(p); }
 	//! Returns true if p is false w.r.t the current assignment.
@@ -571,6 +601,7 @@ public:
 	Literal  decision(uint32 dl)    const { assert(validLevel(dl)); return assign_.trail[ levels_[dl-1].trailPos ]; }
 	//! Returns true, if the current assignment is conflicting.
 	bool     hasConflict()          const { return !conflict_.empty(); }
+	bool     hasStopConflict()      const { return hasConflict() && conflict_[0] == negLit(0); }
 	//! Returns the number of (unprocessed) literals in the propagation queue.
 	uint32   queueSize()            const { return (uint32) assign_.qSize(); }
 	//! Number of problem constraints in this solver.
@@ -595,8 +626,11 @@ public:
 	const LitVec&     conflict()                     const { return conflict_; }
 	//! Returns the most recently derived conflict clause.
 	const LitVec&     conflictClause()               const { return cc_; }
+	//! Returns the set of eliminated literals that are unconstraint w.r.t the last model.
+	const LitVec&     symmetric()                    const { return temp_; }
 	//! Returns the enumeration constraint set by the enumerator used.
-	Constraint*       getEnumerationConstraint()     const { return enum_; }
+	Constraint*       enumerationConstraint()        const { return enum_; }
+	DBRef             constraints()                  const { return constraints_; }
 	//! Returns the idx'th learnt constraint.
 	/*!
 	 * \pre idx < numLearntConstraints()
@@ -606,7 +640,10 @@ public:
 		return *static_cast<LearntConstraint*>(learnts_[ idx ]);
 	}
 	
-	SolveStats  stats;  /**< stores some statistics about the solving process */
+	SolverStrategies  strategy; /**< Strategies used by this object.              */
+	RNG               rng;      /**< Random number generator for this object.     */
+	ValueVec          model;    /**< Stores the last model (if any).              */
+	SolverStats       stats;    /**< Stores statistics about the solving process. */
 	//@}
 
 	/*!
@@ -670,31 +707,24 @@ public:
 	 * know what you are doing!
 	 */
 	//@{
-	//! Sets the initial saved value of v to val if not yet assigned.
-	void initSavedValue(Var v, ValueRep val) {
-		assert(validVar(v));
-		if (assign_.saved(v) == value_free) { assign_.setSavedValue(v, val); }
-	}
-	//! Sets the preferred value of v to val. 
-	/*!
-	 * The preferred value overwrites any sign heuristic.
-	 */
-	void initPrefValue(Var v, ValueRep val) {
-		assert(validVar(v));
-		if (assign_.pref(v) == value_free) { assign_.setPrefValue(v, val); }
-	}
-	//! Returns the default decision literal of v.
-	Literal  defaultLiteral(Var v);
+	bool addPost(PostPropagator* p, bool init) { post_.add(p, p->priority()); return !init || p->init(*this); }
 	//! Updates the reason for p being tue.
 	/*!
 	 * \pre p is true and x is a valid reason for p
 	 */
 	bool setReason(Literal p, const Antecedent& x, uint32 data = UINT32_MAX) {
-		assert(isTrue(p));
+		assert(isTrue(p) || shared_->eliminated(p.var()));
 		assign_.setReason(p.var(), x);
 		if (data != UINT32_MAX) { assign_.setData(p.var(), data); }
 		return true;
 	}
+	//! Request additional reason data slot for variable v.
+	void requestData(Var v) { assign_.requestData(v + 1); }
+	void setPref(Var v, ValueSet::Value which, ValueRep to) {
+		assert(validVar(v) && to <= value_false);
+		assign_.requestPrefs();
+		assign_.setPref(v, which, to);
+	}
 	//! Returns the reason for p being true as a set of literals.
 	void reason(Literal p, LitVec& out) { assert(isTrue(p)); out.clear(); return assign_.reason(p.var()).reason(*this, p, out); }
 	//! Computes a new lbd for the antecedent of p given as the range [first, last).
@@ -709,14 +739,14 @@ public:
 	 *       conflict clause that is currently being derived.
 	 */
 	uint32 updateLearnt(Literal p, const Literal* first, const Literal* last, uint32 cLbd, bool forceUp = false) {
-		uint32 up = strategies().updateLbd;
+		uint32 up = strategy.updateLbd;
 		if ((up || forceUp) && cLbd > 1) {
 			uint32 strict = (up == 2u);
 			uint32 p1     = (up == 3u);
 			uint32 nLbd   = (strict|p1) + countLevels(first, last, cLbd - strict);
 			if (nLbd < cLbd) { cLbd = nLbd - p1; }
 		}
-		if (strategies().bumpVarAct && isTrue(p)) { bumpAct_.push_back(WeightLiteral(p, cLbd)); }
+		if (strategy.bumpVarAct && isTrue(p)) { bumpAct_.push_back(WeightLiteral(p, cLbd)); }
 		return cLbd;
 	}
 	//! Visitor function for antecedents used during conflict clause minimization.
@@ -724,28 +754,22 @@ public:
 		return seen(p.var()) || 
 			(rec && hasLevel(level(p.var())) && rec->checkRecursive(p));
 	}
-	
-	//! Sets a soft memory limit on learnt nogoods.
-	/*!
-	 * The soft limit is checked in search() and when hit
-	 * search() will return value_free.
-	 */
-	void  setMemLimit(uint32 learntLimitInMB);
+
 	//! Returns true if number of learnts exceeds x.learnts or the soft memory limit is exceeded.
-	bool  learntLimit(const SearchLimits& x) const { return numLearntConstraints() > x.learnts || memLimit_ < 0; }
+	bool  learntLimit(const SearchLimits& x) const { return numLearntConstraints() > x.learnts || memUse_ > x.memLimit; }
 	
 	//! Allocates a small block (32-bytes) from the solver's small block pool.
 	void* allocSmall()       { return smallAlloc_->allocate(); }
 	//! Frees a small block previously allocated from the solver's small block pool.
 	void  freeSmall(void* m) { smallAlloc_->free(m);    }
 	
-	void  addLearntBytes(uint32 bytes)  { memLimit_ -= bytes; }
-	void  freeLearntBytes(uint32 bytes) { memLimit_ += bytes; }
+	void  addLearntBytes(uint32 bytes)  { memUse_ += bytes; }
+	void  freeLearntBytes(uint64 bytes) { memUse_ -= (bytes < memUse_) ? bytes : memUse_; }
 	
 	//! simplifies cc and returns finalizeConflictClause(cc, info);
-	uint32  simplifyConflictClause(LitVec& cc, ClauseInfo& info, ClauseHead* rhs);
-	uint32  finalizeConflictClause(LitVec& cc, ClauseInfo& info);
-	uint32  countLevels(const Literal* first, const Literal* last, uint32 maxLevels);
+	uint32 simplifyConflictClause(LitVec& cc, ClauseInfo& info, ClauseHead* rhs);
+	uint32 finalizeConflictClause(LitVec& cc, ClauseInfo& info, uint32 ccRepMode = 0);
+	uint32 countLevels(const Literal* first, const Literal* last, uint32 maxLevels);
 	bool hasLevel(uint32 dl)    const { assert(validLevel(dl)); return levels_[dl-1].marked != 0; }
 	bool frozenLevel(uint32 dl) const { assert(validLevel(dl)); return levels_[dl-1].freeze != 0; }
 	void markLevel(uint32 dl)    { assert(validLevel(dl)); levels_[dl-1].marked = 1; }
@@ -768,25 +792,28 @@ private:
 		uint32        freeze   :  1;
 		ConstraintDB* undo;
 	};
-	typedef PodVector<DLevel>::type     DecisionLevels;  
+	struct DecisionLevels : public PodVector<DLevel>::type {
+		DecisionLevels() : root(0), backtrack(0) {}
+		uint32 root;      // root level
+		uint32 backtrack; // backtrack level
+	};
 	typedef PodVector<Antecedent>::type ReasonVec;
 	typedef PodVector<WatchList>::type  Watches;
 	struct PPList {
 		PPList();
 		~PPList();
-		void add(PostPropagator* p);
+		void add(PostPropagator* p, uint32 prio);
 		void remove(PostPropagator* p);
 		bool propagate(Solver& s, PostPropagator* p) const;
 		void simplify(Solver& s, bool shuf);
-		void reset() const;
+		void cancel()           const;
 		bool isModel(Solver& s) const;
-		bool init(Solver& s, PostPropagator* list) const;
-		bool init(Solver& s) const { return init(s, head); }
-		PostPropagator* enableFunky();
-		PostPropagator* disableFunky();
-		PostPropagator* nil() const { return (PostPropagator*)1; }
-		PostPropagator* head;
-		PostPropagator* look;
+		void disable();
+		void enable();
+		PostPropagator* head()  const { return list; }
+		PostPropagator* active()const { return *act; }
+		PostPropagator* list;// head of pp-list
+		PostPropagator**act; // head of active and initialized pps
 	};
 	struct CmpScore {
 		typedef std::pair<uint32, Activity> ViewPair;
@@ -805,13 +832,11 @@ private:
 		uint32                glue;
 	private: CmpScore& operator=(const CmpScore&);
 	};
-	inline  bool hasStopConflict() const;
 	bool    validWatch(Literal p) const { return p.index() < (uint32)watches_.size(); }
 	void    freeMem();
 	bool    simplifySAT();
-	void    simplifyShort(Literal p);
 	bool    unitPropagate();
-	bool    propagateUnits();
+	void    cancelPropagation() { assign_.qReset(); post_.cancel(); }
 	void    undoLevel(bool sp);
 	uint32  analyzeConflict();
 	void    otfs(Antecedent& lhs, const Antecedent& rhs, Literal p, bool final);
@@ -827,45 +852,48 @@ private:
 	DBInfo  reduceSort(uint32 maxR, const CmpScore& cmp);
 	DBInfo  reduceSortInPlace(uint32 maxR, const CmpScore& cmp, bool onlyPartialSort);
 	ConstraintDB* allocUndo(Constraint* c);
-	SolverStrategies  strategy_;    // Strategies used by this solver-object
+	SharedContext*    shared_;      // initialized by master thread - otherwise read-only!
 	Heuristic         heuristic_;   // active decision heuristic
-	Assignment        assign_;      // three-valued assignment
-	ConstraintDB      constraints_; // problem constraints
-	ConstraintDB      learnts_;     // learnt constraints
-	ImpliedList       impliedLits_; // Lits that were asserted on current dl but are logically implied earlier
-	LitVec            conflict_;    // stores conflict-literals for later analysis
-	LitVec            cc_;          // temporary: stores conflict clause within analyzeConflict
-	LitVec            temp_;        // temporary: eliminated vars that are unconstraint w.r.t the current model
-	WeightLitVec      bumpAct_;     // temporary: lits from current dl whose activity might get an extra bump
-	DecisionLevels    levels_;      // Stores information (e.g. position in trail) on each decision level
-	VarVec            lbdStamp_;    // temporary vector for computing LBD
-	VarVec            cflStamp_;    // temporary vector for computing number of conflicts in branch
-	Watches           watches_;     // for each literal p: list of constraints watching p
-	PPList            post_;        // (possibly empty) list of post propagators
-	ClauseInfo        ccInfo_;      // temporary: stores information about conflict clause cc_
 	CCMinRecursive*   ccMin_;       // additional data for supporting recursive strengthen
 	SmallClauseAlloc* smallAlloc_;  // allocator object for small clauses
-	SharedContext*    shared_;      // initialized by master thread - otherwise read-only!
-	ConstraintDB*     undoHead_;    // Free list of undo DBs
+	ConstraintDB*     undoHead_;    // free list of undo DBs
 	Constraint*       enum_;        // enumeration constraint - set by enumerator
-	int64             memLimit_;    // Soft memory limit on learnt constraints
-	uint32            id_;          // solver id - only needed with multi-threading
-	uint32            units_;       // number of top-level assignments: always marked as seen
-	LitVec::size_type lastSimplify_;// number of top-level assignments on last call to simplify
-	uint32            rootLevel_;   // dl on which search started.
-	uint32            btLevel_;     // When enumerating models: DL of the last unflipped decision from the current model. Can't backjump below this level.
+	uint64            memUse_;      // memory used by learnt constraints (estimate)
+	Assignment        assign_;      // three-valued assignment.
+	DecisionLevels    levels_;      // information (e.g. position in trail) on each decision level
+	ConstraintDB      constraints_; // problem constraints
+	ConstraintDB      learnts_;     // learnt constraints
+	PPList            post_;        // (possibly empty) list of post propagators
+	Watches           watches_;     // for each literal p: list of constraints watching p
+	LitVec            conflict_;    // conflict-literals for later analysis
+	LitVec            cc_;          // temporary: conflict clause within analyzeConflict
+	LitVec            temp_;        // temporary: redundant literals in simplifyConflictClause
+	WeightLitVec      bumpAct_;     // temporary: lits from current dl whose activity might get an extra bump
+	VarVec            lbdStamp_;    // temporary vector for computing LBD
+	VarVec            cflStamp_;    // temporary vector for computing number of conflicts in branch
+	ImpliedList       impliedLits_; // lits that were asserted on current dl but are logically implied earlier
+	ClauseInfo        ccInfo_;      // temporary: information about conflict clause cc_
+	Literal           tag_;         // aux literal for tagging learnt constraints
 	uint32            lbdTime_;     // temporary counter for computing lbd
-	bool              shuffle_;     // shuffle program on next simplify?
+	uint32            dbIdx_;       // position of first new problem constraint in master db 
+	uint32            lastSimp_ :30;// number of top-level assignments on last call to simplify
+	uint32            shufSimp_ : 1;// shuffle db on next simplify?
+	uint32            initPost_ : 1;// initialize new post propagators?
 };
 
 inline bool isRevLit(const Solver& s, Literal p, uint32 maxL) {
 	return s.isFalse(p) && (s.seen(p) || s.level(p.var()) < maxL);
 }
-inline Literal Solver::defaultLiteral(Var v) {
-	uint32 xs = strategies().signDef;
-	if (xs == SolverStrategies::sign_type) { return sharedContext()->preferredLiteralByType(v); }
-	if (xs == SolverStrategies::sign_rnd)  { return Literal(v, strategies().rng.drand() < 0.5); }
-	return Literal(v, xs == SolverStrategies::sign_yes);
+
+template <class C>
+void simplifyDB(Solver& s, C& db, bool shuffle) {
+	typename C::size_type j = 0;
+	for (typename C::size_type i = j, end = db.size(); i != end; ++i) {
+		Constraint* c = db[i];
+		if (c->simplify(s, shuffle)){ c->destroy(&s, false); }
+		else                        { db[j++] = c;  }
+	}
+	shrinkVecTo(db, j);
 }
 
 //@}
@@ -901,12 +929,18 @@ public:
 	virtual void endInit(Solver& /* s */) { }  
 	
 	/*!
-	 * Called for each var that changes its state from eliminated back to normal.
-	 * The default-implementation is a noop.
-	 * \param s Solver in which v is resurrected.
-	 * \param v The variable that is resurrected.
+	 * Called if the state of one or more variables changed. 
+	 * A state change is one of: 
+	 *   - A previously eliminated variable is resurrected.
+	 *   - A new aux variable was added.
+	 *   - An aux variable was removed.
+	 *   .
+	 * \param s Solver in which the state change occurred.
+	 * \param v The first variable affected by the change.
+	 * \param n The range of variables affected, i.e. [v, v+n).
+	 * \note Use s.validVar(v) and s.auxVar(v) to determine the reason for the update.
 	 */
-	virtual void resurrect(const Solver& /* s */, Var /* v */) {}
+	virtual void updateVar(const Solver& /* s */, Var /* v */, uint32 /* n */) = 0;
 	
 	/*!
 	 * Called on decision level 0. Variables that are assigned on this level
@@ -990,12 +1024,25 @@ public:
 	virtual Literal selectRange(Solver& /* s */, const Literal* first, const Literal* /* last */) {
 		return *first;
 	}
-	static Literal selectLiteral(Solver& s, Var v, int sc) {
-		ValueRep xv;
-		if ((xv = s.prefValue(v)) != value_free) { return Literal(v, valSign(xv)); }
-		if ((xv = s.savedValue(v))!= value_free) { return Literal(v, valSign(xv)); }
-		if (sc != 0)                             { return Literal(v, sc < 0); }
-		return s.defaultLiteral(v);
+	static Literal selectLiteral(Solver& s, Var v, int signScore) {
+		ValueSet prefs = s.pref(v);
+		if (signScore != 0 && !prefs.has(ValueSet::user_value | ValueSet::saved_value | ValueSet::pref_value)) {
+			return Literal(v, signScore < 0); 
+		}
+		else if (!prefs.empty()) {
+			return Literal(v, prefs.sign());
+		}
+		return defaultLiteral(s, v);
+	}
+	static Literal defaultLiteral(Solver& s, Var v) {
+		switch(s.strategy.signDef) {
+			default: // 
+			case SolverStrategies::sign_atom: return Literal(v, !s.varInfo(v).has(VarInfo::BODY));
+			case SolverStrategies::sign_no  : return posLit(v);
+			case SolverStrategies::sign_yes : return negLit(v);
+			case SolverStrategies::sign_rnd : return Literal(v, s.rng.drand() < 0.5);
+			case SolverStrategies::sign_disj: return Literal(v, !s.varInfo(v).has(VarInfo::BODY|VarInfo::DISJ));
+		}
 	}
 private:
 	DecisionHeuristic(const DecisionHeuristic&);
@@ -1004,6 +1051,8 @@ private:
 
 //! Selects the first free literal w.r.t to the initial variable order.
 class SelectFirst : public DecisionHeuristic {
+public:
+	void updateVar(const Solver& /* s */, Var /* v */, uint32 /* n */) {}
 private:
 	Literal doSelect(Solver& s);
 };
