@@ -17,14 +17,26 @@
 // along with Clasp; if not, write to the Free Software
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
+#include "example.h"
+#include <clasp/enumerator.h>
+#include <clasp/Solver.h>
+void printModel(const Clasp::SymbolTable& symTab, const Clasp::Model& model) {
+	std::cout << "Model " << model.num << ": \n";
+	// Print each named atom that is true w.r.t the current model.
+	for (Clasp::SymbolTable::const_iterator it = symTab.begin(); it != symTab.end(); ++it) {
+		if (model.isTrue(it->second.lit) && !it->second.name.empty()) {
+			std::cout << it->second.name.c_str() << " ";
+		}
+	}
+	std::cout << std::endl;
+}
 
-#include <iostream>
-void example1();
-void example2();
+#define RUN(x) try { std::cout << "*** Running " << static_cast<const char*>(#x) << " ***" << std::endl; x(); } catch (const std::exception& e) { std::cout << " *** ERROR: " << e.what() << std::endl; }
 
 int main() {
-	std::cout << "*** Running example1 ***" << std::endl;
-	example1();
-	std::cout << "*** Running example2 ***" << std::endl;
-	example2();
-}
\ No newline at end of file
+	RUN(example1);
+	RUN(example2);
+#if WITH_THREADS
+	RUN(example3);
+#endif
+}
