@@ -23,7 +23,6 @@
 #include <algorithm>
 
 namespace Clasp { namespace Detail {
-
 struct GreaterLevel {
 	GreaterLevel(const Solver& s) : solver_(s) {}
 	bool operator()(const Literal& p1, const Literal& p2) const {
@@ -41,20 +40,8 @@ struct Sink {
 	SharedLiterals* clause;   
 };
 
-// returns an abstraction of p's decision level, s.th
-// abstraction(any true literal) > abstraction(any free literal) > abstraction(any false literal).
-static uint32 levelAbstraction(const Solver& s, Literal p) {
-	ValueRep value_p = s.value(p.var());
-	// DL+1,  if isFree(p)
-	// DL(p), if isFalse(p)
-	// ~DL(p),if isTrue(p)
-	uint32   abstr_p = value_p == value_free ? s.decisionLevel()+1 : s.level(p.var()) ^ -(value_p==trueValue(p));
-	assert(abstr_p > 0 || (s.isFalse(p) && s.level(p.var()) == 0));
-	return abstr_p;
-}
-
 void* alloc(uint32 size) {
-	#pragma message TODO("replace with CACHE_LINE_ALIGNED alloc")
+	CLASP_PRAGMA_TODO("replace with CACHE_LINE_ALIGNED alloc")
 	return ::operator new(size);
 }
 void free(void* mem) {
@@ -109,364 +96,256 @@ SharedLiterals* SharedLiterals::share() {
 	++refCount_;
 	return this;
 }
-/////////////////////////////////////////////////////////////////////////////////////////
-// ClauseCreator::Watches
-/////////////////////////////////////////////////////////////////////////////////////////
-struct ClauseCreator::Watches {
-	Watches(const Solver& s, const Literal* begin, const Literal* end, bool knownOrder) {
-		uint32 size   = uint32(end - begin);
-		simpSize      = size;
-		std::fill_n(lits, (int)Clause::MAX_SHORT_LEN, negLit(0));
-		if (!knownOrder) { init(s, (Literal*)begin, size, lits); }
-		else             { 
-			uint32 i= size;
-			while (i-- && Detail::levelAbstraction(s, begin[i]) == 0) {
-				--simpSize;
-			}
-			std::memcpy(lits, begin, std::min(simpSize, uint32(Clause::MAX_SHORT_LEN))*sizeof(Literal));
-			abstr_f = Detail::levelAbstraction(s, lits[0]);
-			abstr_s = Detail::levelAbstraction(s, lits[1]);
-		}
-		assert(simpSize <= size);
-	}
-	Watches(const Solver& s, LitVec& in, bool knownOrder) {
-		assert(in.size() > 1);
-		simpSize = in.size();
-		if (!knownOrder) { init(s, &in[0], in.size(), 0); }
-		else             { abstr_f = Detail::levelAbstraction(s, in[0]); abstr_s = Detail::levelAbstraction(s, in[1]);  }
-		lits[0] = in[0];
-		lits[1] = in[1];
-	}
-	void init(const Solver& s, Literal* in, uint32 size, Literal* out) {
-		abstr_f     = abstr_s = simpSize = 0;
-		uint32 fp   = 0,  sp  = 0;
-		for (uint32 i = 0, j  = 0, abstr_p; i != size; ++i) {
-			Literal p = in[i];
-			abstr_p   = Detail::levelAbstraction(s, p);
-			if (abstr_p) {
-				++simpSize;
-				if (out) {
-					out[j++]= abstr_p > abstr_s ? in[sp] : in[i];
-					if (j  == Clause::MAX_SHORT_LEN) { j = 2; }
-				}
-				if (abstr_p > abstr_f) {
-					assert(abstr_f >= abstr_s);
-					sp      = fp;
-					abstr_s = abstr_f;
-					fp      = i;
-					abstr_f = abstr_p;
-				}
-				else if (abstr_p > abstr_s) {
-					sp      = i;
-					abstr_s = abstr_p;
-				}
-			}
-		}
-		if (out) {
-			if (simpSize > 0) out[0] = in[fp];
-			if (simpSize > 1) out[1] = in[sp];
-		}
-		else if (size > 1) {
-			std::swap(in[0], in[fp]);
-			std::swap(in[1], sp != 0 ? in[sp] : in[fp]);
-		}
-	}
-	uint32  abstr_f;
-	uint32  abstr_s;
-	uint32  simpSize;
-	Literal lits[Clause::MAX_SHORT_LEN];
-};
+
 /////////////////////////////////////////////////////////////////////////////////////////
 // ClauseCreator
 /////////////////////////////////////////////////////////////////////////////////////////
 ClauseCreator::ClauseCreator(Solver* s) 
-: solver_(s) { }
-
-void ClauseCreator::init(ConstraintType t) {
-	assert(!solver_ || solver_->decisionLevel() == 0 || t != Constraint_t::static_constraint);
-	literals_.clear();
-	extra_     = ClauseInfo(t);
-	fwLev_     = 0;
-	swLev_     = 0;
+	: solver_(s)
+	, flags_(0){ 
 }
 
 ClauseCreator& ClauseCreator::start(ConstraintType t) {
-	init(t);
+	assert(solver_ && (solver_->decisionLevel() == 0 || t != Constraint_t::static_constraint));
+	literals_.clear();
+	extra_ = ClauseInfo(t);
 	return *this;
 }
 
-ClauseCreator& ClauseCreator::startAsserting(ConstraintType t, const Literal& p) {
-	init(t);
-	literals_.push_back(p);
-	fwLev_     = UINT32_MAX;
-	return *this;
+uint32 ClauseCreator::watchOrder(const Solver& s, Literal p) {
+	ValueRep value_p = s.value(p.var());
+	// DL+1,  if isFree(p)
+	// DL(p), if isFalse(p)
+	// ~DL(p),if isTrue(p)
+	uint32   abstr_p = value_p == value_free ? s.decisionLevel()+1 : s.level(p.var()) ^ -(value_p==trueValue(p));
+	assert(abstr_p > 0 || (s.isFalse(p) && s.level(p.var()) == 0));
+	return abstr_p;
 }
 
-ClauseCreator& ClauseCreator::add(const Literal& p) {
-	assert(solver_); const Solver& s = *solver_;
-	uint32 abstr_p = Detail::levelAbstraction(s, p);
-	if (abstr_p != 0) {
-		literals_.push_back(p);
-		// watch the two literals with highest abstraction, i.e.
-		// prefer true over free over false literals
-		if (abstr_p > fwLev_) {
-			std::swap(abstr_p, fwLev_);
-			std::swap(literals_[0], literals_.back());
+ClauseRep ClauseCreator::prepare(Solver& s, const Literal* in, uint32 inSize, const ClauseInfo& e, uint32 flags, Literal* out, uint32 outMax) {
+	assert(out && outMax > 2);
+	ClauseRep ret  = ClauseRep::prepared(out, 0, e);
+	uint32 abst_w1 = 0, abst_w2 = 0;
+	bool simplify  = ((flags & clause_force_simplify) != 0) && inSize > 2 && outMax >= inSize;
+	Literal tag    = ~s.tagLiteral();
+	Var     vMax   = 0;
+	for (uint32 i = 0, j = 0, MAX_OUT = outMax - 1; i != inSize; ++i) {
+		Literal p     = in[i];
+		uint32 abst_p = watchOrder(s, p);
+		if ((abst_p + 1) > 1 && (!simplify || !s.seen(p.var()))) {
+			out[j] = p;
+			if (p == tag)         { ret.info.setTagged(true); }
+			if (p.var() > vMax)   { vMax = p.var();}
+			if (simplify)         { s.markSeen(p); }
+			if (abst_p > abst_w1) { std::swap(abst_p, abst_w1); std::swap(out[0], out[j]); }
+			if (abst_p > abst_w2) { std::swap(abst_p, abst_w2); std::swap(out[1], out[j]); }
+			if (j != MAX_OUT)     { ++j;  }
+			++ret.size;
 		}
-		if (abstr_p > swLev_) {
-			std::swap(literals_[1], literals_.back());
-			swLev_= abstr_p;
+		else if (abst_p == UINT32_MAX || (simplify && abst_p && s.seen(~p))) {
+			abst_w1 = UINT32_MAX;
+			break;
 		}
 	}
-	// else: drop uncoditionally false literals 
-	return *this;
+	if (simplify) { 
+		for (uint32 x = 0, end = ret.size; x != end; ++x) { s.clearSeen(out[x].var()); } 
+	}
+	if (abst_w1 == UINT32_MAX || (abst_w2 && out[0].var() == out[1].var())) {
+		out[0]   = abst_w1 == UINT32_MAX || out[0] == ~out[1] ? posLit(0) : out[0]; 
+		ret.size = 1;
+	}
+	ret.info.setAux(s.auxVar(vMax));
+	return ret;
 }
 
-void ClauseCreator::simplify() {
-	if (literals_.empty()) return;
-	swLev_    = 0;
-	uint32 sw = 0;
-	solver_->markSeen(literals_[0]);
-	LitVec::size_type i, j = 1;
-	for (i = 1; i != literals_.size(); ++i) {
-		Literal x = literals_[i];
-		if (solver_->seen(~x)) { 
-			fwLev_     = UINT32_MAX; 
-			continue;
-		}
-		if (!solver_->seen(x)) {
-			solver_->markSeen(x);
-			if (i != j) { literals_[j] = literals_[i]; }
-			if (swLev_ != uint32(-1)) {
-				uint32 xLev = solver_->value(x.var()) == value_free ? uint32(-1) : solver_->level(x.var());
-				if (swLev_ < xLev) {
-					swLev_ = xLev;
-					sw     = (uint32)j;
-				}
-			}
-			++j;
-		}
+
+ClauseRep ClauseCreator::prepare(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& info) {
+	if (lits.empty()) { lits.push_back(negLit(0)); }
+	if ((flags & clause_no_prepare) == 0 || (flags & clause_force_simplify) != 0) {
+		ClauseRep x = prepare(s, &lits[0], (uint32)lits.size(), info, flags, &lits[0]);
+		shrinkVecTo(lits, x.size);
+		return x;
 	}
-	literals_.erase(literals_.begin()+j, literals_.end());
-	for (LitVec::iterator it = literals_.begin(), end = literals_.end(); it != end; ++it) {
-		solver_->clearSeen(it->var());
-	}
-	if (literals_.size() >= 2 && sw != 1) {
-		std::swap(literals_[1], literals_[sw]);
-	}
-	if (fwLev_ == UINT32_MAX) { literals_[0] = posLit(0); }
+	return ClauseRep::prepared(&lits[0], (uint32)lits.size(), info);
 }
 
-ClauseCreator::Status ClauseCreator::status(uint32 dl, uint32 fw, uint32 sw) {
-	if (fw > varMax)  {
-		if (fw == UINT32_MAX)      { return status_subsumed; }
-		if ((fw^UINT32_MAX) >= sw) { return status_asserting;}
-		return status_sat;
-	}
-	if (sw <= dl) {
-		if (fw > dl) { return status_unit; }
-		if (fw > sw) { return status_asserting; }
-		return status_conflicting;
-	}
-	return status_open;
-}
-ClauseCreator::Status ClauseCreator::status() const {
-	if (literals_.empty()) { return status_empty; }
-	return status(solver_->decisionLevel(), Detail::levelAbstraction(*solver_, literals_[0]), swLev_);
+ClauseRep ClauseCreator::prepare(bool forceSimplify) {
+	return prepare(*solver_, literals_, forceSimplify ? clause_force_simplify : 0, extra_);
 }
 
 ClauseCreator::Status ClauseCreator::status(const Solver& s, const Literal* clause_begin, const Literal* clause_end) {
 	if (clause_end <= clause_begin) { return status_empty; }
-	Watches w(s, clause_begin, clause_end, false);
-	return status(s.decisionLevel(), w.abstr_f, w.abstr_s);
+	Literal temp[3];
+	ClauseRep x = prepare(const_cast<Solver&>(s), clause_begin, uint32(clause_end - clause_begin), ClauseInfo(), 0, temp, 3);
+	return status(s, x);
 }
 
-ClauseCreator::Status ClauseCreator::status(const Solver& s, const Watches& w, uint32 flags) {
-	Status x    = status(s.decisionLevel(), w.abstr_f, w.abstr_s);
-	bool notSat = (flags & clause_not_sat) != 0;
-	bool notRoot= (flags & clause_not_root_sat) != 0;
-	if (x == status_conflicting || (x == status_asserting && s.isFalse(w.lits[0]))) {
-		x = (flags & clause_not_conflict) == 0 ? status_unit : status_conflicting;
-	}
-	else if (x == status_sat && (notSat || (notRoot && s.level(w.lits[0].var()) <= s.rootLevel()))) {
-		x = status_subsumed;
-	}
-	return x;
+ClauseCreator::Status ClauseCreator::status(const Solver& s, const ClauseRep& c) {
+	uint32 dl = s.decisionLevel();
+	uint32 fw = c.size     ? watchOrder(s, c.lits[0]) : 0;
+	if (fw == UINT32_MAX) { return status_subsumed; }	
+	uint32 sw = c.size > 1 ? watchOrder(s, c.lits[1]) : 0;
+	uint32 st = status_open;
+	if      (fw > varMax)   { st|= status_sat; fw = ~fw; }
+	else if (fw <= dl)      { st|= (fw ? status_unsat : status_empty); }
+	if (sw <= dl && fw > sw){ st|= status_unit;  }
+	return static_cast<Status>(st);
 }
 
-ClauseCreator::Result ClauseCreator::end() {
+bool ClauseCreator::ignoreClause(const Solver& s, const ClauseRep& c, Status st, uint32 flags) {
+	uint32 x = (st & (status_sat|status_unsat));
+	if (x == status_open)  { return false; }
+	if (x == status_unsat) { return st != status_empty && (flags & clause_not_conflict) != 0; }
+	return st == status_subsumed || (st == status_sat && ( (flags & clause_not_sat) != 0 || ((flags & clause_not_root_sat) != 0 && s.level(c.lits[0].var()) <= s.rootLevel())));
+}
+
+ClauseCreator::Result ClauseCreator::end(uint32 flags) {
 	assert(solver_);
-	Result ret(0);
-	Solver& s     = *solver_;
-	uint32  flags = clause_known_order | clause_not_sat | clause_not_conflict;
-	if (extra_.learnt() && !isSentinel(s.sharedContext()->tagLiteral())) {
-		extra_.setTagged(std::find(literals_.begin(), literals_.end(), ~s.sharedContext()->tagLiteral()) != literals_.end());
-	}
-	return ClauseCreator::create(s, literals_, flags, extra_);
+	flags |= flags_;
+	return ClauseCreator::create_prepared(*solver_, prepare(*solver_, literals_, flags, extra_), flags);
 }
 
-ClauseCreator::Status ClauseCreator::initWatches(Solver& s, LitVec& lits, bool allFree) {
-	uint32 size = static_cast<uint32>(lits.size());
-	if (size < 2) return size == 1 ? status_unit : status_empty;
-	if (allFree) {
-		uint32 fw = 0, sw = 1;
-		uint32 w  = s.strategies().initWatches;
-		if (w == SolverStrategies::watch_rand) {
-			RNG& r = s.strategies().rng;
-			fw = r.irand(size);
-			while ( (sw = r.irand(size)) == fw) {/*intentionally empty*/}
-		}
-		else if (w == SolverStrategies::watch_least) {
-			uint32 watch[2] = {0, size-1};
-			uint32 count[2] = {s.numWatches(~lits[0]), s.numWatches(~lits[size-1])};
-			uint32 maxCount = count[0] < count[1];
-			for (uint32 x = 1, end = size-1; count[maxCount] > 0u && x != end; ++x) {
-				uint32 cxw = s.numWatches(~lits[x]);
-				if (cxw < count[maxCount]) {
-					if (cxw < count[1-maxCount]) {
-						watch[maxCount]   = watch[1-maxCount];
-						count[maxCount]   = count[1-maxCount];
-						watch[1-maxCount] = x;
-						count[1-maxCount] = cxw;
-					}
-					else {
-						watch[maxCount]   = x;
-						count[maxCount]   = cxw;
-					}
-				}
-			}
-			fw = watch[0];
-			sw = watch[1];
-		}
-		std::swap(lits[0], lits[fw]);
-		std::swap(lits[1], lits[sw]);
-		return status_open;
-	}
-	else {
-		Watches w(s, lits, false);
-		return ClauseCreator::status(s.decisionLevel(), w.abstr_f, w.abstr_s);
-	}
-}
-
-ClauseHead* ClauseCreator::newProblemClause(Solver& s, LitVec& lits, const ClauseInfo& e, uint32 flags) {
+ClauseHead* ClauseCreator::newProblemClause(Solver& s, const ClauseRep& clause, uint32 flags) {
 	ClauseHead* ret;
-	initWatches(s, lits, true);
-	if (lits.size() <= Clause::MAX_SHORT_LEN || !s.sharedContext()->physicalShareProblem()) {
-		ret = Clause::newClause(s, &lits[0], (uint32)lits.size(), e);
+	if (clause.size > 2 && s.strategy.initWatches != SolverStrategies::watch_first) {
+		uint32 fw = 0, sw = 1;
+		if (s.strategy.initWatches == SolverStrategies::watch_rand) {
+			fw = s.rng.irand(clause.size);
+			do { sw = s.rng.irand(clause.size); } while (sw == fw); 
+		}
+		else if (s.strategy.initWatches == SolverStrategies::watch_least) {
+			uint32 cw1 = s.numWatches(~clause.lits[0]);
+			uint32 cw2 = s.numWatches(~clause.lits[1]);
+			if (cw1 > cw2) { std::swap(fw, sw); std::swap(cw1, cw2); }
+			for (uint32 i = 2; i != clause.size && cw2; ++i) {
+				uint32 p   = i;
+				uint32 cwp = s.numWatches(~clause.lits[i]);
+				if (cwp < cw1) { std::swap(cwp, cw1); std::swap(fw, p); }
+				if (cwp < cw2) { std::swap(cwp, cw2); std::swap(sw, p); }
+			}
+		}
+		std::swap(clause.lits[0], clause.lits[fw]);
+		std::swap(clause.lits[1], clause.lits[sw]);
+	}
+	if (clause.size <= Clause::MAX_SHORT_LEN || !s.sharedContext()->physicalShareProblem()) {
+		ret = Clause::newClause(s, clause);
 	}
 	else {
-		ret = Clause::newShared(s, SharedLiterals::newShareable(lits, e.type(), 1), e, &lits[0], false);
+		ret = Clause::newShared(s, SharedLiterals::newShareable(clause.lits, clause.size, clause.info.type(), 1), clause.info, clause.lits, false);
 	}
 	if ( (flags & clause_no_add) == 0 ) {
+		assert(!clause.info.aux());
 		s.add(ret);
 	}
 	return ret;
 }
 
-ClauseHead* ClauseCreator::newLearntClause(Solver& s, LitVec& lits, const ClauseInfo& e, uint32 flags) {
+ClauseHead* ClauseCreator::newLearntClause(Solver& s, const ClauseRep& clause, uint32 flags) {
 	ClauseHead* ret;
 	Detail::Sink sharedPtr(0);
-	sharedPtr.clause = s.sharedContext()->distribute(s, &lits[0], static_cast<uint32>(lits.size()), e);
-	if (lits.size() <= Clause::MAX_SHORT_LEN || sharedPtr.clause == 0) {
-		if (!s.isFalse(lits[1]) || lits.size() < s.strategies().compress) {
-			ret = Clause::newLearntClause(s, lits, e);
+	sharedPtr.clause = s.distribute(clause.lits, clause.size, clause.info);
+	if (clause.size <= Clause::MAX_SHORT_LEN || sharedPtr.clause == 0) {
+		if (!s.isFalse(clause.lits[1]) || !s.strategy.compress || clause.size < s.strategy.compress) {
+			ret = Clause::newClause(s, clause);
 		}
 		else {
-			std::stable_sort(lits.begin()+2, lits.end(), Detail::GreaterLevel(s));
-			ret = Clause::newContractedClause(s, lits, e, 2, true);
+			ret = Clause::newContractedClause(s, clause, 2, true);
 		}
 	}
 	else {
-		ret              = Clause::newShared(s, sharedPtr.clause, e, &lits[0], false);
+		ret              = Clause::newShared(s, sharedPtr.clause, clause.info, clause.lits, false);
 		sharedPtr.clause = 0;
 	}
 	if ( (flags & clause_no_add) == 0 ) {
-		s.addLearnt(ret, (uint32)lits.size(), e.type());
+		s.addLearnt(ret, clause.size, clause.info.type());
 	}
 	return ret;
 }
 
-ClauseHead* ClauseCreator::newUnshared(Solver& s, SharedLiterals* clause, const Watches& w, const ClauseInfo& e) {
+ClauseHead* ClauseCreator::newUnshared(Solver& s, SharedLiterals* clause, const Literal* w, const ClauseInfo& e) {
 	LitVec temp; temp.reserve(clause->size());
-	temp.assign(w.lits, w.lits+2);
+	temp.assign(w, w+2);
 	for (const Literal* x = clause->begin(), *end = clause->end(); x != end; ++x) {
-		if (Detail::levelAbstraction(s, *x) > 0 && *x != temp[0] && *x != temp[1]) {
+		if (watchOrder(s, *x) > 0 && *x != temp[0] && *x != temp[1]) {
 			temp.push_back(*x);
 		}
 	}
-	return Clause::newClause(s, &temp[0], (uint32)temp.size(), e);
+	return Clause::newClause(s, ClauseRep::prepared(&temp[0], (uint32)temp.size(), e));
+}
+
+ClauseCreator::Result ClauseCreator::create_prepared(Solver& s, const ClauseRep& clause, uint32 flags) {
+	assert(s.decisionLevel() == 0 || (clause.info.learnt() && clause.prep));
+	Status x = status(s, clause);
+	if (ignoreClause(s, clause, x, flags)){ 
+		return Result(0, x);
+	}
+	if (clause.size > 1) {
+		Result ret(0, x);
+		if (!clause.info.learnt() && s.satPrepro() && !s.sharedContext()->frozen()) {
+			return Result(0, s.satPrepro()->addClause(clause.lits, clause.size) ? x : status_unsat);
+		}
+		if ((flags & clause_no_heuristic) == 0) { s.heuristic()->newConstraint(s, clause.lits, clause.size, clause.info.type()); }
+		if (clause.size > 3 || (flags&clause_explicit) != 0 || !s.allowImplicit(clause)) {
+			ret.local = clause.info.learnt() ? newLearntClause(s, clause, flags) : newProblemClause(s, clause, flags);
+		}
+		else {
+			// add implicit short rep
+			s.add(clause);
+		}
+		if ((x & (status_unit|status_unsat)) != 0) {
+			Antecedent ante(ret.local);
+			if (!ret.local){ ante = clause.size == 3 ? Antecedent(~clause.lits[1], ~clause.lits[2]) : Antecedent(~clause.lits[1]); }
+			ret.status = s.force(clause.lits[0], s.level(clause.lits[1].var()), ante) ? status_unit : status_unsat;
+		}
+		return ret;
+	}
+	s.add(clause);
+	return Result(0, !s.hasConflict() ? status_unit : status_unsat);
 }
 
 ClauseCreator::Result ClauseCreator::create(Solver& s, LitVec& lits, uint32 flags, const ClauseInfo& extra) {
-	assert(s.decisionLevel() == 0 || extra.learnt());
-	Result ret;
-	if (lits.size() > 1) {
-		Watches w(s, lits, (flags & clause_known_order) != 0);
-		Status x   = status(s, w, flags);
-		ret.status = x;
-		if (x == status_subsumed || x == status_conflicting) { 
-			return ret;
-		}
-		if (!extra.learnt() && s.satPrepro()) {
-			if (!s.satPrepro()->addClause(lits)) { ret.status = status_conflicting; }
-			return ret;
-		}		
-		s.heuristic()->newConstraint(s, &lits[0], lits.size(), extra.type());
-		if (lits.size() > 3 || extra.tagged() || (flags&clause_explicit) != 0) {
-			ret.local = extra.learnt() ? newLearntClause(s, lits, extra, flags) : newProblemClause(s, lits, extra, flags);
-		}
-		else {
-			s.addShort(&lits[0], lits.size(), extra);
-		}
-		if ((x & status_unit) != 0) {
-			Antecedent ante(ret.local);
-			if (!ret.local){ ante = lits.size() == 3 ? Antecedent(~lits[1], ~lits[2]) : Antecedent(~lits[1]); }
-			if (!s.force(lits[0], s.level(lits[1].var()), ante)) { ret.status = status_conflicting; }
-		}
-	}
-	else {
-		Literal imp = !lits.empty()	? lits[0] : negLit(0);
-		if (!s.addUnary(imp, extra.type())) { ret.status = status_conflicting; }
-	}	
-	return ret;
+	return create_prepared(s, prepare(s, lits, flags, extra), flags);
 }
 
+ClauseCreator::Result ClauseCreator::create(Solver& s, const ClauseRep& rep, uint32 flags) {
+	return create_prepared(s, (rep.prep == 0 && (flags & clause_no_prepare) == 0 
+		? prepare(s, rep.lits, rep.size, rep.info, flags, rep.lits)
+		: ClauseRep::prepared(rep.lits, rep.size, rep.info)), flags);
+}
 
 ClauseCreator::Result ClauseCreator::integrate(Solver& s, SharedLiterals* clause, uint32 modeFlags, ConstraintType t) {
 	assert(!s.hasConflict() && "ClauseCreator::integrate() - precondition violated!");
 	Detail::Sink shared( (modeFlags & clause_no_release) == 0 ? clause : 0);
 	// determine state of clause
-	Result  result;
-	Watches w(s, clause->begin(), clause->end(), (modeFlags & clause_known_order) != 0);
-	Status  stat( status(s, w, modeFlags) );
-	result.status = stat;
-	if (stat != status_subsumed && (stat != status_conflicting || (modeFlags & clause_not_conflict) == 0)) {
-		ClauseInfo e(t);
-		s.heuristic()->newConstraint(s, clause->begin(), clause->size(), t);
-		uint32 impSize  = (modeFlags & clause_explicit) != 0 || !s.sharedContext()->allowImplicit(t) ? 1 : 3;
-		if (w.simpSize > Clause::MAX_SHORT_LEN && s.sharedContext()->physicalShare(t)) {
-			result.local  = Clause::newShared(s, clause, e, w.lits, shared.clause == 0);
-			shared.clause = 0;
-		}
-		else if (w.simpSize > impSize) {
-			result.local  = w.simpSize <= Clause::MAX_SHORT_LEN
-			              ? Clause::newClause(s, w.lits, w.simpSize, e)
-			              : newUnshared(s, clause, w, e);
-		}
-		else {
-			// unary clause or implicitly shared via binary/ternary implication graph;
-			// only check for implication/conflict but do not create
-			// a local representation for the clause
-			s.stats.addLearnt(w.simpSize, e.type());
-			modeFlags    |= clause_no_add;
-		}
-		if ((modeFlags & clause_no_add) == 0) { s.addLearnt(result.local, w.simpSize, e.type()); }
+	Literal temp[Clause::MAX_SHORT_LEN]; temp[0] = temp[1] = negLit(0);
+	ClauseRep x    = prepare(s, clause->begin(), clause->size(), ClauseInfo(t), 0, temp, Clause::MAX_SHORT_LEN);
+	uint32 impSize = (modeFlags & clause_explicit) != 0 || !s.allowImplicit(x) ? 1 : 3;
+	Status xs      = status(s, x);
+	if (ignoreClause(s, x, xs, modeFlags)) {
+		return Result(0, xs);
 	}
-	if ((stat & status_unit) != 0) {
-		Antecedent ante = result.local ? Antecedent(result.local) : Antecedent(~w.lits[1], ~w.lits[2]);
-		uint32 impLevel = s.level(w.lits[1].var());
-		result.status   = s.force(w.lits[0], impLevel, ante) ? stat : status_conflicting;
-		if (result.local && (modeFlags & clause_int_lbd) != 0 && result.status != status_conflicting) {
+	Result result(0, xs);
+	if ((modeFlags & clause_no_heuristic) == 0) { s.heuristic()->newConstraint(s, clause->begin(), clause->size(), t); }
+	if (x.size > Clause::MAX_SHORT_LEN && s.sharedContext()->physicalShare(t)) {
+		result.local  = Clause::newShared(s, clause, x.info, temp, shared.clause == 0);
+		shared.clause = 0;
+	}
+	else if (x.size > impSize) {
+		result.local  = x.size <= Clause::MAX_SHORT_LEN ? Clause::newClause(s, x) : newUnshared(s, clause, temp, x.info);
+	}
+	else {
+		// unary clause or implicitly shared via binary/ternary implication graph;
+		// only check for implication/conflict but do not create
+		// a local representation for the clause
+		s.stats.addLearnt(x.size, x.info.type());
+		modeFlags    |= clause_no_add;
+	}
+	if ((modeFlags & clause_no_add) == 0) { s.addLearnt(result.local, x.size, x.info.type()); }
+	if ((xs & (status_unit|status_unsat)) != 0) {
+		Antecedent ante = result.local ? Antecedent(result.local) : Antecedent(~temp[1], ~temp[2]);
+		uint32 impLevel = s.level(temp[1].var());
+		result.status   = s.force(temp[0], impLevel, ante) ? status_unit : status_unsat;
+		if (result.local && (modeFlags & clause_int_lbd) != 0 && result.status != status_unsat) {
 			result.local->lbd(s.updateLearnt(negLit(0), clause->begin(), clause->end(), result.local->lbd(), true));
 		}
 	}
@@ -489,30 +368,31 @@ void* Clause::alloc(Solver& s, uint32 lits, bool learnt) {
 	return Detail::alloc(bytes);
 }
 
-ClauseHead* Clause::newClause(void* mem, Solver& s, const Literal* lits, uint32 size, const ClauseInfo& extra) {
-	assert(size >= 2 && mem);
-	return new (mem) Clause(s, extra, lits, size, size, false);
+ClauseHead* Clause::newClause(void* mem, Solver& s, const ClauseRep& rep) {
+	assert(rep.size >= 2 && mem);
+	return new (mem) Clause(s, rep);
 }
 
 ClauseHead* Clause::newShared(Solver& s, SharedLiterals* shared_lits, const ClauseInfo& e, const Literal* lits, bool addRef) {
 	return mt::SharedLitsClause::newClause(s, shared_lits, e, lits, addRef);
 }
 
-ClauseHead* Clause::newContractedClause(Solver& s, const LitVec& lits, const ClauseInfo& e, LitVec::size_type tailStart, bool extend) {
-	assert(lits.size() >= 2);
-	return new (alloc(s, (uint32)lits.size(), e.learnt())) Clause(s, e, &lits[0], static_cast<uint32>(lits.size()), static_cast<uint32>(tailStart), extend);
+ClauseHead* Clause::newContractedClause(Solver& s, const ClauseRep& rep, uint32 tailStart, bool extend) {
+	assert(rep.size >= 2);
+	if (extend) { std::stable_sort(rep.lits+tailStart, rep.lits+rep.size, Detail::GreaterLevel(s)); }
+	return new (alloc(s, rep.size, rep.info.learnt())) Clause(s, rep, tailStart, extend);
 }
 
-Clause::Clause(Solver& s, const ClauseInfo& e, const Literal* a_lits, uint32 size, uint32 tail, bool extend) 
-	: ClauseHead(e) {
-	assert(tail == size || s.isFalse(a_lits[tail]));
-	data_.local.init(size);
+Clause::Clause(Solver& s, const ClauseRep& rep, uint32 tail, bool extend) 
+	: ClauseHead(rep.info) {
+	assert(tail >= rep.size || s.isFalse(rep.lits[tail]));
+	data_.local.init(rep.size);
 	if (!isSmall()) {
 		// copy literals
-		std::memcpy(head_, a_lits, size*sizeof(Literal));
+		std::memcpy(head_, rep.lits, rep.size*sizeof(Literal));
 		tail = std::max(tail, (uint32)ClauseHead::HEAD_LITS);
-		if (tail < size) {       // contracted clause
-			head_[size-1].watch(); // mark last literal of clause
+		if (tail < rep.size) {       // contracted clause
+			head_[rep.size-1].watch(); // mark last literal of clause
 			Literal t = head_[tail];
 			if (s.level(t.var()) > 0) {
 				data_.local.markContracted();
@@ -524,10 +404,10 @@ Clause::Clause(Solver& s, const ClauseInfo& e, const Literal* a_lits, uint32 siz
 		}
 	}
 	else {
-		std::memcpy(head_, a_lits, std::min(size, (uint32)ClauseHead::HEAD_LITS)*sizeof(Literal));
-		data_.lits[0] = size > ClauseHead::HEAD_LITS   ? a_lits[ClauseHead::HEAD_LITS].asUint()   : negLit(0).asUint();
-		data_.lits[1] = size > ClauseHead::HEAD_LITS+1 ? a_lits[ClauseHead::HEAD_LITS+1].asUint() : negLit(0).asUint();
-		assert(isSmall() && Clause::size() == size);
+		std::memcpy(head_, rep.lits, std::min(rep.size, (uint32)ClauseHead::HEAD_LITS)*sizeof(Literal));
+		data_.lits[0] = rep.size > ClauseHead::HEAD_LITS   ? rep.lits[ClauseHead::HEAD_LITS].asUint()   : negLit(0).asUint();
+		data_.lits[1] = rep.size > ClauseHead::HEAD_LITS+1 ? rep.lits[ClauseHead::HEAD_LITS+1].asUint() : negLit(0).asUint();
+		assert(isSmall() && Clause::size() == rep.size);
 	}
 	attach(s);
 }
@@ -688,7 +568,7 @@ bool Clause::simplify(Solver& s, bool reinit) {
 		}
 		if (reinit && size > 3) {
 			detach(s);
-			std::random_shuffle(head_, j, s.strategies().rng);
+			std::random_shuffle(head_, j, s.rng);
 			attach(s);	
 		}
 	}
@@ -784,7 +664,7 @@ ClauseHead::BoolPair Clause::strengthen(Solver& s, Literal p, bool toShort) {
 		ret.first = *it == p;
 		eot       = *it == p ? removeFromTail(s, it, eot) : it + 1;
 	}
-	if (ret.first && ~p == s.sharedContext()->tagLiteral()) {
+	if (ret.first && ~p == s.tagLiteral()) {
 		clearTagged();
 	}
 	ret.second = toShort && eot == t.first && toImplication(s);
@@ -921,7 +801,7 @@ bool SharedLitsClause::simplify(Solver& s, bool reinit) {
 		detach(s);
 		SharedLitsClause::destroy(0, false);
 		// construct short clause in "this"
-		ClauseHead* h = Clause::newClause(this, s, lits, static_cast<uint32>(j-lits), ClauseInfo());
+		ClauseHead* h = Clause::newClause(this, s, ClauseRep::prepared(lits, static_cast<uint32>(j-lits)));
 		// restore extra data - safe because memory layout has not changed!
 		info_.rep = rep;
 		return h->simplify(s, reinit);
@@ -1182,8 +1062,8 @@ bool LoopFormula::locked(const Solver& s) const {
 	return false;
 }
 
-uint32 LoopFormula::isOpen(const Solver& s, const TypeSet& x, LitVec& freeLits) {
-	if (!x.inSet(Constraint_t::learnt_loop) || (other_ != end_-1 && s.isTrue(lits_[other_]))) {
+uint32 LoopFormula::isOpen(const Solver& s, const TypeSet& xs, LitVec& freeLits) {
+	if (!xs.inSet(Constraint_t::learnt_loop) || (other_ != end_-1 && s.isTrue(lits_[other_]))) {
 		return 0;
 	}
 	for (uint32 x = 1; x != end_-1; ++x) {
@@ -1220,8 +1100,7 @@ bool LoopFormula::simplify(Solver& s, bool) {
 			++j;
 		}
 	}
-	uint32  newEnd    = j + 1;
-	uint32  numBodies = j - 1;
+	uint32 newEnd = j + 1;
 	j += 2;
 	// 2. simplify the set of atoms:
 	// - remove all determined atoms
@@ -1230,7 +1109,7 @@ bool LoopFormula::simplify(Solver& s, bool) {
 		if (s.value(lits_[i].var()) == value_free) {
 			if (i != j) { lits_[j] = lits_[i]; }
 			if (lits_[j].watched()) {
-				if (sat || numBodies <= 2) {
+				if (sat) {
 					s.removeWatch(~lits_[j], this);
 					lits_[j].clearWatch();
 				}
@@ -1255,19 +1134,22 @@ bool LoopFormula::simplify(Solver& s, bool) {
 	end_          = newEnd;
 	lits_[end_]   = Literal();
 	lits_[end_-1] = lits_[end_+1];
-	if (sat || numBodies < 3 || size_ == end_ + 1) {
+	ClauseRep act = ClauseRep::create(lits_+1, end_-1, Constraint_t::learnt_loop);
+	if (sat || size_ == end_ + 1 || (act.isImp() && s.allowImplicit(act))) {
 		for (i = 0; i != bw; ++i) {
 			s.removeWatch(~lits_[bodyWatches[i].second], this);
 			lits_[bodyWatches[i].second].clearWatch();
 		}
 		if (sat || size_ == end_+1) { return true; }
 		// replace constraint with short clauses
-		ClauseCreator creator(&s);
-		creator.start();
-		for (i = 1; i != end_; ++i) { creator.add(lits_[i]); }
 		for (i = end_+1; i != size_; ++i) {
-			creator[creator.size()-1] = lits_[i];
-			creator.end();
+			if (lits_[i].watched()) {
+				s.removeWatch(~lits_[i], this);
+				lits_[i].clearWatch();
+			}
+			lits_[end_-1] = lits_[i];
+			ClauseCreator::Result res = ClauseCreator::create(s, act, ClauseCreator::clause_no_add);
+			CLASP_FAIL_IF(!res.ok() || res.local, "LOOP MUST NOT CONTAIN AUX VARS!");
 		}
 		return true;
 	}
