@@ -24,7 +24,7 @@
 #include <clasp/unfounded_check.h>
 #include <clasp/clause.h>
 #include <algorithm>
-
+#include <cmath>
 namespace Clasp { 
 /////////////////////////////////////////////////////////////////////////////////////////
 // DefaultUnfoundedCheck - Construction/Destruction
@@ -72,8 +72,10 @@ namespace Clasp {
 DefaultUnfoundedCheck::DefaultUnfoundedCheck()
 	: solver_(0) 
 	, graph_(0)
+	, mini_(0)
 	, reasons_(0)
 	, strategy_(common_reason) {
+	mini_.release();
 }
 DefaultUnfoundedCheck::~DefaultUnfoundedCheck() { 
 	for (ExtVec::size_type i = 0; i != extended_.size(); ++i) {
@@ -84,27 +86,46 @@ DefaultUnfoundedCheck::~DefaultUnfoundedCheck() {
 /////////////////////////////////////////////////////////////////////////////////////////
 // DefaultUnfoundedCheck - Initialization
 /////////////////////////////////////////////////////////////////////////////////////////
-void DefaultUnfoundedCheck::attachTo(Solver& s, DependencyGraph* graph) {
-	assert(solver_ == 0 && graph_ == 0);
-	solver_  = &s;
-	graph_   = graph;
-	strategy_= static_cast<ReasonStrategy>(s.strategies().loopRep);
-	s.addPost(this);
+void DefaultUnfoundedCheck::addWatch(Literal p, uint32 data, WatchType type) {
+	assert(data < varMax);
+	solver_->addWatch(p, this, static_cast<uint32>((data << 2) | type));
 }
 
 // inits unfounded set checker with graph, i.e.
 // - creates data objects for bodies and atoms
 // - adds necessary watches to the solver
 // - initializes source pointers
-bool DefaultUnfoundedCheck::init(Solver&) {
-	assert(solver_ && graph_ && "DefaultUnfoundedCheck::attachTo() not called!");
+bool DefaultUnfoundedCheck::init(Solver& s) {
+	assert(!graph_ || s.sharedContext()->sccGraph.get() == graph_);
+	if (s.sharedContext()->sccGraph.get() == 0) { 
+		s.removePost(this);
+		return true; 
+	}
+	solver_      = &s;
+	graph_       = s.sharedContext()->sccGraph.get();
+	strategy_    = s.strategy.search != SolverStrategies::no_learning ? static_cast<ReasonStrategy>(s.configuration().loopRep) : no_reason;
+	if (strategy_ == only_reason) {
+		delete [] reasons_;
+		reasons_ = new LitVec[solver_->numVars()];
+	}
+	// process any leftovers from previous steps
+	while (findUfs(s, false) != ufs_none) {
+		while (!ufs_.empty()) {
+			if (!s.force(~graph_->getAtom(ufs_.front()).lit, 0)) { return false; }
+			atoms_[ufs_.pop_ret()].ufs = 0;
+		}
+	}
 	AtomVec::size_type startAtom = atoms_.size();
 	// set up new atoms
 	atoms_.resize(graph_->numAtoms());
+	AtomData& sentinel = atoms_[SharedDependencyGraph::sentinel_atom];
+	sentinel.resurrectSource();
+	sentinel.todo = 1;
+	sentinel.ufs  = 1;
 	// set up new bodies
 	for (uint32 i = (uint32)bodies_.size(); i != graph_->numBodies(); ++i) {
 		bodies_.push_back(BodyData());
-		BodyNodeP n = graph_->getBody(i);
+		BodyPtr n(getBody(i));
 		if (!n.node->extended()) {
 			initBody(n);
 		}
@@ -112,28 +133,27 @@ bool DefaultUnfoundedCheck::init(Solver&) {
 			initExtBody(n);
 		}
 		// when a body becomes false, it can no longer be used as source
-		solver_->addWatch(~n.node->lit, this, (n.id << 1));
+		addWatch(~n.node->lit, n.id, watch_source_false);
 	}
 	// check for initially unfounded atoms
 	propagateSource();
 	for (AtomVec::size_type i = startAtom, end = atoms_.size(); i != end; ++i) {
-		const DependencyGraph::AtomNode& a = graph_->getAtomNode(NodeId(i));
+		const AtomNode& a = graph_->getAtom(NodeId(i));
 		if (!atoms_[i].hasSource() && !solver_->force(~a.lit, 0)) {
 			return false;
 		}
 		if (a.inChoice()) {
-			addExtWatch(~a.lit, NodeId(i));
+			addWatch(~a.lit, NodeId(i), watch_head_false);
 		}
 	}
-	if (strategy_ == only_reason) {
-		delete [] reasons_;
-		reasons_ = new LitVec[solver_->numVars()];
+	if (graph_->numNonHcfs() != 0) {
+		mini_ = new MinimalityCheck(s.searchConfig().fwdCheck);
 	}
 	return true;
 }
 
 // initializes a "normal" body, i.e. a body where lower(B) == size(B)
-void DefaultUnfoundedCheck::initBody(const BodyNodeP& n) {
+void DefaultUnfoundedCheck::initBody(const BodyPtr& n) {
 	assert(n.id < bodies_.size());
 	BodyData& data = bodies_[n.id];
 	// initialize lower to the number of predecessors from same scc that currently
@@ -144,14 +164,14 @@ void DefaultUnfoundedCheck::initBody(const BodyNodeP& n) {
 
 // initializes an "extended" body, i.e. a count/sum
 // creates & populates WS and adds watches to all subgoals
-void DefaultUnfoundedCheck::initExtBody(const BodyNodeP& n) {
+void DefaultUnfoundedCheck::initExtBody(const BodyPtr& n) {
 	assert(n.id < bodies_.size() && n.node->extended());
 	BodyData& data = bodies_[n.id];
 	uint32 preds   = n.node->num_preds();
 	ExtData* extra = new (::operator new(sizeof(ExtData) + (ExtData::flagSize(preds)*sizeof(uint32)))) ExtData(n.node->ext_bound(), preds);
 
-	InitExtWatches addWatches = { this, extra };
-	graph_->visitBodyLiterals(n, addWatches);
+	InitExtWatches addWatches = { this, &n, extra };
+	graph_->visitBodyLiterals(*n.node, addWatches);
 	
 	data.lower_or_ext = (uint32)extended_.size();
 	extended_.push_back(extra);
@@ -159,54 +179,26 @@ void DefaultUnfoundedCheck::initExtBody(const BodyNodeP& n) {
 }
 
 // set n as source for its heads if possible and necessary
-void DefaultUnfoundedCheck::initSuccessors(const BodyNodeP& n, weight_t lower) {
+void DefaultUnfoundedCheck::initSuccessors(const BodyPtr& n, weight_t lower) {
 	if (!solver_->isFalse(n.node->lit)) {
-		for (const NodeId* x = n.node->heads(); x != n.node->heads_end(); ++x) {
-			const AtomNodeP& a = graph_->getAtom(*x);
-			if (a.node->scc != n.node->scc || lower <= 0) {
-				setSource(a, n);
+		for (const NodeId* x = n.node->heads_begin(); x != n.node->heads_end(); ++x) {
+			const AtomNode& a = graph_->getAtom(*x);
+			if (a.scc != n.node->scc || lower <= 0) {
+				setSource(*x, n);
 			}
 		}
 	}
 }
 
 // watches needed to implement extended rules
-void DefaultUnfoundedCheck::addExtWatch(Literal p, const BodyNodeP& B, uint32 data) {
+void DefaultUnfoundedCheck::addExtWatch(Literal p, const BodyPtr& B, uint32 data) {
+	addWatch(p, static_cast<uint32>(watches_.size()), watch_subgoal_false);
 	ExtWatch w = { B.id, data };
-	solver_->addWatch(p, this, (uint32(watches_.size())<<1)+1);
 	watches_.push_back(w);
 }
-void DefaultUnfoundedCheck::addExtWatch(Literal p, uint32 data) {
-	ExtWatch w = { idMax, data };
-	solver_->addWatch(p, this, (uint32(watches_.size())<<1)+1);
-	watches_.push_back(w);
-}
-
 /////////////////////////////////////////////////////////////////////////////////////////
-// DefaultUnfoundedCheck - Constraint interface
+// DefaultUnfoundedCheck - Base interface
 /////////////////////////////////////////////////////////////////////////////////////////
-// a (relevant) literal was assigned. Check which node is affected and update it accordingly
-Constraint::PropResult DefaultUnfoundedCheck::propagate(Solver& s, Literal, uint32& data) {
-	assert(&s == solver_); (void)s;
-	uint32 index  = data >> 1;
-	uint32 type   = data & 1;
-	if (type == 0) { // a body became false
-		// update source if necessary
-		if (bodies_[index].watches > 0) { invalid_.push_back(index); }
-	}
-	else if (watches_[index].bodyId == idMax) { // head of a choice rule became false
-		// normally head false -> body false and hence the head has its source 
-		// autmatically removed when the invalid body is processed;
-		// for choice rules we must force source removal explicity
-		const ExtWatch w = watches_[index];
-		invalidExt_.push_back((w.data << 1) | ExtWatch::watch_choice_false);
-	}
-	else { // a literal relevant to a body became false
-		invalidExt_.push_back((index << 1) | ExtWatch::watch_body_goal_false);
-	}
-	return PropResult(true, true);  // always keep the watch
-}
-
 void DefaultUnfoundedCheck::reason(Solver&, Literal p, LitVec& r) {
 	LitVec::const_iterator it, end;
 	if (!activeClause_.empty() && activeClause_[0] == p) {
@@ -221,50 +213,54 @@ void DefaultUnfoundedCheck::reason(Solver&, Literal p, LitVec& r) {
 	for (; it != end; ++it) r.push_back( ~*it );
 }
 
-/////////////////////////////////////////////////////////////////////////////////////////
-// DefaultUnfoundedCheck - Base interface
-/////////////////////////////////////////////////////////////////////////////////////////
+bool DefaultUnfoundedCheck::propagateFixpoint(Solver& s, PostPropagator* ctx) {
+	bool checkMin = ctx == 0 && mini_.get() && mini_->partialCheck(s.decisionLevel());
+	for (UfsType t; (t = findUfs(s, checkMin)) != ufs_none; ) {
+		if (!falsifyUfs(t)) { 
+			resetTodo(); 
+			return false;
+		}
+	}
+	return true;
+}
+
 void DefaultUnfoundedCheck::reset() {
-	assert(loopAtoms_.empty());
-	for (VarVec::size_type i = 0; i != todo_.vec_.size(); ++i) {
-		atoms_[todo_.vec_[i]].todo = 0;
+	assert(loopAtoms_.empty() && sourceQ_.empty() && ufs_.empty() && todo_.empty());
+	// remember assignments from top-level -
+	// the reset may come from a stop request and we might
+	// want to continue later
+	if (solver_->decisionLevel()) {
+		invalidQ_.clear();
 	}
-	todo_.clear();
-	for (VarVec::size_type i = 0; i != unfounded_.vec_.size(); ++i) {
-		atoms_[unfounded_.vec_[i]].ufs = 0;
-	}
-	unfounded_.clear();
-	invalidExt_.clear();
-	invalid_.clear();
-	activeClause_.clear();
 }
-
-bool DefaultUnfoundedCheck::propagateFixpoint(Solver&) {
-	return solver_->strategies().search != SolverStrategies::no_learning 
-		? assertAtom()
-		: assertSet();
+bool DefaultUnfoundedCheck::valid(Solver& s) {
+	if (!mini_.get() || findNonHcfUfs(s) == ufs_none) { return true; }
+	falsifyUfs(ufs_non_poly);
+	return false;
+}
+bool DefaultUnfoundedCheck::isModel(Solver& s) {
+	return DefaultUnfoundedCheck::valid(s);
 }
-
 /////////////////////////////////////////////////////////////////////////////////////////
 // DefaultUnfoundedCheck - source pointer propagation
 /////////////////////////////////////////////////////////////////////////////////////////
 // propagates recently set source pointers within one strong component.
-void DefaultUnfoundedCheck::propagateSource(bool forceTodo) {
+void DefaultUnfoundedCheck::propagateSource() {
 	for (LitVec::size_type i = 0; i < sourceQ_.size(); ++i) {
 		NodeId atom = sourceQ_[i];
 		if (atoms_[atom].hasSource()) {
 			// propagate a newly added source-pointer
-			graph_->visitAtomSuccessors(atom, AddSource(this));
+			graph_->getAtom(atom).visitSuccessors(AddSource(this));
 		}
 		else {
-			graph_->visitAtomSuccessors(atom, RemoveSource(this, forceTodo));
+			graph_->getAtom(atom).visitSuccessors(RemoveSource(this));
 		}
 	}
 	sourceQ_.clear();
 }
 
 // replaces current source of atom with n
-void DefaultUnfoundedCheck::updateSource(AtomData& atom, const BodyNodeP& n) {
+void DefaultUnfoundedCheck::updateSource(AtomData& atom, const BodyPtr& n) {
 	if (atom.watch() != AtomData::nill_source) {
 		--bodies_[atom.watch()].watches;
 	}
@@ -273,10 +269,9 @@ void DefaultUnfoundedCheck::updateSource(AtomData& atom, const BodyNodeP& n) {
 }
 
 // an atom in extended body n has a new source, check if n is now a valid source
-void DefaultUnfoundedCheck::AddSource::operator()(const BodyNodeP& n, NodeId atomId, uint32 idx) const {
-	assert(n.node->extended() && n.node->get_pred(idx) == atomId);
-	(void)atomId;
-	ExtData* ext = self->extended_[self->bodies_[n.id].lower_or_ext];
+void DefaultUnfoundedCheck::AddSource::operator()(NodeId bodyId, uint32 idx) const {
+	BodyPtr n(self->getBody(bodyId));
+	ExtData* ext = self->extended_[self->bodies_[bodyId].lower_or_ext];
 	if (ext->lower > 0 || self->bodies_[n.id].watches == 0) {
 		// currently not a source - safely add pred to our watch set
 		ext->addToWs(idx, n.node->pred_weight(idx, false));
@@ -287,9 +282,9 @@ void DefaultUnfoundedCheck::AddSource::operator()(const BodyNodeP& n, NodeId ato
 	}
 }
 // an atom in extended body n has lost its source, check if n is no longer a valid source
-void DefaultUnfoundedCheck::RemoveSource::operator()(const BodyNodeP& n, NodeId atomId, uint32 idx) const {
-	assert(n.node->extended() && n.node->get_pred(idx) == atomId); (void)atomId;
-	ExtData* ext = self->extended_[self->bodies_[n.id].lower_or_ext];
+void DefaultUnfoundedCheck::RemoveSource::operator()(NodeId bodyId, uint32 idx) const {
+	BodyPtr n(self->getBody(bodyId));
+	ExtData* ext = self->extended_[self->bodies_[bodyId].lower_or_ext];
 	ext->removeFromWs(idx, n.node->pred_weight(idx, false));
 	if (ext->lower > 0 && self->bodies_[n.id].watches > 0) {
 		// extended bodies don't always become false if a predecessor becomes false
@@ -299,21 +294,21 @@ void DefaultUnfoundedCheck::RemoveSource::operator()(const BodyNodeP& n, NodeId
 }
 
 // n is a valid source again, forward propagate this information to its heads
-void DefaultUnfoundedCheck::forwardSource(const BodyNodeP& n) {
-	for (const NodeId* x = n.node->heads(); x != n.node->heads_end(); ++x) {
-		setSource(graph_->getAtom(*x), n);
+void DefaultUnfoundedCheck::forwardSource(const BodyPtr& n) {
+	for (const NodeId* x = n.node->heads_begin(); x != n.node->heads_end(); ++x) {
+		setSource(*x, n);
 	}
 }
 
 // n is no longer a valid source, forward propagate this information to its heads
-void DefaultUnfoundedCheck::forwardUnsource(const BodyNodeP& n, bool add) {
-	for (const NodeId* x = n.node->heads(); x != n.node->heads_end() && graph_->getAtomNode(*x).scc == n.node->scc; ++x) {
+void DefaultUnfoundedCheck::forwardUnsource(const BodyPtr& n, bool add) {
+	for (const NodeId* x = n.node->heads_begin(); x != n.node->heads_end() && graph_->getAtom(*x).scc == n.node->scc; ++x) {
 		if (atoms_[*x].hasSource() && atoms_[*x].watch() == n.id) {
 			atoms_[*x].markSourceInvalid();
 			sourceQ_.push_back(*x);
 		}
 		if (add && atoms_[*x].watch() == n.id) {
-			enqueueTodo(*x);
+			pushTodo(*x);
 		}
 	}
 }
@@ -321,15 +316,15 @@ void DefaultUnfoundedCheck::forwardUnsource(const BodyNodeP& n, bool add) {
 // sets body as source for head if necessary.
 // PRE: value(body) != value_false
 // POST: source(head) != 0
-void DefaultUnfoundedCheck::setSource(const AtomNodeP& head, const BodyNodeP& body) {
+void DefaultUnfoundedCheck::setSource(NodeId head, const BodyPtr& body) {
 	assert(!solver_->isFalse(body.node->lit));
 	// For normal rules from not false B follows not false head, but
 	// for choice rules this is not the case. Therefore, the 
 	// check for isFalse(head) is needed so that we do not inadvertantly
 	// source a head that is currently false.
-	if (!atoms_[head.id].hasSource() && !solver_->isFalse(head.node->lit)) {
-		updateSource(atoms_[head.id], body);
-		sourceQ_.push_back(head.id);
+	if (!atoms_[head].hasSource() && !solver_->isFalse(graph_->getAtom(head).lit)) {
+		updateSource(atoms_[head], body);
+		sourceQ_.push_back(head);
 	}
 }
 
@@ -337,14 +332,14 @@ void DefaultUnfoundedCheck::setSource(const AtomNodeP& head, const BodyNodeP& bo
 // Heads having the body as source have their source invalidated and are added
 // to the todo queue. Furthermore, source pointer removal is propagated forward
 void DefaultUnfoundedCheck::removeSource(NodeId bodyId) {
-	const DependencyGraph::BodyNode& body = graph_->getBodyNode(bodyId);
-	for (const NodeId* x = body.heads(); x != body.heads_end(); ++x) {
+	const BodyNode& body = graph_->getBody(bodyId);
+	for (const NodeId* x = body.heads_begin(); x != body.heads_end(); ++x) {
 		if (atoms_[*x].watch() == bodyId) {
 			if (atoms_[*x].hasSource()) {
 				atoms_[*x].markSourceInvalid();
 				sourceQ_.push_back(*x);
 			}
-			enqueueTodo(*x);
+			pushTodo(*x);
 		}
 	}
 	propagateSource();
@@ -353,107 +348,121 @@ void DefaultUnfoundedCheck::removeSource(NodeId bodyId) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // DefaultUnfoundedCheck - Finding & propagating unfounded sets
 /////////////////////////////////////////////////////////////////////////////////////////
-bool DefaultUnfoundedCheck::findUnfoundedSet() {
+void DefaultUnfoundedCheck::updateAssignment(Solver& s) {
+	assert(sourceQ_.empty() && ufs_.empty() && pickedExt_.empty());
+	for (VarVec::const_iterator it = invalidQ_.begin(), end = invalidQ_.end(); it != end; ++it) {
+		uint32 index = (*it) >> 2;
+		uint32 type  = (*it) & 3u;
+		if (type == watch_source_false) { 
+			// a body became false - update atoms having the body as source
+			removeSource(index);
+		}
+		else if (type == watch_head_false) {
+			// an atom in the head of a choice rule became false
+			// normally head false -> body false and hence the head has its source autmatically removed
+			// for choice rules we must force source removal explicity
+			if (atoms_[index].hasSource() && !s.isFalse(graph_->getBody(atoms_[index].watch()).lit)) {
+				atoms_[index].markSourceInvalid();
+				graph_->getAtom(index).visitSuccessors(RemoveSource(this, true));
+				propagateSource();
+			}
+		}
+		else if (type == watch_head_true) { 
+			// TODO: approx. ufs for disjunctive lp
+		}
+		else if (type == watch_subgoal_false) { // a literal p relevant to an extended body became false
+			assert(index < watches_.size());
+			const ExtWatch& w    = watches_[index];
+			const BodyNode& body = graph_->getBody(w.bodyId);
+			ExtData*        ext  = extended_[bodies_[w.bodyId].lower_or_ext];
+			ext->removeFromWs(w.data>>1, body.pred_weight(w.data>>1, test_bit(w.data, 0) != 0));
+			if (ext->lower > 0 && bodies_[w.bodyId].watches && !bodies_[w.bodyId].picked && !s.isFalse(body.lit)) {
+				// The body is not a valid source but at least one head atom 
+				// still depends on it: mark body as invalid source
+				removeSource(w.bodyId);
+				pickedExt_.push_back(w.bodyId);
+				bodies_[w.bodyId].picked = 1;
+			}
+		}
+	}
+	for (VarVec::size_type i = 0, end = pickedExt_.size(); i != end; ++i) {
+		bodies_[pickedExt_[i]].picked = 0;
+	}
+	pickedExt_.clear();
+	invalidQ_.clear();
+}
+
+DefaultUnfoundedCheck::UfsType DefaultUnfoundedCheck::findUfs(Solver& s, bool checkMin) {
 	// first: remove all sources that were recently falsified
-	VarVec::size_type unpick = invalid_.size();
-	for (VarVec::size_type i = 0; i != invalidExt_.size(); ++i) {
-		uint32        id = invalidExt_[i] >> 1;
-		ExtWatch::Type t = static_cast<ExtWatch::Type>(invalidExt_[i] & 1u);
-		if (t == ExtWatch::watch_choice_false) {
-			if (atoms_[id].hasSource() && !solver_->isFalse(graph_->getBodyNode(atoms_[id].watch()).lit)) {
-				atoms_[id].markSourceInvalid();
-				sourceQ_.push_back(id);
-				propagateSource(true);
-			}
-		}
-		else if (t == ExtWatch::watch_body_goal_false) {
-			typedef DependencyGraph::BodyNode BodyNode;
-			const ExtWatch& w = watches_[id];
-			BodyData&       b = bodies_[w.bodyId];
-			ExtData*      ext = extended_[b.lower_or_ext];
-			const BodyNode& B = graph_->getBodyNode(w.bodyId);
-			NodeId       pred = w.data >> 1;
-			ext->removeFromWs(pred, B.pred_weight(pred, test_bit(w.data, 0) != 0));
-			if (ext->lower > 0 && b.watches && b.picked == 0 && !solver_->isFalse(B.lit)) {
-				invalid_.push_back(w.bodyId);
-				b.picked = 1;
-			}
-		}
-	}
-	for (VarVec::size_type i = 0; i != invalid_.size(); ++i) {
-		removeSource(invalid_[i]);
-	}
-	for (VarVec::size_type i = unpick; i != invalid_.size(); ++i) {
-		bodies_[invalid_[i]].picked = 0;
-	}
-	invalid_.clear(); invalidExt_.clear();
-	assert(sourceQ_.empty() && unfounded_.empty());
+	updateAssignment(s);
 	// second: try to re-establish sources.
 	while (!todo_.empty()) {
-		NodeId head = dequeueTodo();
-		if (!atoms_[head].hasSource() && !solver_->isFalse(graph_->getAtomNode(head).lit) && !findSource(head)) {
-			return true;  // found an unfounded set - contained in unfounded_
+		NodeId head       = todo_.pop_ret();
+		atoms_[head].todo = 0;
+		if (!atoms_[head].hasSource() && !s.isFalse(graph_->getAtom(head).lit) && !findSource(head)) {
+			return ufs_poly;  // found an unfounded set - contained in unfounded_
 		}
 		assert(sourceQ_.empty());
 	}
 	todo_.clear();
-	return false;     // no unfounded sets
+	return !checkMin ? ufs_none : findNonHcfUfs(s);
 }
 
 // searches a new source for the atom node head.
 // If a new source is found the function returns true.
 // Otherwise the function returns false and unfounded_ contains head
 // as well as atoms with no source that circularly depend on head.
-bool DefaultUnfoundedCheck::findSource(NodeId head) {
-	assert(unfounded_.empty());
-	enqueueUnfounded(head);	// unfounded, unless we find a new source
-	VarVec noSourceYet;
-	bool changed = false;
+bool DefaultUnfoundedCheck::findSource(NodeId headId) {
+	assert(ufs_.empty() && invalidQ_.empty());
 	const NodeId* bodyIt, *bodyEnd;
-	while (!unfounded_.empty()) {
-		head = unfounded_.front();
-		if (!atoms_[head].hasSource()) { // no source
-			unfounded_.pop_front();        // note: current atom is still marked 
-			AtomNodeP headNode(graph_->getAtom(head));
-			for (bodyIt = headNode.node->bodies(), bodyEnd = headNode.node->bodies_end(); bodyIt != bodyEnd; ++bodyIt) {
-				BodyNodeP bodyNode(graph_->getBody(*bodyIt));
+	uint32 newSource = 0;
+	pushUfs(headId); // unfounded, unless we find a new source
+	while (!ufs_.empty()) {
+		headId         = ufs_.pop_ret(); // still marked and in vector!
+		AtomData& head = atoms_[headId];
+		if (!head.hasSource()) {
+			const AtomNode& headNode = graph_->getAtom(headId);
+			for (bodyIt = headNode.bodies_begin(), bodyEnd = headNode.bodies_end(); bodyIt != bodyEnd; ++bodyIt) {
+				BodyPtr bodyNode(getBody(*bodyIt));
 				if (!solver_->isFalse(bodyNode.node->lit)) {
-					if (bodyNode.node->scc != headNode.node->scc || isValidSource(bodyNode)) {
-						atoms_[head].ufs = 0;          // found a new source,
-						setSource(headNode, bodyNode); // set the new source
-						propagateSource();             // and propagate it forward
-						changed = true;                // may source atoms in noSourceYet!
+					if (bodyNode.node->scc != headNode.scc || isValidSource(bodyNode)) {
+						head.ufs = 0;               // found a new source,
+						setSource(headId, bodyNode);// set the new source
+						propagateSource();          // and propagate it forward
+						++newSource;
 						break;
 					}
 					else { addUnsourced(bodyNode); }
 				}
 			}
-			if (!atoms_[head].hasSource()) {
-				noSourceYet.push_back(head);// no source found
+			if (!head.hasSource()) { // still no source - check again once we are done
+				invalidQ_.push_back(headId); 
 			}
 		}
-		else {  // head has a source
-			dequeueUnfounded();
+		else {  // head has a source and is thus not unfounded
+			++newSource;
+			head.ufs = 0;
 		}
 	} // while unfounded_.emtpy() == false
-	unfounded_.clear();
-	if (changed) {
-		// remove all atoms that have a source as they are not unfounded
-		VarVec::iterator it;
-		for (it = noSourceYet.begin(); it != noSourceYet.end(); ++it) {
-			if ( atoms_[*it].hasSource() )   { atoms_[*it].ufs = 0; }
-			else                             { unfounded_.push_back(*it); }
+	ufs_.rewind();
+	if (newSource != 0) {
+		// some atoms in unfounded_ have a new source
+		// clear queue and check possible candidates for atoms that are still unfounded
+		uint32 visited = ufs_.vec.size();
+		ufs_.clear();
+		if (visited != newSource) {
+			// add elements that are still unfounded
+			for (VarVec::iterator it = invalidQ_.begin(), end = invalidQ_.end(); it != end; ++it) {
+				if ( (atoms_[*it].ufs = (1u - atoms_[*it].validS)) == 1 ) { ufs_.push(*it); }
+			}
 		}
 	}
-	else {
-		// all atoms in noSourceYet are unfounded!
-		noSourceYet.swap(unfounded_.vec_);
-	}
-	return unfounded_.empty();
+	invalidQ_.clear();
+	return ufs_.empty();
 }
 
 // checks whether the body can source its heads
-bool DefaultUnfoundedCheck::isValidSource(const BodyNodeP& n) {
+bool DefaultUnfoundedCheck::isValidSource(const BodyPtr& n) {
 	if (!n.node->extended()) {
 		return bodies_[n.id].lower_or_ext == 0;
 	}
@@ -469,7 +478,7 @@ bool DefaultUnfoundedCheck::isValidSource(const BodyNodeP& n) {
 		const NodeId* x  = n.node->preds();
 		uint32       p   = 0;
 		for (; *x != idMax; x += inc, ++p) {
-			if (atoms_[*x].hasSource() && !ext->inWs(p) && !solver_->isFalse(graph_->getAtomNode(*x).lit)) {
+			if (atoms_[*x].hasSource() && !ext->inWs(p) && !solver_->isFalse(graph_->getAtom(*x).lit)) {
 				ext->addToWs(p, n.node->pred_weight(p, false));
 			}
 		}
@@ -487,57 +496,41 @@ bool DefaultUnfoundedCheck::isValidSource(const BodyNodeP& n) {
 
 // enqueues all predecessors of this body that currently lack a source
 // PRE: isValidSource(n) == false
-void DefaultUnfoundedCheck::addUnsourced(const BodyNodeP& n) {
+void DefaultUnfoundedCheck::addUnsourced(const BodyPtr& n) {
 	const uint32 inc = n.node->pred_inc();
 	for (const NodeId* x = n.node->preds(); *x != idMax; x += inc) {
-		if (!atoms_[*x].hasSource() && !solver_->isFalse(graph_->getAtomNode(*x).lit)) {
-			enqueueUnfounded(*x);
+		if (!atoms_[*x].hasSource() && !solver_->isFalse(graph_->getAtom(*x).lit)) {
+			pushUfs(*x);
 		}
 	}
 }
 
-// asserts all atoms of the unfounded set, then propagates
-bool DefaultUnfoundedCheck::assertSet() {
+// falsifies the atoms one by one from the unfounded set stored in unfounded_
+bool DefaultUnfoundedCheck::falsifyUfs(UfsType t) {
 	activeClause_.clear();
-	while (findUnfoundedSet()) {
-		while (!unfounded_.empty() && solver_->force(~graph_->getAtomNode(unfounded_.front()).lit, 0)) {
-			dequeueUnfounded();
-		}
-		if (!unfounded_.empty() || !solver_->propagateUntil(this)) {
-			return false;
-		}
-	}
-	return true;
-}
-
-// as long as an unfounded set U is not empty,
-// - asserts the first non-false atom
-// - propagates
-// - removes the atom from U
-bool DefaultUnfoundedCheck::assertAtom() {
-	while(findUnfoundedSet()) {
-		activeClause_.clear();
-		while (!unfounded_.empty()) {
-			Literal a = graph_->getAtomNode(unfounded_.front()).lit;
-			if (!solver_->isFalse(a) && !(assertAtom(a) && solver_->propagateUntil(this))) {
-				break;
+	for (uint32 rDL = 0; !ufs_.empty(); ) {
+		Literal a = graph_->getAtom(ufs_.front()).lit;
+		if (!solver_->isFalse(a) && !(assertAtom(a, t) && solver_->propagateUntil(this))) {
+			if (t == ufs_non_poly) {
+				mini_->schedNext(solver_->decisionLevel(), false);
 			}
-			assert(solver_->isFalse(a));
-			dequeueUnfounded();
+			break;
 		}
-		if (!loopAtoms_.empty()) {
-			createLoopFormula();
-		}
-		if (!unfounded_.empty()) { 
-			return false; 
-		}
-		unfounded_.clear(); // clear underlying vector
+		assert(solver_->isFalse(a));
+		atoms_[ufs_.pop_ret()].ufs = 0;
+		if      (ufs_.qFront == 1)               { rDL = solver_->decisionLevel(); }
+		else if (solver_->decisionLevel() != rDL){ break; /* atoms may no longer be unfounded after backtracking */ }
 	}
-	return true;
+	if (!loopAtoms_.empty()) {
+		createLoopFormula();
+	}
+	resetUfs();
+	activeClause_.clear();
+	return !solver_->hasConflict();
 }
 
 // asserts an unfounded atom using the selected reason strategy
-bool DefaultUnfoundedCheck::assertAtom(Literal a) {
+bool DefaultUnfoundedCheck::assertAtom(Literal a, UfsType t) {
 	if (solver_->isTrue(a) || strategy_ == distinct_reason || activeClause_.empty()) {
 		// Conflict, first atom of unfounded set, or distinct reason for each atom requested -
 		// compute reason for a being unfounded.
@@ -547,25 +540,25 @@ bool DefaultUnfoundedCheck::assertAtom(Literal a) {
 			createLoopFormula(); 
 		}
 		activeClause_.assign(1, ~a);
-		computeReason();
+		computeReason(t);
 	}
 	activeClause_[0] = ~a;
-	bool noClause = solver_->isTrue(a) || strategy_ == only_reason || (strategy_ == shared_reason && activeClause_.size() > 3);
+	bool noClause = solver_->isTrue(a) || strategy_ == no_reason || strategy_ == only_reason || (strategy_ == shared_reason && activeClause_.size() > 3);
 	if (noClause) {
-		if (!solver_->force(~a, this)){ return false; }
-		if (strategy_ == only_reason) { reasons_[a.var()-1].assign(activeClause_.begin()+1, activeClause_.end()); }
-		else                          { loopAtoms_.push_back(~a); }
+		if (!solver_->force(~a, this))  { return false; }
+		if (strategy_ == only_reason)   { reasons_[a.var()-1].assign(activeClause_.begin()+1, activeClause_.end()); }
+		else if (strategy_ != no_reason){ loopAtoms_.push_back(~a); }
 		return true;
 	}
 	else { // learn nogood and assert ~a
-		return ClauseCreator::create(*solver_, activeClause_, info_).ok();
+		return ClauseCreator::create(*solver_, activeClause_, ClauseCreator::clause_no_prepare, info_).ok();
 	}
 }
 void DefaultUnfoundedCheck::createLoopFormula() {
 	assert(activeClause_.size() > 3);
 	if (loopAtoms_.size() == 1) {
 		activeClause_[0] = loopAtoms_[0];
-		Constraint* ante = ClauseCreator::create(*solver_, activeClause_, info_).local;
+		Constraint* ante = ClauseCreator::create(*solver_, activeClause_, ClauseCreator::clause_no_prepare, info_).local;
 		assert(ante != 0 && solver_->isTrue(loopAtoms_[0]) && solver_->reason(loopAtoms_[0]) == this);
 		solver_->setReason(loopAtoms_[0], ante);
 	}
@@ -584,36 +577,68 @@ void DefaultUnfoundedCheck::createLoopFormula() {
 }
 
 // computes the reason why a set of atoms is unfounded
-void DefaultUnfoundedCheck::computeReason() {
-	assert(reasonExt_.empty());
-	uint32 ufsScc = graph_->getAtomNode(unfounded_.front()).scc;
-	for (VarVec::size_type i = unfounded_.front_; i != unfounded_.vec_.size(); ++i) {
-		const DependencyGraph::AtomNode& atom = graph_->getAtomNode(unfounded_.vec_[i]);
+void DefaultUnfoundedCheck::computeReason(UfsType t) {
+	if (strategy_ == no_reason) { return; }
+	uint32 ufsScc = graph_->getAtom(ufs_.front()).scc;
+	for (VarVec::size_type i = ufs_.qFront; i != ufs_.vec.size(); ++i) {
+		const AtomNode& atom = graph_->getAtom(ufs_.vec[i]);
 		if (!solver_->isFalse(atom.lit)) {
 			assert(atom.scc == ufsScc);
-			for (const NodeId* x = atom.bodies(); x != atom.bodies_end(); ++x) {
-				addIfReason(graph_->getBody(*x), ufsScc);
+			for (const NodeId* x = atom.bodies_begin(); x != atom.bodies_end(); ++x) {
+				BodyPtr body(getBody(*x));
+				if (t == ufs_poly || !body.node->delta()){ addIfReason(body, ufsScc); }
+				else                                     { addDeltaReason(body, ufsScc); }
 			}
 		}
 	}
-	for (VarVec::size_type i = 0; i != reasonExt_.size(); ++i) { bodies_[reasonExt_[i]].picked = 0; }
-	reasonExt_.clear();
+	for (VarVec::size_type i = 0; i != pickedExt_.size(); ++i) { bodies_[pickedExt_[i]].picked = 0; }
+	pickedExt_.clear();
 	info_     = ClauseInfo(Constraint_t::learnt_loop);
-	uint32 dl = solver_->finalizeConflictClause(activeClause_, info_);
-	assert(dl == solver_->decisionLevel() && "Loop nogood must contain a literal from current DL!");
-	double ratio = activeClause_.size()/double(solver_->decisionLevel()+1);
-	if (ratio > 10 && activeClause_.size() > 100 && !solver_->isFalse(activeClause_[0])) {
-		activeClause_.resize(1);
-		// replace reason with current decision sequence
-		// start with last decision so that watches are correctly initialized
-		for (uint32 x = dl; x; --x) {
-			activeClause_.push_back(~solver_->decision(x));
-		}
- 	}
+	uint32 rc = !solver_->isFalse(activeClause_[0]) && activeClause_.size() > 100 && activeClause_.size() > (solver_->decisionLevel() * 10);
+	uint32 dl = solver_->finalizeConflictClause(activeClause_, info_, rc);
+	uint32 cDL= solver_->decisionLevel();
+	assert((t == ufs_non_poly || dl == cDL) && "Loop nogood must contain a literal from current DL!");
+	if (dl < cDL && (dl = solver_->undoUntil(dl, false)) < cDL) {
+		// cancel any active propagation on cDL
+		invalidQ_.clear();
+		for (PostPropagator* n = this->next; n; n = n->next) { n->reset(); }
+	}
 }
 
+void DefaultUnfoundedCheck::addDeltaReason(const BodyPtr& n, uint32 uScc) {
+	if (bodies_[n.id].picked != 0) return;
+	uint32 bodyAbst = solver_->isFalse(n.node->lit) ? solver_->level(n.node->lit.var()) : solver_->decisionLevel() + 1;
+	for (const NodeId* x = n.node->heads_begin(); x != n.node->heads_end(); ++x) {
+		if (*x != SharedDependencyGraph::sentinel_atom) { // normal head
+			if (graph_->getAtom(*x).scc == uScc) {
+				addIfReason(n, uScc);
+			}
+			continue;
+		}
+		else { // disjunctive head
+			uint32  reasonAbst= bodyAbst;
+			Literal reasonLit = n.node->lit;
+			bool    inUfs     = false;
+			Literal aLit;
+			for (++x; *x; ++x) {
+				if (atoms_[*x].ufs == 1) {
+					inUfs = true;
+				}
+				else if (solver_->isTrue(aLit = graph_->getAtom(*x).lit) && solver_->level(aLit.var()) < reasonAbst) {
+					reasonLit  = ~aLit;
+					reasonAbst = solver_->level(reasonLit.var());
+				}
+			}
+			if (inUfs && reasonAbst && reasonAbst <= solver_->decisionLevel()) {
+				addReasonLit(reasonLit);
+			}
+		}
+	}
+	bodies_[n.id].picked = 1;
+	pickedExt_.push_back(n.id);
+}
 // check if n is part of the reason for the current unfounded set
-void DefaultUnfoundedCheck::addIfReason(const BodyNodeP& n, uint32 uScc) {
+void DefaultUnfoundedCheck::addIfReason(const BodyPtr& n, uint32 uScc) {
 	if (solver_->isFalse(n.node->lit)) {
 		if (n.node->scc != uScc) {
 			addReasonLit(n.node->lit);
@@ -622,7 +647,8 @@ void DefaultUnfoundedCheck::addIfReason(const BodyNodeP& n, uint32 uScc) {
 			if (!n.node->extended()) {
 				// body is only a reason if it does not depend on the atoms from the unfounded set
 				for (const NodeId* x = n.node->preds(); *x != idMax; ++x) {
-					if (atoms_[*x].ufs && !solver_->isFalse(graph_->getAtomNode(*x).lit)) {
+					if (atoms_[*x].ufs && !solver_->isFalse(graph_->getAtom(*x).lit)) {
+						// TODO: add to picked?
 						return;
 					}
 				}
@@ -630,7 +656,7 @@ void DefaultUnfoundedCheck::addIfReason(const BodyNodeP& n, uint32 uScc) {
 			}
 			else if (bodies_[n.id].picked == 0) {
 				bodies_[n.id].picked = 1;
-				reasonExt_.push_back(n.id);
+				pickedExt_.push_back(n.id);
 				// Check if the body depends on the atoms from the unfounded set. I.e.
 				// would the body still be false if all but its unfounded literals would be true?
 				ExtData* ext     = extended_[bodies_[n.id].lower_or_ext];
@@ -639,7 +665,7 @@ void DefaultUnfoundedCheck::addIfReason(const BodyNodeP& n, uint32 uScc) {
 				const uint32 inc = n.node->pred_inc();
 				uint32       p   = 0;
 				for (; *x != idMax; x += inc, ++p) {
-					if (!ext->inWs(p) && (atoms_[*x].ufs == 0 || solver_->isFalse(graph_->getAtomNode(*x).lit))) {
+					if (!ext->inWs(p) && (atoms_[*x].ufs == 0 || solver_->isFalse(graph_->getAtom(*x).lit))) {
 						if ( (temp -= n.node->pred_weight(p, false)) <= 0 ) {
 							addReasonLit(n.node->lit);
 							return;
@@ -657,10 +683,10 @@ void DefaultUnfoundedCheck::addIfReason(const BodyNodeP& n, uint32 uScc) {
 	}
 	else if (n.node->scc == uScc && n.node->extended() && bodies_[n.id].picked == 0) {
 		bodies_[n.id].picked = 1;
-		reasonExt_.push_back(n.id);
+		pickedExt_.push_back(n.id);
 		// body is neither false nor a valid source - add all false lits to reason set
 		AddReasonLit addFalseLits = { this };
-		graph_->visitBodyLiterals(n, addFalseLits);
+		graph_->visitBodyLiterals(*n.node, addFalseLits);
 	}
 }
 
@@ -675,4 +701,64 @@ void DefaultUnfoundedCheck::addReasonLit(Literal p) {
 	}
 }
 
+/////////////////////////////////////////////////////////////////////////////////////////
+// DefaultUnfoundedCheck - Minimality check for disjunctive logic programs
+/////////////////////////////////////////////////////////////////////////////////////////
+DefaultUnfoundedCheck::UfsType DefaultUnfoundedCheck::findNonHcfUfs(Solver& s) {
+	assert(invalidQ_.empty() && loopAtoms_.empty());
+	for (SharedDependencyGraph::NonHcfIter hcc = graph_->nonHcfBegin(), hccEnd = graph_->nonHcfEnd(); hcc != hccEnd; ++hcc) {
+		s.stats.addTest(s.numFreeVars() != 0);
+		hcc->second->assumptionsFromAssignment(s, loopAtoms_);
+		if (!hcc->second->test(hcc->first, s, loopAtoms_, invalidQ_) || s.hasConflict()) {
+			uint32 pos = 0, minDL = UINT32_MAX;
+			for (VarVec::const_iterator it = invalidQ_.begin(), end = invalidQ_.end(); it != end; ++it) {
+				if (s.isTrue(graph_->getAtom(*it).lit) && s.level(graph_->getAtom(*it).lit.var()) < minDL) {
+					minDL = s.level(graph_->getAtom(*it).lit.var());
+					pos   = (uint32)ufs_.vec.size();
+				}
+				pushUfs(*it);
+			}
+			if (pos) {
+				std::swap(ufs_.vec.front(), ufs_.vec[pos]);
+			}
+			invalidQ_.clear();
+			loopAtoms_.clear();
+			return ufs_non_poly;
+		}
+		loopAtoms_.clear();
+	}
+	mini_->schedNext(s.decisionLevel(), true);
+	return ufs_none;
+}
+
+DefaultUnfoundedCheck::MinimalityCheck::MinimalityCheck(const FwdCheck& afwd) : fwd(afwd), high(UINT32_MAX), low(0), next(0) {	
+	if (fwd.highPct > 100) { fwd.highPct = 100; }
+	if (fwd.initHigh != 0) { high        = fwd.initHigh; }
+}
+
+bool DefaultUnfoundedCheck::MinimalityCheck::partialCheck(uint32 level) { 
+	if (level < low) {
+		next -= (low - level);
+		low   = level;
+	}
+	return !next || next == level;
+}
+
+void DefaultUnfoundedCheck::MinimalityCheck::schedNext(uint32 level, bool ok) {
+	low  = 0;
+	next = UINT32_MAX;
+	if (!ok) {
+		high = level;
+		next = 0;
+	}
+	else if (fwd.highPct != 0) {
+		double p = fwd.highPct / 100.0;
+		high     = std::max(high, level);
+		low      = level;
+		if (low >= high && fwd.incHigh) {
+			high   = (uint32)std::ceil(low + (low * p));
+		}
+		next     = low + (uint32)std::ceil((high - low) * p);
+	}
+}
 }
