@@ -24,127 +24,27 @@
 #include <clasp/util/misc_types.h>
 #include <algorithm>
 
-namespace Clasp { namespace mt {
-
-struct SharedWeightLits {
-	SharedWeightLits(uint32 size, bool hasWeight) : shared_lits(size, true, hasWeight) {
-		shared_counter = 1; 
-	}
-	Clasp::atomic<int> shared_counter;
-	WeightLits         shared_lits;
-};
-
-SharedWeightLits* toSharedLits(WeightLits* x) {
-	assert(x->sharable() && "WeightConstraintLits are not shareable!");
-#define OFFSETOF(S, x)  (size_t)&reinterpret_cast<const volatile unsigned char&>((((S*)0)->*&S::x))
-	return reinterpret_cast<SharedWeightLits*>(
-		reinterpret_cast<unsigned char*>(x) - OFFSETOF(SharedWeightLits, shared_lits)
-	);
-#undef OFFSETOF 
-}
-
-} // Clasp::mt
-
-WeightLits* WeightLits::clone() {
-	if (sharable()) {
-		++mt::toSharedLits(this)->shared_counter;
-		return this;
-	}
-	else {
-		std::size_t bytes = sizeof(WeightLits) + (size() << uint32(hasWeights()))*sizeof(Literal);
-		WeightLits* x     = new (::operator new(bytes)) WeightLits();
-		std::memcpy(x, this, bytes);
-		return x;
-	}
-}
-
-void WeightLits::destroy() {
-	if (!sharable()) {
-		::operator delete (this);
-	}
-	else if (--mt::toSharedLits(this)->shared_counter == 0) {
-		void* m = mt::toSharedLits(this);
-		::operator delete (m);
-	}
-}
-
-bool WeightLits::unique() const {
-	return !sharable() || mt::toSharedLits(const_cast<WeightLits*>(this))->shared_counter == 1;
-}
-
+namespace Clasp {
 /////////////////////////////////////////////////////////////////////////////////////////
-// WeightConstraint
+// WeightLitsRep
 /////////////////////////////////////////////////////////////////////////////////////////
-bool WeightConstraint::newWeightConstraint(SharedContext& ctx, Literal con, WeightLitVec& lits, weight_t bound, Constraint** out) {
-	assert(ctx.master()->decisionLevel() == 0 && !ctx.isShared());
-	if (out) *out = 0;
-	weight_t sumW = canonicalize(*ctx.master(), lits, bound);
-	if      (bound <= 0)  { return ctx.addUnary(con); } // trivially SAT
-	else if (bound > sumW){ return ctx.addUnary(~con);} // trivially UNSAT
-	Solver& s = *ctx.master();
-	if (s.value(con.var()) != value_free) {
-		// backward propagate
-		weight_t b  = s.isTrue(con) ? (sumW-bound)+1 : bound;
-		bool bpTrue = s.isTrue(con);
-		uint32 i;
-		for (i = 0; i != lits.size() && (b - lits[i].second) <= 0; ++i) {
-			sumW -= lits[i].second;
-			if (bpTrue) bound -= lits[i].second;
-			if (!s.force(bpTrue ? lits[i].first : ~lits[i].first, 0)) {
-				return false;
-			}
-		}
-		if (i == lits.size() || bound <= 0) {
-			return true;
-		}
-		lits.erase(lits.begin(), lits.begin()+i);
-		if (lits.front().second == lits.back().second && lits.front().second != 1) {
-			// disguised cardinality constraint
-			weight_t w = lits.front().second;
-			bound      = (bound+(w-1))/w;
-			sumW       = (sumW+(w-1))/w;
-		}
-		if (bound == 1) { // clause
-			assert(s.isTrue(con));
-			ClauseCreator clause(&s); clause.reserve(lits.size());
-			clause.start();
-			for (i = 0; i != lits.size(); ++i) { clause.add(lits[i].first); }
-			return clause.end();
-		}
-	}
-	bool   hasW = lits.front().second != lits.back().second;
-	uint32 size = (uint32)lits.size()+1;
-	uint32 nb   = sizeof(WeightConstraint) + (size+hasW)*sizeof(UndoInfo);
-	void*  m    = 0;
-	WL*    sL   = 0;
-	if (!ctx.physicalShareProblem()) {
-		nb        = ((nb + 3) / 4)*4;
-		uint32 ls = sizeof(WL) + (size << uint32(hasW))*sizeof(Literal);
-		m         = ::operator new (nb + ls);
-		sL        = new (reinterpret_cast<unsigned char*>(m)+nb) WL(size, false, hasW);
-	}
-	else {
-		m         = ::operator new(nb);
-		void* t   = ::operator new(sizeof(mt::SharedWeightLits) + (size << uint32(hasW))*sizeof(Literal));
-		sL        = &(new (t) mt::SharedWeightLits(size, hasW))->shared_lits;
-	}
-	assert(m && (reinterpret_cast<uintp>(m) & 7u) == 0);
-	Constraint* c = new (m) WeightConstraint(ctx, con, lits, bound, sumW, sL);
-	if (out) *out = c;
-	ctx.add(c);
-	return !s.hasConflict();
-}
-
-// removes assigned and merges duplicate/complementary literals
+// Removes assigned and merges duplicate/complementary literals.
 // return: achievable weight
 // post  : lits is sorted by decreasing weights
-weight_t WeightConstraint::canonicalize(Solver& s, WeightLitVec& lits, weight_t& bound) {
+WeightLitsRep WeightLitsRep::create(Solver& s, WeightLitVec& lits, weight_t bound) {
 	// Step 1: remove assigned/superfluous literals and merge duplicate/complementary ones
 	LitVec::size_type j = 0, other;
+	const weight_t MAX_W= std::numeric_limits<weight_t>::max();
 	for (LitVec::size_type i = 0; i != lits.size(); ++i) {
 		lits[i].first.clearWatch();
 		Literal x = lits[i].first;
-		if (s.value(x.var()) == value_free && lits[i].second > 0) {
+		if (lits[i].second != 0 && s.topValue(x.var()) == value_free) {
+			if (lits[i].second < 0) {
+				lits[i].second = -lits[i].second;
+				lits[i].first  = x = ~lits[i].first;
+				CLASP_ASSERT_CONTRACT_MSG(bound < 0 || (MAX_W-bound) >= lits[i].second, "bound out of range");
+				bound         += lits[i].second;
+			}
 			if (!s.seen(x.var())) { // first time we see x, keep and mark x
 				if (i != j) lits[j] = lits[i];
 				s.markSeen(x);
@@ -175,19 +75,19 @@ weight_t WeightConstraint::canonicalize(Solver& s, WeightLitVec& lits, weight_t&
 		else if (s.isTrue(x)) { bound -= lits[i].second; }
 	}
 	lits.erase(lits.begin()+j, lits.end());
+	if (bound < 0) { bound = 0; }
 	// Step 2: compute min,max, achievable weight and clear flags set in step 1
 	weight_t sumW = 0;
-	uint32 minW   = std::numeric_limits<uint32>::max(), maxW = 0;
+	weight_t minW = MAX_W, maxW = 1;
+	weight_t B    = std::max(bound, 1);
 	for (LitVec::size_type i = 0; i != lits.size(); ++i) {
 		assert(lits[i].second > 0);
 		s.clearSeen(lits[i].first.var());
-		if (uint32(lits[i].second) > maxW) { maxW = lits[i].second; }
-		if (uint32(lits[i].second) < minW) { minW = lits[i].second; }
+		if (lits[i].second > maxW) { maxW = lits[i].second = std::min(lits[i].second, B);  }
+		if (lits[i].second < minW) { minW = lits[i].second; }
+		CLASP_ASSERT_CONTRACT_MSG((MAX_W - sumW) >= lits[i].second, "Sum of weights out of range");
 		sumW += lits[i].second;
 	}
-	if (bound < 0) {
-		bound = 0;
-	}
 	// Step 3: sort by decreasing weight
 	if (maxW != minW) {
 		std::stable_sort(lits.begin(), lits.end(), compose22(
@@ -201,46 +101,174 @@ weight_t WeightConstraint::canonicalize(Solver& s, WeightLitVec& lits, weight_t&
 		sumW  = (sumW+(minW-1))/minW;
 		for (LitVec::size_type i = 0; i != lits.size(); ++i) { lits[i].second = 1; }
 	}
-	return sumW;
+	WeightLitsRep result = { !lits.empty() ? &lits[0] : 0, (uint32)lits.size(), bound, sumW };
+	return result;
 }
 
-WeightConstraint::WeightConstraint(SharedContext& ctx, Literal W, const WeightLitVec& lits, uint32 bound, uint32 sumWeights, WL* out) {
-	const bool hasW = lits.front().second != lits.back().second;
+// Propagates top-level assignment.
+bool WeightLitsRep::propagate(Solver& s, Literal W) {
+	if      (bound <= 0)    { return s.force(W); } // trivially SAT
+	else if (bound > reach) { return s.force(~W);} // trivially UNSAT
+	else if (s.topValue(W.var()) == value_free) { 
+		return true; 
+	}
+	// backward propagate
+	bool bpTrue = s.isTrue(W);
+	weight_t B  = bpTrue ? (reach-bound)+1 : bound;
+	while (lits->second >= B) {
+		reach -= lits->second;
+		if (!s.force(bpTrue ? lits->first : ~lits->first, 0))        { return false; }
+		if ((bpTrue && (bound -= lits->second) <= 0) || --size == 0) { return true;  }
+		++lits;
+	}
+	if (lits->second > 1 && lits->second == lits[size-1].second) {
+		B     = lits->second;
+		bound = (bound + (B-1)) / B;
+		reach = (reach + (B-1)) / B;
+		for (uint32 i = 0; i != size && lits[i].second != 1; ++i) { lits[i].second = 1; }
+	}
+	return true;
+}
+/////////////////////////////////////////////////////////////////////////////////////////
+// WeightConstraint::WL
+/////////////////////////////////////////////////////////////////////////////////////////
+WeightConstraint::WL::WL(uint32 s, bool shared, bool hasW) : sz(s), rc(shared), w(hasW) { }
+uint8* WeightConstraint::WL::address() { return reinterpret_cast<unsigned char*>(this) - (sizeof(uint32) * rc); }
+WeightConstraint::WL* WeightConstraint::WL::clone(){
+	if (shareable()) {
+		++*reinterpret_cast<Clasp::atomic<uint32>*>(address());
+		return this;
+	}
+	else {
+		uint32 litSize = (size() << uint32(weights()))*sizeof(Literal);
+		WL* x          = new (::operator new(sizeof(WL) + litSize)) WL(size(), false, weights());
+		std::memcpy(x->lits, this->lits, litSize);
+		return x;
+	}
+}
+void WeightConstraint::WL::release() {
+	unsigned char* x = address();
+	if (!shareable() || --*reinterpret_cast<Clasp::atomic<uint32>*>(x) == 0) {
+		::operator delete(x);
+	}
+}
+uint32 WeightConstraint::WL::refCount() const {
+	assert(shareable());
+	return *reinterpret_cast<const Clasp::atomic<uint32>*>(const_cast<WL*>(this)->address());
+}
+/////////////////////////////////////////////////////////////////////////////////////////
+// WeightConstraint
+/////////////////////////////////////////////////////////////////////////////////////////
+WeightConstraint::CPair WeightConstraint::create(Solver& s, Literal W, WeightLitVec& lits, weight_t bound, uint32 flags) {
+	return WeightConstraint::create(s, W, WeightLitsRep::create(s, lits, bound), flags);
+}
+WeightConstraint::CPair WeightConstraint::create(Solver& s, Literal W, WeightLitsRep rep, uint32 flags) {
+	CPair res;
+	if (!s.sharedContext()->physicalShareProblem()) { flags |= create_no_share; }
+	if (s.sharedContext()->frozen())                { flags |= (create_no_freeze|create_no_share); }
+	res.con[0] = createImpl(s, W, rep, flags);
+	if ((flags & create_eq_bound) != 0 && res.ok()) {
+		// Adds constraints for W == (lits == bound)
+		++rep.bound;
+		res.con[1] = createImpl(s, ~W, rep, flags);
+	}
+	return res;
+}
+
+WeightConstraint* WeightConstraint::createImpl(Solver& s, Literal W, WeightLitsRep& rep, uint32 flags) {
+	WeightConstraint* conflict = (WeightConstraint*)0x1;
+	bool addSat = (flags&create_sat) != 0 && rep.size;
+	if (!rep.propagate(s, W))                 { return conflict; }
+	if (rep.unsat() || (rep.sat() && !addSat)){ return 0; }
+	if ((rep.bound == 1 || rep.bound == rep.reach) && (flags & create_explicit) == 0) {
+		LitVec clause; clause.reserve(1 + rep.size);
+		Literal bin[2];
+		bool disj = rep.bound == 1; // con == disjunction or con == conjunction
+		bool sat  = false;
+		clause.push_back(W ^ disj);
+		for (LitVec::size_type i = 0; i != rep.size; ++i) { 
+			bin[0] = ~clause[0];
+			bin[1] = rep.lits[i].first ^ disj;
+			if (bin[0] != ~bin[1]) {
+				if (bin[0] != bin[1])                 { clause.push_back(~bin[1]); }
+				if (!s.add(ClauseRep::create(bin, 2))){ return conflict; }
+			}
+			else { sat = true; }
+		}
+		return (sat || ClauseCreator::create(s, clause, 0)) ? 0 : conflict;
+	}
+	assert(rep.open() || (rep.sat() && addSat));
+	uint32 act  = 3u;
+	if (flags & (create_only_btb|create_only_bfb)) {
+		act  = ((flags & (create_only_btb|create_only_bfb)) / create_only_btb) - 1u;
+		act += (act == 2u);
+	}
+	bool   hasW = rep.hasWeights();
+	uint32 size = 1 + rep.size;
+	uint32 nb   = sizeof(WeightConstraint) + (size+uint32(hasW))*sizeof(UndoInfo);
+	uint32 wls  = sizeof(WL) + (size << uint32(hasW))*sizeof(Literal);
+	void*  m    = 0;
+	WL*    sL   = 0;
+	if ((flags & create_no_share) != 0) {
+		nb        = ((nb + 3) / 4)*4;
+		m         = ::operator new (nb + wls);
+		sL        = new (reinterpret_cast<unsigned char*>(m)+nb) WL(size, false, hasW);
+	}
+	else {
+		static_assert(sizeof(Clasp::atomic<uint32>) == sizeof(uint32), "Invalid size!");
+		m         = ::operator new(nb);
+		uint8* t  = (uint8*)::operator new(wls + sizeof(uint32));
+		*(new (t) Clasp::atomic<uint32>()) = 1;
+		sL        = new (t+sizeof(uint32)) WL(size, true, hasW);
+	}
+	assert(m && (reinterpret_cast<uintp>(m) & 7u) == 0);
+	SharedContext*  ctx = (flags & create_no_freeze) == 0 ? const_cast<SharedContext*>(s.sharedContext()) : 0;
+	WeightConstraint* c = new (m) WeightConstraint(s, ctx, W, rep, sL, act);
+	if (!c->integrateRoot(s)) {
+		c->destroy(&s, true);
+		return conflict;
+	}
+	if ((flags & create_no_add) == 0) { s.add(c); }
+	return c;
+}
+WeightConstraint::WeightConstraint(Solver& s, SharedContext* ctx, Literal W, const WeightLitsRep& rep, WL* out, uint32 act) {
+	const bool hasW = rep.hasWeights();
 	lits_           = out;
-	ownsLit_        = !out->sharable();
-	Literal* p      = lits_->lits_;
+	active_         = act;
+	ownsLit_        = !out->shareable();
+	Literal* p      = lits_->lits;
 	Literal* h      = (Literal*)undo_;
-	Solver&  s      = *ctx.master();
 	weight_t w      = 1;
 	Var      mv     = W.var();
-	bound_[FFB_BTB]	= (sumWeights-bound)+1;// ffb-btb
-	bound_[FTB_BFB]	= bound;               // ftb-bfb
-	*p++            = ~W;                  // store constraint literal
-	if (hasW) *p++  = Literal::fromRep(w); // and weight if necessary
-	ctx.setFrozen(W.var(), true);          // exempt from variable elimination
-	active_ = s.value(W.var())==value_free // if unassigned
-		? NOT_ACTIVE                         // both constraints are relevant
-		: FFB_BTB+s.isFalse(W);              // otherwise only one is relevant
-	for (uint32 i = 0, j = 1, end = (uint32)lits.size(); i != end; ++i, ++j) {
-		*p++ = h[j] = lits[i].first;         // store constraint literal
-		w    = lits[i].second;               // followed by weight 
-		if (hasW) *p++= Literal::fromRep(w); // if necessary
-		addWatch(s, j, FTB_BFB);             // watches  lits[idx]
-		addWatch(s, j, FFB_BTB);             // watches ~lits[idx]
-		ctx.setFrozen(h[j].var(), true);     // exempt from variable elimination
+	bound_[FFB_BTB]	= (rep.reach-rep.bound)+1; // ffb-btb
+	bound_[FTB_BFB]	= rep.bound;               // ftb-bfb
+	*p++            = ~W;                      // store constraint literal
+	if (hasW) *p++  = Literal::fromRep(w);     // and weight if necessary
+	if (ctx) ctx->setFrozen(W.var(), true);    // exempt from variable elimination
+	if (s.topValue(W.var()) != value_free) {   // only one direction is relevant
+		active_ = FFB_BTB+s.isFalse(W);
+	}
+	watched_        = 3u - (active_ != 3u || ctx == 0);
+	for (uint32 i = 0, j = 1, end = rep.size; i != end; ++i, ++j) {
+		*p++ = h[j] = rep.lits[i].first;         // store constraint literal
+		w    = rep.lits[i].second;               // followed by weight 
+		if (hasW) *p++= Literal::fromRep(w);     // if necessary
+		addWatch(s, j, FTB_BFB);                 // watches  lits[idx]
+		addWatch(s, j, FFB_BTB);                 // watches ~lits[idx]
+		if (ctx) ctx->setFrozen(h[j].var(), true);// exempt from variable elimination
 		mv   = std::max(mv, h[j].var());
 	}
-	if (isWeight()) ctx.requestData(mv+1); // weight constraints can become unit more than once
+	if (hasW) s.requestData(mv+1);         // weight constraints can become unit more than once
 	// init heuristic
 	uint32 off = active_ != NOT_ACTIVE;
 	h[0]       = W;
-	s.heuristic()->newConstraint(s, h+off, lits.size()+(1-off), Constraint_t::static_constraint);
+	s.heuristic()->newConstraint(s, h+off, rep.size+(1-off), Constraint_t::static_constraint);
 	// init undo stack
 	up_                 = undoStart();     // undo stack is initially empty
 	undo_[0].data       = 0;
 	undo_[up_].data     = 0;
 	setBpIndex(1);                         // where to start back propagation
-	if (active_ == NOT_ACTIVE) {
+	if (s.topValue(W.var()) == value_free){
 		addWatch(s, 0, FTB_BFB);             // watch con in both phases
 		addWatch(s, 0, FFB_BTB);             // in order to allow for backpropagation
 	}
@@ -251,13 +279,14 @@ WeightConstraint::WeightConstraint(SharedContext& ctx, Literal W, const WeightLi
 }
 
 WeightConstraint::WeightConstraint(Solver& s, const WeightConstraint& other) {
-	lits_           = other.lits_->clone();
-	ownsLit_        = 0;
-	Literal* heu    = (Literal*)undo_;
-	heu[0]          = ~lits_->lit(0);
-	bound_[0]	      = other.bound_[0];
-	bound_[1]	      = other.bound_[1];
-	active_         = other.active_;
+	lits_        = other.lits_->clone();
+	ownsLit_     = 0;
+	Literal* heu = (Literal*)undo_;
+	heu[0]       = ~lits_->lit(0);
+	bound_[0]	   = other.bound_[0];
+	bound_[1]	   = other.bound_[1];
+	active_      = other.active_;
+	watched_     = other.watched_;
 	if (active_ == NOT_ACTIVE && s.value(heu[0].var()) == value_free) {
 		addWatch(s, 0, FTB_BFB);  // watch con in both phases
 		addWatch(s, 0, FFB_BTB);  // in order to allow for backpropagation
@@ -277,14 +306,43 @@ WeightConstraint::WeightConstraint(Solver& s, const WeightConstraint& other) {
 	up_ = other.up_;
 }
 
-WeightConstraint::~WeightConstraint() {
-}
-
+WeightConstraint::~WeightConstraint() {}
 Constraint* WeightConstraint::cloneAttach(Solver& other) { 
 	void* m = ::operator new(sizeof(WeightConstraint) + (size()+isWeight())*sizeof(UndoInfo));
 	return new (m) WeightConstraint(other, *this);
 }
 
+bool WeightConstraint::integrateRoot(Solver& s) {
+	if (!s.decisionLevel() || highestUndoLevel(s) >= s.rootLevel() || s.hasConflict()) { return !s.hasConflict(); }
+	// check if constraint has assigned literals
+	uint32 low = s.decisionLevel(), vDL;
+	uint32 np  = 0;
+	for (uint32 i = 0, end = size(); i != end; ++i) {
+		Var v = lits_->var(i);
+		if (s.value(v) != value_free && (vDL = s.level(v)) != 0) { 
+			++np; 
+			s.markSeen(v);
+			low = std::min(low, vDL);
+		}
+	}
+	// propagate assigned literals in assignment order
+	const LitVec& trail = s.trail();
+	const uint32  end   = trail.size() - s.queueSize();
+	GenericWatch* w     = 0;
+	for (uint32 i = s.levelStart(low); i != end && np; ++i) {
+		Literal p = trail[i];
+		if (s.seen(p) && np--) {
+			s.clearSeen(p.var());
+			if (!s.hasConflict() && (w = s.getWatch(trail[i], this)) != 0) {
+				w->propagate(s, p);
+			}
+		}
+	}
+	for (uint32 i = end; i != trail.size() && np; ++i) {
+		if (s.seen(trail[i].var())) { --np; s.clearSeen(trail[i].var()); }
+	}
+	return !s.hasConflict();
+}
 void WeightConstraint::addWatch(Solver& s, uint32 idx, ActiveConstraint c) {
 	// Add watch only if c is relevant.
 	if (uint32(c^1) != active_) {
@@ -293,13 +351,26 @@ void WeightConstraint::addWatch(Solver& s, uint32 idx, ActiveConstraint c) {
 	}
 }
 
-void WeightConstraint::destroy(Solver*, bool) {
-	if (ownsLit_ == 0) { lits_->destroy(); }
+void WeightConstraint::destroy(Solver* s, bool detach) {
+	if (s && detach) {
+		for (uint32 i = 0, end = size(); i != end; ++i) {
+			s->removeWatch( lits_->lit(i), this );
+			s->removeWatch(~lits_->lit(i), this );
+		}
+		for (uint32 i = s->decisionLevel(); i; --i) {
+			s->removeUndoWatch(i, this);
+		}
+	}
+	if (ownsLit_ == 0) { lits_->release(); }
 	void* mem    = static_cast<Constraint*>(this);
 	this->~WeightConstraint();
 	::operator delete(mem);	
 }
 
+void WeightConstraint::setBpIndex(uint32 n){ 
+	if (isWeight()) undo_[0].data = (n<<1)+(undo_[0].data&1); 
+}
+
 // Returns the numerical highest decision level watched by this constraint.
 uint32 WeightConstraint::highestUndoLevel(Solver& s) const {
 	return up_ != undoStart() 
@@ -344,15 +415,22 @@ void WeightConstraint::updateConstraint(Solver& s, uint32 idx, ActiveConstraint
 Constraint::PropResult WeightConstraint::propagate(Solver& s, Literal, uint32& d) {
 	// determine the affected constraint and its body literal
 	ActiveConstraint c  = (ActiveConstraint)(d&1);
+	const uint32   idx  = d >> 1;
 	Literal body        = lit(0, c);
 	if ( uint32(c^1) == active_ || s.isTrue(body) ) {
 		// the other constraint is active or this constraint is already satisfied; 
 		// nothing to do
 		return PropResult(true, true);		
 	}
+	if (idx == 0 && s.decisionLevel() == s.rootLevel() && watched_ == 3u) {
+		watched_ = c;
+		for (uint32 i = 1, end = size(); i != end; ++i) {
+			s.removeWatch(lit(i, c), this);
+		}
+	}
 	// the constraint is not yet satisfied; update it and
 	// check if we can now propagate any literals.
-	updateConstraint(s, d >> 1, c);
+	updateConstraint(s, idx, c);
 	if (bound_[c] <= 0 || (isWeight() && litSeen(0))) {
 		uint32 reasonData = !isWeight() ? UINT32_MAX : up_;
 		if (!litSeen(0)) {
@@ -424,7 +502,16 @@ void WeightConstraint::undoLevel(Solver& s) {
 		bound_[u.constraint()] += weight(u.idx());
 		--up_;
 	}
-	if (!litSeen(0)) active_ = NOT_ACTIVE;
+	if (!litSeen(0)) {
+		active_ = NOT_ACTIVE;
+		if (watched_ < 2u) {
+			ActiveConstraint other = static_cast<ActiveConstraint>(watched_^1);
+			for (uint32 i = 1, end = size(); i != end; ++i) {
+				addWatch(s, i, other);
+			}
+			watched_ = 3u;
+		}
+	}
 }
 
 bool WeightConstraint::simplify(Solver& s, bool) {
@@ -445,9 +532,9 @@ bool WeightConstraint::simplify(Solver& s, bool) {
 		}
 	}
 	if (lits_->unique() && size() > 4 && (up_ - undoStart()) > size()/2) {
-		Literal*     lits = lits_->lits_;
-		const uint32 inc  = 1 + lits_->hasWeights();
-		uint32       end  = lits_->size_*inc;
+		Literal*     lits = lits_->lits;
+		const uint32 inc  = 1 + lits_->weights();
+		uint32       end  = lits_->size()*inc;
 		uint32 i, j, idx  = 1;
 		// find first assigned literal - must be there otherwise undo stack would be empty
 		for (i = inc; s.value(lits[i].var()) == value_free; i += inc) {
@@ -459,7 +546,7 @@ bool WeightConstraint::simplify(Solver& s, bool) {
 		for (j = i, i += inc; i != end; i += inc) {
 			if (s.value(lits[i].var()) == value_free) {
 				lits[j] = lits[i];
-				if (lits_->hasWeights()) { lits[j+1] = lits[i+1]; }
+				if (lits_->weights()) { lits[j+1] = lits[i+1]; }
 				undo_[idx].data = 0;
 				assert(!litSeen(idx));
 				if (Clasp::GenericWatch* w = s.getWatch(lits[i], this)) {
@@ -479,20 +566,18 @@ bool WeightConstraint::simplify(Solver& s, bool) {
 		// clear undo stack & update to new size
 		up_ = undoStart();
 		setBpIndex(1);
-		lits_->size_ = idx;
+		lits_->sz = idx;
 	}
 	return false;
 }
 
 uint32 WeightConstraint::estimateComplexity(const Solver& s) const {
-	weight_t w1 = bound_[0];
-	weight_t w2 = bound_[1];
-	uint32 r    = 2;
-	for (uint32 i = 1; i != size(); ++i) {
+	weight_t B = std::min(bound_[0], bound_[1]);
+	uint32 r   = 2;
+	for (uint32 i = 1; i != size() && B > 0; ++i) {
 		if (s.value(lits_->var(i)) == value_free) {
 			++r;
-			if ( (w1 -= weight(i)) <= 0 ) break;
-			if ( (w2 -= weight(i)) <= 0 ) break;
+			B -= weight(i);
 		}
 	}
 	return r;
