@@ -20,11 +20,11 @@
 #include "test.h"
 #include <clasp/heuristics.h>
 #include <clasp/lookahead.h>
-#include <clasp/program_builder.h>
+#include <clasp/logic_program.h>
 #include <clasp/clause.h>
 #include <clasp/solver.h>
 namespace Clasp { namespace Test {
-
+using namespace Clasp::Asp;
 class DecisionHeuristicTest : public CppUnit::TestFixture {
   CPPUNIT_TEST_SUITE(DecisionHeuristicTest);
 	CPPUNIT_TEST(testTrivial);
@@ -33,33 +33,40 @@ class DecisionHeuristicTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testLookaheadBugNoSimplify);
 	CPPUNIT_TEST(testLookaheadBugDepsNotCleared);
 	CPPUNIT_TEST(testLookaheadBugNoDeps);
+	CPPUNIT_TEST(testLookaheadBugNoNant);
 	CPPUNIT_TEST(testLookaheadStopConflict);
 
 	CPPUNIT_TEST(testBerkmin);
 	CPPUNIT_TEST(testVmtf);
 	CPPUNIT_TEST(testVsids);
+	CPPUNIT_TEST(testVsidsAux);
 	CPPUNIT_TEST(testResurrect);
 	CPPUNIT_TEST(testStrangeLookSeq);
 	CPPUNIT_TEST(testStrangeLookSeq2);
+	CPPUNIT_TEST(testRestrictDetach);
+
+	CPPUNIT_TEST(testDomSignPos);
+	CPPUNIT_TEST(testDomSignNeg);
+	CPPUNIT_TEST(testDomLevel);
+	CPPUNIT_TEST(testDomDynamic);
+	CPPUNIT_TEST(testDomPrio);
+	CPPUNIT_TEST(testDomPrio2);
+	CPPUNIT_TEST(testDomInit);
+	CPPUNIT_TEST(testDomInc);
 	CPPUNIT_TEST_SUITE_END();
 public:
 	void testTrivial() {
 		SharedContext ctx;
-		ctx.master()->setHeuristic(0, new UnitHeuristic(Lookahead::hybrid_lookahead));
+		ctx.master()->heuristic().reset(new UnitHeuristic(Lookahead::hybrid_lookahead));
 		ctx.startAddConstraints();
 		CPPUNIT_ASSERT_EQUAL(true, ctx.endInit());
 	}
 	void testBodyLookahead() {
-		SharedContext ctx1;
-		SharedContext ctx2(ctx1, SharedContext::init_share_symbols);
-		SharedContext ctx3(ctx1, SharedContext::init_share_symbols);
-		
-		ctx1.master()->addPost(new Lookahead(Lookahead::body_lookahead, false));
-		ctx2.master()->addPost(new Lookahead(Lookahead::atom_lookahead, false));
-		ctx3.master()->addPost(new Lookahead(Lookahead::hybrid_lookahead, false));
-
-		ProgramBuilder api;
-		api.startProgram(ctx1, ProgramBuilder::EqOptions().noEq().noScc())
+		SharedContext ctx;
+		ctx.setConcurrency(2);
+		Solver& s = ctx.addSolver();
+		LogicProgram api;
+		api.start(ctx, LogicProgram::AspOptions().noEq().noScc())
 			.setAtomName(1, "x").setAtomName(2, "a1").setAtomName(3, "a2").setAtomName(4, "a3")
 			.setAtomName(5, "b1").setAtomName(6, "b2").setAtomName(7, "b3")
 			.startRule().addHead(1).addToBody(1, false).endRule()
@@ -73,22 +80,22 @@ public:
 			.startRule().addHead(4).addToBody(7, false).endRule()
 			.startRule().addHead(7).addToBody(4, false).endRule()
 		;
-		
-		CPPUNIT_ASSERT_EQUAL(false, api.endProgram() && ctx1.endInit());
-		CPPUNIT_ASSERT_EQUAL(true, api.cloneProgram(ctx2) && ctx2.endInit());
-		CPPUNIT_ASSERT_EQUAL(false, api.cloneProgram(ctx3) && ctx3.endInit());
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		s.addPost(new Lookahead(Lookahead::Params(Lookahead::body_lookahead).addImps(false)));
+		CPPUNIT_ASSERT_EQUAL(false, ctx.attach(s));
+		ctx.detach(s, true);
+		s.addPost(new Lookahead(Lookahead::Params(Lookahead::atom_lookahead).addImps(false)));
+		CPPUNIT_ASSERT_EQUAL(true, ctx.attach(s));
+		ctx.detach(s, true);
+		s.addPost(new Lookahead(Lookahead::Params(Lookahead::hybrid_lookahead).addImps(false)));
+		CPPUNIT_ASSERT_EQUAL(false, ctx.attach(s));
 	}
 	void testAtomLookahead() {
-		SharedContext ctx1;
-		SharedContext ctx2(ctx1, SharedContext::init_share_symbols);
-		SharedContext ctx3(ctx1, SharedContext::init_share_symbols);
-		
-		ctx1.master()->addPost(new Lookahead(Lookahead::body_lookahead, false));
-		ctx2.master()->addPost(new Lookahead(Lookahead::atom_lookahead, false));
-		ctx3.master()->addPost(new Lookahead(Lookahead::hybrid_lookahead, false));
-		
-		ProgramBuilder api;
-		api.startProgram(ctx1, ProgramBuilder::EqOptions().noEq().noScc())
+		SharedContext ctx;
+		ctx.setConcurrency(2);
+		Solver& s = ctx.addSolver();
+		LogicProgram api;
+		api.start(ctx, LogicProgram::AspOptions().noEq().noScc())
 			.setAtomName(1, "x").setAtomName(2, "c1").setAtomName(3, "c2").setAtomName(4, "c3")
 			.setAtomName(5, "a1").setAtomName(6, "a2").setAtomName(7, "a3")
 			.setAtomName(8, "b1").setAtomName(9, "b2").setAtomName(10, "b3")
@@ -106,15 +113,22 @@ public:
 			.startRule().addHead(7).addToBody(10, false).endRule()
 			.startRule().addHead(10).addToBody(7, false).endRule()
 		;
-		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx1.endInit());
-		CPPUNIT_ASSERT_EQUAL(false, api.cloneProgram(ctx2) && ctx2.endInit());
-		CPPUNIT_ASSERT_EQUAL(false, api.cloneProgram(ctx3) && ctx3.endInit());
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		Lookahead::Params p; p.addImps(false);
+		s.addPost(new Lookahead(p.lookahead(Lookahead::body_lookahead)));
+		CPPUNIT_ASSERT_EQUAL(true, ctx.attach(s));
+		ctx.detach(s, true);
+		s.addPost(new Lookahead(p.lookahead(Lookahead::atom_lookahead)));
+		CPPUNIT_ASSERT_EQUAL(false, ctx.attach(s));
+		ctx.detach(s, true);
+		s.addPost(new Lookahead(p.lookahead(Lookahead::hybrid_lookahead)));
+		CPPUNIT_ASSERT_EQUAL(false, ctx.attach(s));
 	}
 
 	void testLookaheadBugNoSimplify() {
 		DecisionHeuristic* lookAtom = new UnitHeuristic(Lookahead::atom_lookahead);
 		SharedContext ctx;
-		ctx.master()->setHeuristic(0, lookAtom);
+		ctx.master()->heuristic().reset(lookAtom);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Literal e = posLit(ctx.addVar(Var_t::atom_var));
@@ -131,11 +145,10 @@ public:
 	void testLookaheadBugDepsNotCleared() {
 		DecisionHeuristic* lookAtom = new UnitHeuristic(Lookahead::atom_lookahead);
 		SharedContext ctx;
-		ctx.master()->setHeuristic(0, lookAtom);
+		ctx.master()->heuristic().reset(lookAtom);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Literal c = posLit(ctx.addVar(Var_t::atom_var));
-		Literal d = posLit(ctx.addVar(Var_t::atom_var));
 		Literal e = posLit(ctx.addVar(Var_t::atom_var));
 		Literal f = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
@@ -154,11 +167,10 @@ public:
 	void testLookaheadBugNoDeps() {
 		DecisionHeuristic* lookAtom = new UnitHeuristic(Lookahead::atom_lookahead);
 		SharedContext ctx;
-		ctx.master()->setHeuristic(0, lookAtom);
+		ctx.master()->heuristic().reset(lookAtom);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Literal c = posLit(ctx.addVar(Var_t::atom_var));
-		Literal d = posLit(ctx.addVar(Var_t::atom_var));
 		Literal e = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
 		ctx.addBinary(a, b);
@@ -172,21 +184,37 @@ public:
 		CPPUNIT_ASSERT(lookAtom->select(s));
 		CPPUNIT_ASSERT(s.isFalse(b));
 	}
+	void testLookaheadBugNoNant() {
+		Clasp::Lookahead::Params p(Lookahead::atom_lookahead);
+		p.restrictNant = true;
+		DecisionHeuristic* lookAtom = new UnitHeuristic(p);
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(lookAtom);
+		Literal a = posLit(ctx.addVar(Var_t::atom_var));
+		Literal b = posLit(ctx.addVar(Var_t::atom_var));
+		Literal c = posLit(ctx.addVar(Var_t::atom_var));
+		Solver& s = ctx.startAddConstraints();
+		ctx.addBinary(a, b);
+		ctx.addBinary(b, c);
+		ctx.addBinary(~a, ~b);
+		ctx.addBinary(~b, ~c);
+		ctx.endInit();
+		uint32 n = s.numFreeVars();
+		CPPUNIT_ASSERT(lookAtom->select(s) && s.numFreeVars() != n);
+	}
 	
 	void testLookaheadStopConflict() {
 		DecisionHeuristic* lookAtom = new UnitHeuristic(Lookahead::atom_lookahead);
 		SharedContext ctx;
-		ctx.master()->setHeuristic(0, lookAtom);
+		ctx.master()->heuristic().reset(lookAtom);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
-		Literal c = posLit(ctx.addVar(Var_t::atom_var));
-		Literal d = posLit(ctx.addVar(Var_t::atom_var));
-		Literal e = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
 		ctx.addBinary(a, b);
 		ctx.endInit();
 		struct StopConflict : public PostPropagator {
-			bool propagate(Solver& s) { s.setStopConflict(); return false; }
+			bool propagateFixpoint(Solver& s, PostPropagator*) { s.setStopConflict(); return false; }
+			uint32 priority() const   { return PostPropagator::priority_class_simple; }
 		}* x = new StopConflict;
 		s.addPost(x);
 		CPPUNIT_ASSERT(!lookAtom->select(s) && s.hasConflict());
@@ -194,10 +222,9 @@ public:
 	}
 
 	void testBerkmin() {
-		ClaspBerkmin* berkmin = new ClaspBerkmin();
+		ClaspBerkmin* berkmin = new ClaspBerkmin(0, HeuParams().other(3).init(1));
 		SharedContext ctx;
-		ctx.master()->strategies().heuOther = 3;
-		ctx.master()->setHeuristic(1, berkmin);
+		ctx.master()->heuristic().reset(berkmin);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Literal c = posLit(ctx.addVar(Var_t::atom_var));
@@ -222,12 +249,12 @@ public:
 		s.assume( ~c );
 		s.assume( ~d );
 		ClauseCreator cc(&s);
-		cc.startAsserting(Constraint_t::learnt_conflict, a).add(b).add(c).add(d).end();
+		cc.start(Constraint_t::learnt_conflict).add(a).add(b).add(c).add(d).end();
 		s.undoUntil(0);
 		s.assume( ~e );
 		s.assume( ~f );
 		s.assume( ~g );
-		cc.startAsserting(Constraint_t::learnt_loop, d).add(e).add(f).add(g).end();
+		cc.start(Constraint_t::learnt_loop).add(d).add(e).add(f).add(g).end();
 		s.undoUntil(0);
 		CPPUNIT_ASSERT_EQUAL(0u, s.numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, berkmin->select(s));
@@ -242,9 +269,9 @@ public:
 	void testVmtf() {
 		ClaspVmtf* vmtf = new ClaspVmtf;
 		SharedContext ctx;
-		ctx.master()->setHeuristic(2, vmtf);
-		Literal a = posLit(ctx.addVar(Var_t::atom_var));
-		Literal b = posLit(ctx.addVar(Var_t::atom_var));
+		ctx.master()->heuristic().reset(vmtf);
+		ctx.addVar(Var_t::atom_var);
+		ctx.addVar(Var_t::atom_var);
 		Solver& s = ctx.startAddConstraints();
 		ctx.endInit();
 		CPPUNIT_ASSERT_EQUAL(true, vmtf->select(s));
@@ -257,7 +284,7 @@ public:
 	void testVsids() {
 		ClaspVsids* vsids = new ClaspVsids;
 		SharedContext ctx;
-		ctx.master()->setHeuristic(3, vsids);
+		ctx.master()->heuristic().reset(vsids);
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
@@ -271,38 +298,66 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, s.trail().back() == ~b && s.propagate());
 		CPPUNIT_ASSERT_EQUAL(false, vsids->select(s));
 	}
+	void testVsidsAux() {
+		ClaspVsids* vsids = new ClaspVsids;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(vsids);
+		Literal a = posLit(ctx.addVar(Var_t::atom_var));
+		Literal b = posLit(ctx.addVar(Var_t::atom_var));
+		Solver& s = ctx.startAddConstraints();
+		ctx.endInit();
+		Var v = s.pushAuxVar();
+		LitVec up;
+		vsids->updateReason(s,up,posLit(v));
+		CPPUNIT_ASSERT_EQUAL(true, vsids->select(s));
+		CPPUNIT_ASSERT(s.value(v) != value_free);
+		s.popAuxVar(1);
+		CPPUNIT_ASSERT_EQUAL(true, vsids->select(s));
+		CPPUNIT_ASSERT(s.trail().back().var() != v);
+	}
 
 	void testStrangeLookSeq() {
 		SharedContext ctx;
 		Literal a = posLit(ctx.addVar(Var_t::body_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
-		Lookahead* look = new Lookahead(Lookahead::body_lookahead);
-		s.addPost(look);
-		ctx.endInit();	
-		s.removePost(look);
-		s.addUnary(a, Constraint_t::learnt_conflict);
+		s.heuristic().reset(UnitHeuristic::restricted(Lookahead::body_lookahead, 3, s.heuristic().release()));
+		ctx.endInit();
+		s.force(a);
 		s.simplify();
-		RestrictedUnit::decorate(s, 1, look);
 		bool x = s.decideNextBranch();
 		CPPUNIT_ASSERT(x == true && s.value(b.var()) != value_free);
 	}
 
 	void testStrangeLookSeq2() {
 		SharedContext ctx;
-		Lookahead* look = new Lookahead(Lookahead::atom_lookahead);
-		ctx.master()->setHeuristic(0, new SelectFirst);
+		ctx.master()->heuristic().reset(UnitHeuristic::restricted(Lookahead::atom_lookahead, 2, new SelectFirst));
 		Literal a = posLit(ctx.addVar(Var_t::atom_var));
 		Literal b = posLit(ctx.addVar(Var_t::atom_var));
 		Solver& s = ctx.startAddConstraints();
-		look->init(s);
 		ctx.addBinary(a, b);
 		ctx.addBinary(a, ~b);
 		ctx.addBinary(~a, b);
 		ctx.endInit();
-		RestrictedUnit::decorate(s, 1, look);
 		bool x = ctx.master()->decideNextBranch();
-		CPPUNIT_ASSERT(x == true && s.decisionLevel() == 0);
+		CPPUNIT_ASSERT(x == false && s.decisionLevel() == 0 && s.numFreeVars() == 0);
+	}
+
+	void testRestrictDetach() {
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(UnitHeuristic::restricted(Lookahead::atom_lookahead, 3, new SelectFirst));
+		Literal a = posLit(ctx.addVar(Var_t::atom_var));
+		Literal b = posLit(ctx.addVar(Var_t::atom_var));
+		Solver& s = ctx.startAddConstraints();
+		ctx.addBinary(a, b);
+		ctx.endInit();
+		bool x = ctx.master()->decideNextBranch();
+		CPPUNIT_ASSERT(x == true && s.decisionLevel() == 1);
+		s.propagate();
+		CPPUNIT_ASSERT(s.numFreeVars() == 0);
+		CPPUNIT_ASSERT(s.getPost(PostPropagator::priority_reserved_look) != 0);
+		s.heuristic().reset(new SelectFirst);
+		CPPUNIT_ASSERT(s.getPost(PostPropagator::priority_reserved_look) == 0);
 	}
 
 	void testResurrect() {
@@ -315,7 +370,7 @@ public:
 		};
 		for (uint32 i = 0; i != 3; ++i) {
 			SharedContext ctx;
-			ctx.master()->strategies().heuristic.reset(heus[i].second);
+			ctx.master()->strategy.heuristic.reset(heus[i].second);
 			Var v1 = ctx.addVar(Var_t::atom_var);
 			Var v2 = ctx.addVar(Var_t::atom_var);
 			Var v3 = ctx.addVar(Var_t::atom_var);
@@ -323,17 +378,185 @@ public:
 			ctx.endInit();
 			Solver& s = *ctx.master();
 			s.eliminate(v1, true);
-			while (s.strategies().heuristic->select(s) && s.propagate()) { ; }
+			while (s.strategy.heuristic->select(s) && s.propagate()) { ; }
 			CPPUNIT_ASSERT(2u == s.stats.choices);
 			CPPUNIT_ASSERT_EQUAL_MESSAGE(heus[i].first, 0u, s.numFreeVars());
 			s.eliminate(v1, false);
 			CPPUNIT_ASSERT_EQUAL(value_free, s.value(v1));
-			CPPUNIT_ASSERT_EQUAL_MESSAGE(heus[i].first, true, s.strategies().heuristic->select(s));
+			CPPUNIT_ASSERT_EQUAL_MESSAGE(heus[i].first, true, s.strategy.heuristic->select(s));
 			CPPUNIT_ASSERT_MESSAGE(heus[i].first, value_free != s.value(v1));
 		}
 		*/
 		CPPUNIT_FAIL("TODO - Resurrection not yet supported!");
 	}
+
+	void testDomSignPos() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx).setAtomName(1, "a").startRule(Asp::CHOICERULE).addHead(1).endRule();
+		addDomRule(api, "_heuristic(a,sign,1)");
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(1)->literal());
+	}
+	void testDomSignNeg() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx).setAtomName(1, "a").startRule(Asp::CHOICERULE).addHead(1).endRule();
+		addDomRule(api, "_heuristic(a,sign,-1)");
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == ~api.getAtom(1)->literal());
+	}
+	void testDomLevel() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx).setAtomName(1, "a").setAtomName(2,"b").startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		addDomRule(api, "_heuristic(a,sign,1)");
+		addDomRule(api, "_heuristic(b,sign,1)");
+		addDomRule(api, "_heuristic(a,level,10)");
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(1)->literal());
+	}
+
+	void testDomDynamic() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx)
+			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule()
+			.startRule().addHead(4).addToBody(1, true).addToBody(2,true).endRule()
+			.setCompute(4, false);
+		addDomRule(api, "_heuristic(a,sign,1)");
+		addDomRule(api, "_heuristic(b,sign,1)");
+		addDomRule(api, "_heuristic(a,level,10)");
+		addDomRule(api, "_heuristic(c,sign,1)", posLit(2));
+		addDomRule(api, "_heuristic(c,sign,-1)", negLit(2));
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(1)->literal());
+		s.propagate();
+		CPPUNIT_ASSERT(s.isFalse(api.getAtom(2)->literal()));
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == ~api.getAtom(3)->literal());
+		s.clearAssumptions();
+		uint32 n = s.numWatches(posLit(2));
+		// test removal of watches
+		ctx.master()->heuristic().reset(new SelectFirst());
+		CPPUNIT_ASSERT_MESSAGE("Heuristic not detached", s.numWatches(posLit(2)) != n);
+	}
+
+	void testDomPrio() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx)
+			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule()
+			.startRule().addHead(4).addToBody(1, true).addToBody(2,true).endRule()
+			.setCompute(4, false);
+		addDomRule(api, "_heuristic(b,sign,1)");
+		addDomRule(api, "_heuristic(a,true,10)");
+		addDomRule(api, "_heuristic(c,sign,1,10)");
+		addDomRule(api, "_heuristic(c,sign,-1,20)", negLit(2));
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(1)->literal());
+		s.propagate();
+		CPPUNIT_ASSERT(s.isFalse(api.getAtom(2)->literal()));
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == ~api.getAtom(3)->literal());
+	}
+	void testDomPrio2() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx)
+			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).endRule()
+			.startRule().addHead(4).addToBody(1, true).addToBody(2,true).endRule()
+			.setCompute(4, false);
+		addDomRule(api, "_heuristic(b,sign,1)");
+		addDomRule(api, "_heuristic(a,true,10)");
+		addDomRule(api, "_heuristic(c,sign,1,30)");
+		addDomRule(api, "_heuristic(c,sign,-1,20)", negLit(2));
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(1)->literal());
+		s.propagate();
+		CPPUNIT_ASSERT(s.isFalse(api.getAtom(2)->literal()));
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back() == api.getAtom(3)->literal());
+	}
+	void testDomInit() {
+		DomainHeuristic* dom = new DomainHeuristic;
+		SharedContext ctx;
+		ctx.master()->heuristic().reset(dom);
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx)
+			.setAtomName(1, "a").setAtomName(2, "b")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).endRule();
+		addDomRule(api, "_heuristic(a,init,10,20)");
+		addDomRule(api, "_heuristic(a,init,50,10)");
+		addDomRule(api, "_heuristic(b,init,10,10)");
+		addDomRule(api, "_heuristic(b,init,30,20)");
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		CPPUNIT_ASSERT(dom->select(s));
+		CPPUNIT_ASSERT(s.trail().back().var() == api.getAtom(2)->var());
+	}
+	void testDomInc() {
+		SharedContext ctx;
+		Solver& s = *ctx.master();
+		LogicProgram api;
+		api.start(ctx).updateProgram();
+		api.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "x").setAtomName(4, "y")
+			.startRule(Asp::CHOICERULE).addHead(1).addHead(2).addHead(3).addHead(4).endRule();
+		addDomRule(api, "_heuristic(a,level,1)", posLit(3));
+		addDomRule(api, "_heuristic(b,level,1)", posLit(4));
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		uint32 n = s.numWatches(posLit(3));
+		DomainHeuristic* dom = new DomainHeuristic;
+		dom->startInit(s);
+		dom->endInit(s);
+		CPPUNIT_ASSERT(s.numWatches(posLit(3)) > n);
+		CPPUNIT_ASSERT(api.updateProgram());
+		Var c = api.newAtom();
+		api.setAtomName(c, "c").startRule(Asp::CHOICERULE).addHead(c).endRule();
+		addDomRule(api, "_heuristic(c,level,1)", posLit(3));
+		CPPUNIT_ASSERT_EQUAL(true, api.endProgram() && ctx.endInit());
+		dom->startInit(s);
+		dom->endInit(s);
+		delete dom;
+		CPPUNIT_ASSERT_MESSAGE("Heuristic not detached", s.numWatches(posLit(3)) == n);
+	}
+	void addDomRule(LogicProgram& prg, const char* heu, Literal pre = posLit(0)) {
+		Var h = prg.newAtom();
+		prg.setAtomName(h, heu);
+		prg.startRule().addHead(h);
+		if (pre != posLit(0)) prg.addToBody(pre.var(), pre.sign() == false);
+		prg.endRule();
+	}
 };
 
 CPPUNIT_TEST_SUITE_REGISTRATION(DecisionHeuristicTest);
