@@ -20,18 +20,19 @@
 #include "test.h"
 #include <algorithm>
 #include <clasp/minimize_constraint.h>
-
 #include <clasp/solver.h>
-
+#include <clasp/solve_algorithms.h>
 namespace Clasp { namespace Test {
-class MinimizeTest : public CppUnit::TestFixture {
-	CPPUNIT_TEST_SUITE(MinimizeTest);
+class DefaultMinimizeTest : public CppUnit::TestFixture {
+	CPPUNIT_TEST_SUITE(DefaultMinimizeTest);
 	CPPUNIT_TEST(testEmpty);
 	CPPUNIT_TEST(testOneLevelLits);
 	CPPUNIT_TEST(testMultiLevelLits);
 	CPPUNIT_TEST(testMultiLevelWeightsAreReused);
 	CPPUNIT_TEST(testMergeComplementaryLits);
 	CPPUNIT_TEST(testMergeComplementaryLits2);
+	CPPUNIT_TEST(testNegativeLowerInit);
+	CPPUNIT_TEST(testNegativeLower);
 
 	CPPUNIT_TEST(testOrder);
 	CPPUNIT_TEST(testSkipLevel);
@@ -46,6 +47,7 @@ class MinimizeTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testRootLevelMadness);
 	CPPUNIT_TEST(testIntegrateOptimum);
 	CPPUNIT_TEST(testIntegrateOptimumConflict);
+	CPPUNIT_TEST(testIntegrateBug);
 
 	CPPUNIT_TEST(testReasonBug);
 	CPPUNIT_TEST(testSmartBacktrack);
@@ -63,28 +65,27 @@ class MinimizeTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testHierarchicalSetModel);
 	CPPUNIT_TEST(testHierarchical);
 	CPPUNIT_TEST(testInconsistent);
-
 	CPPUNIT_TEST_SUITE_END(); 
 public:
-	MinimizeTest() {
-		a     = posLit(ctx.addVar(Var_t::atom_var));
-		b     = posLit(ctx.addVar(Var_t::atom_var));
-		c     = posLit(ctx.addVar(Var_t::atom_var));
-		d     = posLit(ctx.addVar(Var_t::atom_var));
-		e     = posLit(ctx.addVar(Var_t::atom_var));
-		f     = posLit(ctx.addVar(Var_t::atom_var));
-		x     = posLit(ctx.addVar(Var_t::atom_var));
-		y     = posLit(ctx.addVar(Var_t::atom_var));
+	DefaultMinimizeTest() {
+		a = posLit(ctx.addVar(Var_t::atom_var));
+		b = posLit(ctx.addVar(Var_t::atom_var));
+		c = posLit(ctx.addVar(Var_t::atom_var));
+		d = posLit(ctx.addVar(Var_t::atom_var));
+		e = posLit(ctx.addVar(Var_t::atom_var));
+		f = posLit(ctx.addVar(Var_t::atom_var));
+		x = posLit(ctx.addVar(Var_t::atom_var));
+		y = posLit(ctx.addVar(Var_t::atom_var));
 		ctx.startAddConstraints();
 	}
 	void setUp()    { newMin = 0; data = 0; }
 	void tearDown() { 
 		if (newMin) { newMin->destroy(ctx.master(), true);  } 
-		if (data)   { data->destroy(); }
+		if (data)   { data->release(); }
 	}
-
 	void testEmpty() {
-		newMin = MinimizeBuilder().buildAndAttach(ctx);
+		MinimizeBuilder b;
+		newMin = buildAndAttach(b);
 		CPPUNIT_ASSERT(countMinLits() == 0);
 	}
 	
@@ -98,16 +99,16 @@ public:
 		min.push_back( WeightLiteral(a, 2) ); // duplicate lit
 		min.push_back( WeightLiteral(d, 1) );
 		min.push_back( WeightLiteral(e, 2) ); // false lit
-		newMin = MinimizeBuilder().addRule(min).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(min));
 		CPPUNIT_ASSERT(newMin->numRules() == 1);
 		CPPUNIT_ASSERT(countMinLits() == 3);
-		CPPUNIT_ASSERT(newMin->shared()->sum(0) == 1);
+		CPPUNIT_ASSERT(newMin->shared()->adjust(0) == 1);
 		const SharedMinimizeData* data = newMin->shared();
 		CPPUNIT_ASSERT(data->lits[0].first == a && data->lits[0].second == 3);
 		for (const WeightLiteral* it = data->lits; !isSentinel(it->first); ++it) {
 			CPPUNIT_ASSERT_MESSAGE("Minimize lits not sorted!", it->second >= (it+1)->second);
 			CPPUNIT_ASSERT(ctx.master()->hasWatch(it->first, newMin));
-			CPPUNIT_ASSERT(ctx.frozen(it->first.var()));
+			CPPUNIT_ASSERT(ctx.varInfo(it->first.var()).frozen());
 		}
 		newMin->destroy(ctx.master(), true);
 		newMin = 0;
@@ -142,11 +143,11 @@ public:
 		min.push_back( WeightLiteral(d, 1) ); // duplicate lit
 		builder.addRule(min);
 		
-		newMin = builder.buildAndAttach(ctx);
+		newMin = buildAndAttach(builder);
 		CPPUNIT_ASSERT(newMin->numRules() == 3);
 		CPPUNIT_ASSERT(countMinLits() == 6);
 		const SharedMinimizeData* data = newMin->shared();
-		CPPUNIT_ASSERT(data->sum(0) == 1 && data->sum(1) == 0 && data->sum(2) == 2);
+		CPPUNIT_ASSERT(data->adjust(0) == 1 && data->adjust(1) == 0 && data->adjust(2) == 2);
 		CPPUNIT_ASSERT(data->lits[0].first == b);
 		CPPUNIT_ASSERT(data->weights.size() == 11);
 		for (const WeightLiteral* it = data->lits; !isSentinel(it->first); ++it) {
@@ -169,7 +170,7 @@ public:
 		min.push_back( WeightLiteral(d, 1) );
 		builder.addRule(min);
 		
-		newMin = builder.buildAndAttach(ctx);
+		newMin = buildAndAttach(builder);
 		// b = 0
 		// d = 0
 		// a = 2
@@ -191,12 +192,12 @@ public:
 		min.push_back( WeightLiteral(~c, 1) );
 		min.push_back( WeightLiteral(e, 1) );
 		builder.addRule(min);
-		newMin = builder.buildAndAttach(ctx);
+		newMin = buildAndAttach(builder);
 		CPPUNIT_ASSERT(countMinLits() == 5);
-		CPPUNIT_ASSERT(newMin->shared()->sum(0) == 1 && newMin->shared()->sum(1) == 1);
+		CPPUNIT_ASSERT(newMin->shared()->adjust(0) == 1 && newMin->shared()->adjust(1) == 1);
 
 		ctx.master()->assume(c) && ctx.master()->propagate();
-		CPPUNIT_ASSERT(newMin->sum(1) == 0);
+		CPPUNIT_ASSERT(newMin->sum(1, true) == 0);
 
 	}
 
@@ -209,28 +210,73 @@ public:
 		min.clear();
 		min.push_back( WeightLiteral(a, 1) );
 		builder.addRule(min);
-		newMin = builder.buildAndAttach(ctx);
+		newMin = buildAndAttach(builder);
 		CPPUNIT_ASSERT(countMinLits() == 1);
-		CPPUNIT_ASSERT(newMin->shared()->sum(0) == 1 && newMin->shared()->sum(1) == 0);
+		CPPUNIT_ASSERT(newMin->shared()->adjust(0) == 1 && newMin->shared()->adjust(1) == 0);
+	}
+	void testNegativeLowerInit() {
+		WeightLitVec aMin, bMin, cMin;
+		aMin.push_back( WeightLiteral(a, 1) );
+		aMin.push_back( WeightLiteral(b, 1) );
+		bMin.push_back( WeightLiteral(~a, 1) );
+		bMin.push_back( WeightLiteral(~b, 1) );
+		cMin.push_back( WeightLiteral(a, 1) );
+		cMin.push_back( WeightLiteral(b, 1) );
+		data = MinimizeBuilder()
+			.addRule(aMin)
+			.addRule(bMin)
+			.addRule(cMin)
+			.build(ctx);
+		CPPUNIT_ASSERT(data->lower(0) == 0);
+		CPPUNIT_ASSERT(data->lower(1) == -2);
+		CPPUNIT_ASSERT(data->lower(2) == 0);
+	}
+	void testNegativeLower() {
+		ctx.addBinary(a, b);
+		WeightLitVec aMin, bMin, cMin;
+		aMin.push_back( WeightLiteral(a, 1) );
+		aMin.push_back( WeightLiteral(b, 1) );
+		bMin.push_back( WeightLiteral(~a, 1) );
+		cMin.push_back( WeightLiteral(a, 1) );
+		data = MinimizeBuilder()
+			.addRule(aMin)
+			.addRule(bMin)
+			.addRule(cMin)
+			.build(ctx);
+		newMin = createMin(ctx, *ctx.master(), data, SolverStrategies::opt_hier);
+		newMin->integrateBound(*ctx.master());
+		
+		Solver& s = *ctx.master();
+		s.assume(b) && s.propagate();
+		s.assume(~a) && s.propagate();
+		newMin->handleModel(s);
+		s.undoUntil(0);
+		CPPUNIT_ASSERT(!newMin->integrate(s) || !s.propagate());
+		LitVec ignore;
+		CPPUNIT_ASSERT(newMin->handleUnsat(s, true, ignore));
+		CPPUNIT_ASSERT(newMin->integrate(s) && s.propagate());
+		
+		CPPUNIT_ASSERT(s.isFalse(b));
+		CPPUNIT_ASSERT(s.assume(a) && s.propagate());
+		newMin->handleModel(s);
 	}
 	
 	void testOrder() {
 		WeightLitVec aMin, bMin;
 		aMin.push_back( WeightLiteral(a, 1) );
 		bMin.push_back( WeightLiteral(b, 1) );
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		
 		Solver& solver = *ctx.master();
 		solver.assume(b);
 		CPPUNIT_ASSERT_EQUAL(true, solver.propagate());
 		solver.assume(~a);
 		CPPUNIT_ASSERT_EQUAL(true, solver.propagate());
-		CPPUNIT_ASSERT(newMin->sum(0) == 0 && newMin->sum(1) == 1);
-		solver.undoUntil(newMin->commitCurrent(solver));
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->sum(0, true) == 0 && newMin->sum(1, true) == 1);
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.isFalse(a));
 		CPPUNIT_ASSERT(solver.isFalse(b));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 0);
@@ -245,20 +291,19 @@ public:
 		bMin.push_back( WeightLiteral(a, 1) );
 		bMin.push_back( WeightLiteral(b, 1) );
 
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~d) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
 		solver.force(~e, 0); solver.propagate();
-		CPPUNIT_ASSERT_EQUAL(uint32(3), newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.backtrack();
 		solver.force(e, 0);
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 2 && solver.backtrackLevel() == 2);
 		CPPUNIT_ASSERT(solver.isFalse(c) && solver.isFalse(e));
 		solver.backtrack();
@@ -268,17 +313,17 @@ public:
 		WeightLitVec aMin;
 		aMin.push_back( WeightLiteral(x, 1) );
 		aMin.push_back( WeightLiteral(y, 1) );
-		newMin = MinimizeBuilder().addRule(aMin).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(aMin));
 		// disable backjumping
 		ctx.setProject(a.var(), true);
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~x) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(y) && solver.propagate());
-		CPPUNIT_ASSERT_EQUAL(uint32(2), newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(1);
 		solver.setBacktrackLevel(1);
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 1 && solver.isTrue(~x));
 		solver.backtrack();
 		CPPUNIT_ASSERT(solver.decisionLevel() == 0 && solver.isTrue(~x));
@@ -293,10 +338,9 @@ public:
 		bMin.push_back( WeightLiteral(a, 1) );
 		bMin.push_back( WeightLiteral(b, 1) );
 
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		Solver& solver = *ctx.master();
 				
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
@@ -304,9 +348,9 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~d) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
 		solver.force(~e, 0); solver.propagate();
-		CPPUNIT_ASSERT_EQUAL(uint32(3), newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(0);
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		solver.force(c, 0);
 		solver.force(d, 0);
@@ -324,16 +368,15 @@ public:
 		aMin.push_back( WeightLiteral(b, 1) );
 		bMin.push_back( WeightLiteral(c, 1) );
 		bMin.push_back( WeightLiteral(d, 1) );
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
-		CPPUNIT_ASSERT_EQUAL(1u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(0);
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(d) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~a));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~b));
@@ -346,22 +389,24 @@ public:
 		aMin.push_back( WeightLiteral(b, 1) );
 		bMin.push_back( WeightLiteral(c, 1) );
 		bMin.push_back( WeightLiteral(d, 1) );
-		newMin = MinimizeBuilder()
+		wsum_t bound[2] = {1,1};
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
 			.addRule(bMin)
-			.setOptimum(0, 1)
-			.setOptimum(1, 1)
-			.buildAndAttach(ctx, MinimizeMode_t::enumerate);
-		
+			, MinimizeMode_t::enumerate, bound, 2);
+
 		Solver& solver = *ctx.master();
-		
 		CPPUNIT_ASSERT_EQUAL(true, solver.propagate());
 		
-		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
+		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
 
-		CPPUNIT_ASSERT_EQUAL(1u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(0);
+		newMin->relaxBound();
+		newMin->integrateBound(solver);
+		CPPUNIT_ASSERT(solver.queueSize() == 0);
+
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(d) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT(solver.isTrue(~c));
@@ -374,17 +419,16 @@ public:
 		aMin.push_back( WeightLiteral(b, 1) );
 		aMin.push_back( WeightLiteral(c, 1) );
 		aMin.push_back( WeightLiteral(d, 2) );
-		newMin = MinimizeBuilder()
-			.addRule(aMin)
-			.buildAndAttach(ctx, MinimizeMode_t::optimize);
+		newMin = buildAndAttach(MinimizeBuilder()
+			.addRule(aMin));
 		Solver& solver = *ctx.master();
 		solver.assume(a) && solver.propagate();
 		solver.force(b,0)&& solver.propagate();
 		solver.assume(c) && solver.propagate();
 		solver.force(~d,0) && solver.propagate();
-		CPPUNIT_ASSERT_EQUAL(1u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(1u);
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.isFalse(d));
 		LitVec r1, r2;
 		solver.reason(~d, r1);
@@ -400,12 +444,14 @@ public:
 	void testSetModelMayBacktrackMultiLevels() {
 		WeightLitVec aMin;
 		aMin.push_back( WeightLiteral(a, 1) );
-		newMin = MinimizeBuilder().addRule(aMin).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(aMin));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
-		CPPUNIT_ASSERT_EQUAL(0u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
+		newMin->integrateBound(solver);
+		CPPUNIT_ASSERT(solver.decisionLevel() == 0);
 	}
 
 	void testPriorityBug() {
@@ -416,10 +462,9 @@ public:
 		bMin.push_back( WeightLiteral(d, 1) );
 		bMin.push_back( WeightLiteral(e, 1) );
 		bMin.push_back( WeightLiteral(f, 1) );
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		// disbale backjumping
 		ctx.setProject(a.var(), true);
 		ctx.setProject(e.var(), true);
@@ -427,9 +472,9 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(e) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(f) && solver.propagate());
-		CPPUNIT_ASSERT_EQUAL(2u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.backtrack();
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 2);
 		solver.backtrack();
 		CPPUNIT_ASSERT_EQUAL(true, solver.propagate());
@@ -446,9 +491,9 @@ public:
 		aMin.push_back( WeightLiteral(a, 1) );
 		aMin.push_back( WeightLiteral(b, 2) );
 		aMin.push_back( WeightLiteral(c, 1) );
-		newMin = MinimizeBuilder().addRule(aMin).setOptimum(0, 2).buildAndAttach(ctx);
+		wsum_t bound   = 2;
+		newMin         = buildAndAttach(MinimizeBuilder().addRule(aMin), MinimizeMode_t::optimize, &bound, 1);
 		Solver& solver = *ctx.master();
-	
 		solver.assume(a) && solver.propagate();
 		solver.setBacktrackLevel(1);
 		ctx.setProject(a.var(), true);
@@ -460,9 +505,9 @@ public:
 		solver.assume(y) && solver.propagate();
 		solver.assume(c) && solver.propagate();
 		solver.assume(e) && solver.propagate();
-		CPPUNIT_ASSERT(newMin->commitCurrent(solver) == 3);
+		newMin->commitUpperBound(solver);
 		solver.undoUntil(3);
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 1 && solver.isTrue(~b));
 		r.clear();
 		solver.reason(~b, r);
@@ -477,15 +522,14 @@ public:
 		bMin.push_back( WeightLiteral(d, 1) );
 		bMin.push_back( WeightLiteral(e, 2) );
 		bMin.push_back( WeightLiteral(f, 1) );
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		Solver& solver = *ctx.master();
 		solver.assume(a) && solver.propagate();
 		solver.assume(c) && solver.propagate();
 		solver.pushRootLevel(2);
-		MinimizeConstraint::SumVec opt(newMin->numRules());
+		SumVec opt(newMin->numRules());
 		opt[0] = 2;
 		opt[1] = 3;
 		setOptimum(solver, opt, false);
@@ -515,14 +559,15 @@ public:
 		aMin.push_back( WeightLiteral(d, 1) );
 		Solver& solver = *ctx.master();
 		data   = MinimizeBuilder().addRule(aMin).build(ctx);
-		newMin = createMin(solver, data);
+		newMin = createMin(ctx, solver, data);
+		newMin->integrateBound(solver);
 		solver.assume(a) && solver.propagate();
 		solver.assume(e) && solver.propagate();
 		solver.assume(b) && solver.propagate();
 		solver.assume(f) && solver.propagate();
 		solver.assume(c) && solver.propagate();
 		
-		SharedMinimizeData::SumVec opt(data->numRules());
+		SumVec opt(data->numRules());
 		opt[0] = 2;
 		CPPUNIT_ASSERT(setOptimum(solver, opt, true));
 		CPPUNIT_ASSERT(solver.decisionLevel() == 1 && solver.queueSize() == 3);
@@ -538,8 +583,9 @@ public:
 		aMin.push_back( WeightLiteral(d, 1) );
 		Solver& solver = *ctx.master();
 		data = MinimizeBuilder().addRule(aMin).build(ctx);
-		SharedMinimizeData::SumVec opt(data->numRules());
-		newMin = createMin(solver, data);
+		SumVec opt(data->numRules());
+		newMin = createMin(ctx, solver, data);
+		newMin->integrateBound(solver);
 		solver.assume(a) && solver.propagate();
 		solver.assume(e) && solver.propagate();
 		solver.assume(b) && solver.propagate();
@@ -556,12 +602,27 @@ public:
 		CPPUNIT_ASSERT(!setOptimum(solver, opt, true));
 		CPPUNIT_ASSERT(solver.hasConflict());
 		CPPUNIT_ASSERT(solver.clearAssumptions());
-		CPPUNIT_ASSERT(newMin->integrateNext(solver) == false && solver.hasConflict());
+		CPPUNIT_ASSERT(newMin->integrateBound(solver) == false && solver.hasConflict());
 		CPPUNIT_ASSERT(!solver.resolveConflict());
 		newMin->destroy(&solver, true);
 		newMin = 0;
 	}
-
+	void testIntegrateBug() {
+		WeightLitVec min;
+		min.push_back( WeightLiteral(a, 1) );
+		min.push_back( WeightLiteral(b, 1) );
+		min.push_back( WeightLiteral(c, 1) );
+		data   = MinimizeBuilder().addRule(min).build(ctx);
+		wsum_t bound = 2;
+		Solver& s = *ctx.master();
+		newMin = createMin(ctx, s, data, SolverStrategies::opt_dec);
+		data->setMode(MinimizeMode_t::optimize, &bound, 1);
+		s.pushRoot(s.tagLiteral());
+		s.pushRoot(a);
+		s.pushRoot(b);
+		s.pushRoot(c);
+		CPPUNIT_ASSERT(!newMin->integrateBound(s));
+	}
 	void testReasonBug() {
 		WeightLitVec aMin, bMin;
 		aMin.push_back( WeightLiteral(a, 1) );
@@ -570,17 +631,16 @@ public:
 		bMin.push_back( WeightLiteral(e, 1) );
 		bMin.push_back( WeightLiteral(f, 3) );
 		
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(aMin)
-			.addRule(bMin)
-			.buildAndAttach(ctx);
+			.addRule(bMin));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(d) && solver.propagate());
-		CPPUNIT_ASSERT_EQUAL(2u, newMin->commitCurrent(solver));
+		newMin->commitUpperBound(solver);
 		solver.backtrack();
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(e) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~f));
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
@@ -596,14 +656,14 @@ public:
 		aMin.push_back( WeightLiteral(a, 1) );
 		aMin.push_back( WeightLiteral(b, 1) );
 		aMin.push_back( WeightLiteral(c, 1) );
-		newMin = MinimizeBuilder().addRule(aMin).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(aMin));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~c) && solver.propagate());
 		
-		solver.undoUntil(newMin->commitCurrent(solver));
-		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateNext(solver));
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT_EQUAL(true, newMin->integrateBound(solver));
 		
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~b));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~c));
@@ -613,17 +673,16 @@ public:
 		WeightLitVec min1, min2;
 		min1.push_back( WeightLiteral(a, 1) );
 		min2.push_back( WeightLiteral(b, 1) );
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(min1)
-			.addRule(min2)
-			.buildAndAttach(ctx);
+			.addRule(min2));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.force(b, 0) && solver.propagate());
 			
-		CPPUNIT_ASSERT(newMin->commitCurrent(solver) == 0);
+		newMin->commitUpperBound(solver);
 		solver.backtrack();
-		CPPUNIT_ASSERT_EQUAL(false, newMin->integrateNext(solver));
+		CPPUNIT_ASSERT_EQUAL(false, newMin->integrateBound(solver));
 	}
 
 	void testMultiAssignment() {
@@ -636,12 +695,11 @@ public:
 
 		min2.push_back( WeightLiteral(f, 1) );
 		
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(min1)
-			.addRule(min2)
-			.buildAndAttach(ctx);
+			.addRule(min2));
 		Solver& solver = *ctx.master();
-		MinimizeConstraint::SumVec opt(newMin->numRules());
+		SumVec opt(newMin->numRules());
 		opt[0] = 3;
 		opt[1] = 0;
 		CPPUNIT_ASSERT_EQUAL(true, setOptimum(solver, opt, false));
@@ -663,10 +721,9 @@ public:
 		min2.push_back( WeightLiteral(e, 1) );
 		min2.push_back( WeightLiteral(f, 1) );
 		
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(min1)
-			.addRule(min2)
-			.buildAndAttach(ctx);
+			.addRule(min2));
 		ctx.setProject(x.var(), true);
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
@@ -678,18 +735,18 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, solver.force(e,0) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.force(~f,0) && solver.propagate());
 		
-		solver.undoUntil(newMin->commitCurrent(solver));
-		CPPUNIT_ASSERT(3 == solver.decisionLevel());
+		newMin->commitUpperBound(solver);
+		solver.undoUntil(3);
 		solver.setBacktrackLevel(3);
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT_EQUAL(true, solver.force(f,0) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~d));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~e));
 		
-		uint32 dl = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(2 == dl);
+		newMin->commitUpperBound(solver);
+		solver.undoUntil(2);
 		solver.backtrack();
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~x));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~f));
 		CPPUNIT_ASSERT_EQUAL(true, solver.isTrue(~d));
@@ -706,24 +763,22 @@ public:
 		min2.push_back( WeightLiteral(b, 1) );
 		min2.push_back( WeightLiteral(c, 1) );
 		
-		newMin = MinimizeBuilder()
+		newMin = buildAndAttach(MinimizeBuilder()
 			.addRule(min1)
-			.addRule(min2)
-			.buildAndAttach(ctx);
+			.addRule(min2));
 		Solver& solver = *ctx.master();
 		
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
 
-		solver.undoUntil(newMin->commitCurrent(solver));
-		CPPUNIT_ASSERT(newMin->integrateNext(solver) && 1 == solver.decisionLevel());
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT(newMin->integrateBound(solver) && 1 == solver.decisionLevel());
 		CPPUNIT_ASSERT(solver.propagate());
 		
-		uint32 dl = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(dl == 0);
-		solver.undoUntil(dl);
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		newMin->commitUpperBound(solver);
+		solver.undoUntil(0);
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.isTrue(~a));
 		CPPUNIT_ASSERT(solver.value(b.var()) == value_free);
 	}
@@ -737,11 +792,11 @@ public:
 		data = MinimizeBuilder()
 			.addRule(aMin)
 			.addRule(bMin)
-			.setOptimum(0, 1)
-			.setOptimum(1, 1)
 			.build(ctx);
-		data->setMode(MinimizeMode_t::optimize, 1);
-		newMin = data->attach(*ctx.master());
+		wsum_t bound[2] = {wsum_t(1),wsum_t(1)};
+		data->setMode(MinimizeMode_t::optimize, bound, 2);
+		newMin = createMin(ctx, *ctx.master(), data, SolverStrategies::opt_hier);
+		newMin->integrateBound(*ctx.master());
 		CPPUNIT_ASSERT(ctx.master()->value(a.var()) == value_free);
 		ctx.master()->assume(a) && ctx.master()->propagate();
 		CPPUNIT_ASSERT(ctx.master()->isFalse(b));
@@ -760,9 +815,8 @@ public:
 			.addRule(aMin)
 			.addRule(bMin)
 			.build(ctx);
-		data->setMode(MinimizeMode_t::optimize, 0);
-		newMin = data->attach(*ctx.master());
-		SharedMinimizeData::SumVec opt;
+		newMin = static_cast<DefaultMinimize*>(data->attach(*ctx.master()));
+		SumVec opt;
 		opt.push_back(1);
 		opt.push_back(1);
 		setOptimum(*ctx.master(), opt, true);
@@ -777,12 +831,12 @@ public:
 		WeightLitVec min1;
 		min1.push_back( WeightLiteral(a, 1) );
 		min1.push_back( WeightLiteral(b, 0) );
-		newMin = MinimizeBuilder().addRule(min1).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(min1));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.force(~c,0) && solver.propagate());
-		solver.undoUntil(newMin->commitCurrent(solver));
-		newMin->integrateNext(solver);
+		newMin->commitUpperBound(solver);
+		newMin->integrateBound(solver);
 		CPPUNIT_ASSERT(0u == solver.decisionLevel() && solver.isFalse(a));
 	}
 
@@ -790,21 +844,18 @@ public:
 		WeightLitVec min1;
 		min1.push_back( WeightLiteral(a, 1) );
 		min1.push_back( WeightLiteral(b, 1) );
-		newMin = MinimizeBuilder().addRule(min1, -2).buildAndAttach(ctx);
+		newMin = buildAndAttach(MinimizeBuilder().addRule(min1, -2));
 		Solver& solver = *ctx.master();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
-		uint32 DL = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(DL < solver.decisionLevel());
-		solver.undoUntil(DL);
-		CPPUNIT_ASSERT(0 == newMin->shared()->optimum()->opt[0]);
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT(0 == newMin->shared()->optimum(0));
 		
 		solver.clearAssumptions();
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~a) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(~b) && solver.propagate());
-		DL = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(DL == UINT32_MAX);
-		CPPUNIT_ASSERT(-2 == newMin->shared()->optimum()->opt[0]);
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT(-2 == newMin->shared()->optimum(0));
 	}
 
 	void testAdjustFact() {
@@ -815,30 +866,33 @@ public:
 		min1.push_back( WeightLiteral(d, 1) );
 		data = MinimizeBuilder().addRule(min1, -2).build(ctx);
 		Solver& solver = *ctx.master();
-		solver.addUnary(a, Constraint_t::static_constraint);
+		ctx.addUnary(a);
 		solver.propagate();
-		newMin = createMin(solver, data);
+		newMin = createMin(ctx, solver, data);
+		newMin->integrateBound(solver);
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(b) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(c) && solver.propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver.assume(d) && solver.propagate());
-		solver.undoUntil(newMin->commitCurrent(solver));
-		newMin->integrateNext(solver);
+		newMin->commitUpperBound(solver);
+		newMin->integrateBound(solver);
 		CPPUNIT_ASSERT(2 == solver.decisionLevel());
 	}
 
 	void testAssumption() {
+		SharedContext ctx;
+		a = posLit(ctx.addVar(Var_t::atom_var));
+		b = posLit(ctx.addVar(Var_t::atom_var));
+		c = posLit(ctx.addVar(Var_t::atom_var));
+		ctx.startAddConstraints();
 		WeightLitVec min1;
 		min1.push_back( WeightLiteral(a, 1) );
 		min1.push_back( WeightLiteral(b, 1) );
 		min1.push_back( WeightLiteral(c, 1) );
 		Solver& solver = *ctx.master();
-		Literal minAssume = d;
-		data   = MinimizeBuilder().addRule(min1).build(ctx, minAssume);
-		newMin = createMin(solver, data);
-		solver.assume(minAssume);
-		solver.pushRootLevel(1);
-		solver.propagate();
-		SharedMinimizeData::SumVec opt(1, 0);
+		data   = MinimizeBuilder().addRule(min1).build(ctx);
+		newMin = createMin(ctx, solver, data, SolverStrategies::opt_inc);
+		Literal minAssume = posLit(solver.pushTagVar(true));
+		SumVec opt(1, 0);
 		setOptimum(solver, opt, false);
 		CPPUNIT_ASSERT(solver.isFalse(a) && solver.reason(~a).constraint() == newMin);
 		CPPUNIT_ASSERT(solver.isFalse(b) && solver.reason(~b).constraint() == newMin);
@@ -849,7 +903,6 @@ public:
 		CPPUNIT_ASSERT(r.size() == 1 && r[0] == minAssume);
 
 		solver.clearAssumptions();
-		newMin->integrateNext(solver);
 		CPPUNIT_ASSERT(solver.numAssignedVars() == 0);
 	}
 
@@ -867,19 +920,19 @@ public:
 		builder.addRule(min);
 		Solver& solver = *ctx.master();
 		data   = builder.build(ctx);
-		newMin = createMin(solver, data);
+		newMin = createMin(ctx, solver, data, SolverStrategies::opt_hier);
+		newMin->integrateBound(solver);
 		solver.assume(a); solver.propagate();
 		solver.assume(b); solver.propagate();
 		solver.assume(c); solver.propagate();
 		solver.assume(d); solver.propagate();
 		solver.assume(e); solver.propagate();
 		solver.assume(f); solver.propagate();
-		data->setMode(MinimizeMode_t::optimize, true);
-		CPPUNIT_ASSERT(newMin->commitCurrent(solver) == 2);
-		solver.undoUntil(2);
-		newMin->integrateNext(solver);
+		newMin->commitUpperBound(solver);
+		solver.undoUntil(solver.level(b.var()));
+		newMin->integrateBound(solver);
 		CPPUNIT_ASSERT(solver.isFalse(c));
-		CPPUNIT_ASSERT(solver.numAssignedVars() == 3);
+		CPPUNIT_ASSERT(solver.numAssignedVars() == 4);
 	}
 
 	void testHierarchical() {
@@ -895,20 +948,16 @@ public:
 		ctx.addBinary(a, ~b);
 		Solver& solver = *ctx.master();
 		data   = builder.build(ctx);
-		newMin = createMin(solver, data);
-
-		data->setMode(MinimizeMode_t::optimize, 1);
-
-		ctx.endInit();
+		newMin = createMin(ctx, solver, data, SolverStrategies::opt_hier);
+		newMin->integrateBound(solver);
 		solver.assume(a);
 		solver.propagate();
-		uint32 x = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(x == 0);
+		newMin->commitUpperBound(solver);
 		solver.backtrack();
-		CPPUNIT_ASSERT(newMin->integrateNext(solver));
+		CPPUNIT_ASSERT(newMin->integrateBound(solver));
 		CPPUNIT_ASSERT(solver.propagate() == true);
-		x = newMin->commitCurrent(solver);
-		CPPUNIT_ASSERT(x == UINT32_MAX);
+		newMin->commitUpperBound(solver);
+		CPPUNIT_ASSERT(!newMin->integrateBound(solver));
 	}
 	void testInconsistent() {
 		MinimizeBuilder builder;
@@ -917,11 +966,10 @@ public:
 		min.push_back( WeightLiteral(b, 1) );
 		min.push_back( WeightLiteral(c, 1) );
 		builder.addRule(min);
-		builder.setOptimum(0, 1);
+		wsum_t bound = 1;
 		ctx.addUnary(a);
 		ctx.addUnary(b);
-		CPPUNIT_ASSERT(builder.build(ctx) == 0);
-		CPPUNIT_ASSERT(builder.buildAndAttach(ctx) == 0);
+		CPPUNIT_ASSERT(buildAndAttach(builder, MinimizeMode_t::optimize, &bound, 1) == 0);
 	}
 private:
 	uint32 countMinLits() const {
@@ -931,21 +979,214 @@ private:
 		return lits;
 	}
 	
-	bool setOptimum(Solver& s, SharedMinimizeData::SumVec& vec, bool less) {
+	bool setOptimum(Solver& s, SumVec& vec, bool less) {
 		SharedMinimizeData* data = const_cast<SharedMinimizeData*>(newMin->shared());
 		if (!less) {
 			vec.back() += 1;
 		}
 		data->setOptimum(&vec[0]);
-		return newMin->integrateNext(s);
+		while (!newMin->integrateBound(s)) {
+			if (!s.resolveConflict()) { return false; }
+		}
+		return true;
 	}
-	MinimizeConstraint* createMin(Solver& s, SharedMinimizeData* data) {
-		return data->attach(s);
+	DefaultMinimize* createMin(SharedContext& ctx, Solver& s, SharedMinimizeData* data, SolverStrategies::OptStrategy str = SolverStrategies::opt_def) {
+		ctx.endInit();
+		return static_cast<DefaultMinimize*>(data->attach(s, str));
 	}	
+	DefaultMinimize* buildAndAttach(MinimizeBuilder& x, MinimizeMode m = MinimizeMode_t::optimize, const wsum_t* b = 0, uint32 bs = 0) {
+		DefaultMinimize* con = 0;
+		if ( (data = x.build(ctx)) != 0 && data->setMode(m, b, bs) ) {
+			con = createMin(ctx, *ctx.master(), data);
+			con->integrateBound(*ctx.master());
+		}
+		return con;
+	}
 	SharedContext ctx;
-	MinimizeConstraint* newMin;
-	SharedMinimizeData* data;
+	DefaultMinimize* newMin;
+	SharedMinimizeData*   data;
 	Literal a, b, c, d, e, f, x, y;
 };
-CPPUNIT_TEST_SUITE_REGISTRATION(MinimizeTest);
+
+class UncoreMinimizeTest : public CppUnit::TestFixture {
+	CPPUNIT_TEST_SUITE(UncoreMinimizeTest);
+	CPPUNIT_TEST(testEmpty);
+	CPPUNIT_TEST(testEnumerate);
+	CPPUNIT_TEST(testOptimize);
+	CPPUNIT_TEST(testOptimizeGP);
+	CPPUNIT_TEST(testMtBug1);
+	CPPUNIT_TEST(testNegativeLower);
+	CPPUNIT_TEST_SUITE_END(); 
+public:
+	void setUp() {
+		data = 0;
+		min  = 0;
+	}
+	void tearDown() {
+		if (min) {
+			min->destroy(0,false);
+		}
+		if (data) {
+			data->release();
+		}
+	}
+	void testEmpty() {
+		MinimizeBuilder b;
+		ctx.startAddConstraints();
+		data = b.build(ctx);
+		ctx.endInit();
+		min = data->attach(*ctx.master(), SolverStrategies::opt_unsat_pre);
+		CPPUNIT_ASSERT(min->integrate(*ctx.master()));
+	}
+	void testEnumerate() {
+		Var a = ctx.addVar(Var_t::atom_var);
+		Var b = ctx.addVar(Var_t::atom_var);
+		Var c = ctx.addVar(Var_t::atom_var);
+		ctx.startAddConstraints();
+		WeightLitVec lits;
+		MinimizeBuilder builder;
+		lits.push_back(WeightLiteral(posLit(a), 1));
+		lits.push_back(WeightLiteral(posLit(b), 1));
+		lits.push_back(WeightLiteral(posLit(c), 1));
+		builder.addRule(lits);
+		data = builder.build(ctx);
+		ctx.endInit();
+		wsum_t bound = 1;
+		data->setMode(MinimizeMode_t::enumerate, &bound, 1);
+		min = data->attach(*ctx.master(), SolverStrategies::opt_unsat_pre);
+		CPPUNIT_ASSERT(min->integrate(*ctx.master()));
+		
+		ctx.master()->assume(posLit(a));
+		ctx.master()->propagate();
+		CPPUNIT_ASSERT(ctx.master()->isFalse(posLit(b)));
+		CPPUNIT_ASSERT(ctx.master()->isFalse(posLit(c)));
+	}
+	void testOptimize() {
+		Var a = ctx.addVar(Var_t::atom_var);
+		Var b = ctx.addVar(Var_t::atom_var);
+		Solver& s = ctx.startAddConstraints();
+		WeightLitVec lits;
+		MinimizeBuilder builder;
+		lits.push_back(WeightLiteral(posLit(a), 1));
+		lits.push_back(WeightLiteral(posLit(b), 1));
+		builder.addRule(lits);
+		data = builder.build(ctx);
+		ctx.endInit();
+		data->setMode(MinimizeMode_t::optimize);
+		min = data->attach(s, SolverStrategies::opt_unsat_pre);
+		BasicSolve solve(s);
+		LitVec gp;
+		while (min->integrate(s) || min->handleUnsat(s, true, gp)) {
+			ValueRep v = solve.solve();
+			CPPUNIT_ASSERT(v == value_true);
+			min->handleModel(s);
+		}
+		CPPUNIT_ASSERT(gp.empty());
+		CPPUNIT_ASSERT(s.isFalse(lits[0].first) && s.isFalse(lits[1].first));
+	}
+	void testOptimizeGP() {
+		Var a = ctx.addVar(Var_t::atom_var);
+		Var b = ctx.addVar(Var_t::atom_var);
+		Solver& s = ctx.startAddConstraints();
+		WeightLitVec lits;
+		MinimizeBuilder builder;
+		lits.push_back(WeightLiteral(posLit(a), 1));
+		lits.push_back(WeightLiteral(posLit(b), 1));
+		builder.addRule(lits);
+		data = builder.build(ctx);
+		ctx.endInit();
+		data->setMode(MinimizeMode_t::optimize);
+		min = data->attach(s, SolverStrategies::opt_unsat_pre);
+		BasicSolve solve(s);
+		LitVec gp; gp.push_back(posLit(a));
+		solve.assume(gp);
+		gp.clear();
+		while (min->integrate(s) || min->handleUnsat(s, true, gp)) {
+			ValueRep v = solve.solve();
+			CPPUNIT_ASSERT(v == value_true);
+			min->handleModel(s);
+		}
+		CPPUNIT_ASSERT(s.rootLevel() == 1 && s.decision(1) == posLit(a));
+		CPPUNIT_ASSERT(s.isTrue(lits[0].first) && s.isFalse(lits[1].first));
+	}
+
+	void testMtBug1() {
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(posLit(ctx.addVar(Var_t::atom_var)), 1));
+		lits.push_back(WeightLiteral(posLit(ctx.addVar(Var_t::atom_var)), 1));
+		Solver& s1 = ctx.startAddConstraints();
+		data       = MinimizeBuilder().addRule(lits).build(ctx);
+		ctx.setConcurrency(2);
+		ctx.endInit(true);
+		Solver& s2 = *ctx.solver(1);
+		data->setMode(MinimizeMode_t::enumOpt);
+		MinimizeConstraint* m1 = data->attach(s1, SolverStrategies::opt_unsat_pre);
+		MinimizeConstraint* m2 = data->attach(s2, SolverStrategies::opt_unsat_pre);
+		s1.setEnumerationConstraint(m1);
+		s2.setEnumerationConstraint(m2);
+		BasicSolve solve(s1);
+		LitVec gp;
+		while (m1->integrate(s1) || m1->handleUnsat(s1, true, gp)) {
+			ValueRep v = solve.solve();
+			CPPUNIT_ASSERT(v == value_true);
+			m1->handleModel(s1);
+		}
+		data->markOptimal();
+		m1->relax(s1, true);
+		m2->relax(s2, true);
+		solve.reset(s2, s2.searchConfig());
+		uint32 numModels = 0;
+		s2.setPref(1, ValueSet::user_value, value_true);
+		for (;;) {
+			ValueRep v = m2->integrate(s2) ? solve.solve() : value_false;
+			if (v == value_true) {
+				++numModels;
+				CPPUNIT_ASSERT(s2.isFalse(lits[0].first));
+				CPPUNIT_ASSERT(s2.isFalse(lits[1].first));
+				if (!s2.backtrack()) { break; }
+			}
+			else if (v == value_false) {
+				break;
+			}
+		}
+		CPPUNIT_ASSERT(numModels == 1);
+	}
+	void testNegativeLower() {
+		Literal a = posLit(ctx.addVar(Var_t::atom_var));
+		Literal b = posLit(ctx.addVar(Var_t::atom_var));
+		Solver& s = ctx.startAddConstraints();
+		ctx.addBinary(a, b);
+		WeightLitVec aMin, bMin;
+		aMin.push_back( WeightLiteral(a, 1) );
+		aMin.push_back( WeightLiteral(b, 1) );
+		bMin.push_back( WeightLiteral(~a, 1) );
+		data = MinimizeBuilder()
+			.addRule(aMin)
+			.addRule(bMin)
+			.build(ctx);
+		ctx.endInit();
+		data->setMode(MinimizeMode_t::optimize);
+		min = data->attach(s, SolverStrategies::opt_unsat_pre);
+		
+		LitVec ignore;
+		while (!min->integrate(*ctx.master()) && min->handleUnsat(s, true, ignore)) { ; }
+		CPPUNIT_ASSERT(s.assume(~a) && s.propagate());
+		CPPUNIT_ASSERT(s.assume(b) && s.propagate());
+		CPPUNIT_ASSERT(min->handleModel(s));
+		s.undoUntil(0);
+		while (!min->integrate(*ctx.master()) && min->handleUnsat(s, true, ignore)) { ; }
+		CPPUNIT_ASSERT(s.assume(a) && s.propagate());
+		CPPUNIT_ASSERT(s.isFalse(b));
+		min->handleModel(s);
+		s.undoUntil(0);
+		while (!min->integrate(*ctx.master()) && min->handleUnsat(s, true, ignore)) { ; }
+		CPPUNIT_ASSERT(s.hasConflict());
+	}
+private:
+	SharedContext       ctx;
+	SharedMinimizeData* data;
+	MinimizeConstraint* min;
+};
+CPPUNIT_TEST_SUITE_REGISTRATION(DefaultMinimizeTest);
+CPPUNIT_TEST_SUITE_REGISTRATION(UncoreMinimizeTest);
 } } 
