@@ -24,6 +24,7 @@
 #include <memory>
 
 namespace Clasp { namespace Test {
+	using namespace Clasp::Asp;
 	class UnfoundedCheckTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST_SUITE(UnfoundedCheckTest);
 	CPPUNIT_TEST(testAllUncoloredNoUnfounded);
@@ -41,12 +42,22 @@ namespace Clasp { namespace Test {
 	CPPUNIT_TEST(testChoiceCardInterplay);
 	CPPUNIT_TEST(testCardInterplayOnBT);
 
+	CPPUNIT_TEST(testInitNoSource);
+
 	CPPUNIT_TEST(testIncrementalUfs);
+	CPPUNIT_TEST(testInitialStopConflict);
+	CPPUNIT_TEST(testIncrementalLearnFact);
+
+	CPPUNIT_TEST(testApproxUfs);
 	CPPUNIT_TEST_SUITE_END(); 
 public:
+	class WrapDefaultUnfoundedCheck : public DefaultUnfoundedCheck {
+	public:
+		bool propagate(Solver& s) { return DefaultUnfoundedCheck::propagateFixpoint(s, 0); }
+	};
 	UnfoundedCheckTest() : ufs(0) { }
 	void setUp() { 
-		ufs   = new DefaultUnfoundedCheck();
+		ufs   = new WrapDefaultUnfoundedCheck();
 	}
 	void tearDown() {
 		ufs   = 0;
@@ -61,7 +72,7 @@ public:
 	
 	void testAlternativeSourceNotUnfounded() {
 		setupSimpleProgram();
-		solver().assume( ctx.symTab()[6].lit );
+		solver().assume( ctx.symbolTable()[6].lit );
 		solver().propagateUntil(ufs.get());
 		uint32 old = solver().numAssignedVars();
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
@@ -70,21 +81,21 @@ public:
 	
 	void testOnlyOneSourceUnfoundedIfMinus() {
 		setupSimpleProgram();
-		solver().assume( ctx.symTab()[6].lit );
-		solver().assume( ctx.symTab()[5].lit );
+		solver().assume( ctx.symbolTable()[6].lit );
+		solver().assume( ctx.symbolTable()[5].lit );
 		solver().propagateUntil(ufs.get());
 		uint32 old = solver().numAssignedVars();
 		uint32 oldC = ctx.numConstraints();
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
 		CPPUNIT_ASSERT(old < solver().numAssignedVars());
-		CPPUNIT_ASSERT(solver().isFalse(ctx.symTab()[4].lit));
-		CPPUNIT_ASSERT(solver().isFalse(ctx.symTab()[1].lit));
-		CPPUNIT_ASSERT(!solver().isFalse(ctx.symTab()[2].lit));
+		CPPUNIT_ASSERT(solver().isFalse(ctx.symbolTable()[4].lit));
+		CPPUNIT_ASSERT(solver().isFalse(ctx.symbolTable()[1].lit));
+		CPPUNIT_ASSERT(!solver().isFalse(ctx.symbolTable()[2].lit));
 		CPPUNIT_ASSERT(oldC+1 == ctx.numConstraints() + ctx.numLearntShort());
 	}
 	
 	void testWithSimpleCardinalityConstraint() {
-		builder.startProgram(ctx)
+		builder.start(ctx)
 			.setAtomName(1, "a").setAtomName(2, "b")
 			.startRule(CHOICERULE).addHead(2).endRule()
 			.startRule(CONSTRAINTRULE, 1).addHead(1).addToBody(1, true).addToBody(2,true).endRule()
@@ -97,18 +108,18 @@ public:
 		CPPUNIT_ASSERT_EQUAL(0u, solver().numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()) );
 		CPPUNIT_ASSERT_EQUAL(0u, solver().numAssignedVars());
-		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symTab()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symbolTable()[2].lit));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(1u, solver().numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()) );
 		CPPUNIT_ASSERT_EQUAL(2u, solver().numAssignedVars());
 		LitVec r;
-		solver().reason(~ctx.symTab()[1].lit, r);
+		solver().reason(~ctx.symbolTable()[1].lit, r);
 		CPPUNIT_ASSERT(1 == r.size());
-		CPPUNIT_ASSERT(r[0] == ~ctx.symTab()[2].lit);
+		CPPUNIT_ASSERT(r[0] == ~ctx.symbolTable()[2].lit);
 	}
 	void testWithSimpleWeightConstraint() {
-		builder.startProgram(ctx)
+		builder.start(ctx)
 			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
 			.startRule(CHOICERULE).addHead(2).addHead(3).endRule()
 			.startRule(WEIGHTRULE, 2).addHead(1).addToBody(1, true, 2).addToBody(2,true, 2).addToBody(3, true, 1).endRule()
@@ -121,13 +132,13 @@ public:
 		CPPUNIT_ASSERT_EQUAL(0u, solver().numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()) );
 		CPPUNIT_ASSERT_EQUAL(0u, solver().numAssignedVars());
-		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symTab()[3].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symbolTable()[3].lit));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(1u, solver().numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()) );
 		CPPUNIT_ASSERT_EQUAL(1u, solver().numAssignedVars());
 
-		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symTab()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symbolTable()[2].lit));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(2u, solver().numAssignedVars());
 
@@ -135,25 +146,25 @@ public:
 		CPPUNIT_ASSERT_EQUAL(3u, solver().numAssignedVars());
 		
 		LitVec r;
-		solver().reason(~ctx.symTab()[1].lit, r);
+		solver().reason(~ctx.symbolTable()[1].lit, r);
 		CPPUNIT_ASSERT(2 == r.size());
-		CPPUNIT_ASSERT(r[0] == ~ctx.symTab()[2].lit);
-		CPPUNIT_ASSERT(r[1] == ~ctx.symTab()[3].lit);
+		CPPUNIT_ASSERT(r[0] == ~ctx.symbolTable()[2].lit);
+		CPPUNIT_ASSERT(r[1] == ~ctx.symbolTable()[3].lit);
 
 		solver().undoUntil(0);
-		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symTab()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().assume(~ctx.symbolTable()[2].lit));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(1u, solver().numAssignedVars());
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()) );
 		CPPUNIT_ASSERT_EQUAL(2u, solver().numAssignedVars());
 		r.clear();
-		solver().reason(~ctx.symTab()[1].lit, r);
+		solver().reason(~ctx.symbolTable()[1].lit, r);
 		CPPUNIT_ASSERT(1 == r.size());
-		CPPUNIT_ASSERT(r[0] == ~ctx.symTab()[2].lit);
+		CPPUNIT_ASSERT(r[0] == ~ctx.symbolTable()[2].lit);
 	}
 	
 	void testNtExtendedBug() {
-		builder.startProgram(ctx)
+		builder.start(ctx)
 			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c").setAtomName(4, "t").setAtomName(5, "x")
 			.startRule(CHOICERULE).addHead(1).addHead(2).addHead(3).endRule() // {a,b,c}.
 			.startRule(CONSTRAINTRULE, 2).addHead(4).addToBody(2, true).addToBody(4, true).addToBody(5,true).endRule()
@@ -167,7 +178,7 @@ public:
 		solver().assume(Literal(6, false));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
-		solver().assume(~ctx.symTab()[1].lit);
+		solver().assume(~ctx.symbolTable()[1].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(false, ufs->propagate(solver()));  // {x, t} are unfounded!
 		
@@ -178,17 +189,17 @@ public:
 		solver().assume(Literal(6, true));
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		// F: a
-		solver().assume(~ctx.symTab()[1].lit);
+		solver().assume(~ctx.symbolTable()[1].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		// x is false because both of its bodies are false
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[5].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[5].lit));
 	
 		// t is now unfounded but its defining body is not
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[4].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[4].lit));
 		LitVec r;
-		solver().reason(~ctx.symTab()[4].lit, r);
-		CPPUNIT_ASSERT(r.size() == 1 && r[0] == ~ctx.symTab()[5].lit);
+		solver().reason(~ctx.symbolTable()[4].lit, r);
+		CPPUNIT_ASSERT(r.size() == 1 && r[0] == ~ctx.symbolTable()[5].lit);
 	}
 	
 	void testNtExtendedFalse() {
@@ -198,7 +209,7 @@ public:
 		// r :- s, z.
 		// x :- not z.
 		// y :- not z.
-		builder.startProgram(ctx, ProgramBuilder::EqOptions().noEq())
+		builder.start(ctx, LogicProgram::AspOptions().noEq())
 			.setAtomName(1, "x").setAtomName(2, "y").setAtomName(3, "z").setAtomName(4, "r").setAtomName(5, "s")
 			.startRule(CHOICERULE).addHead(3).endRule() // {z}.
 			.startRule().addHead(1).addToBody(3,false).endRule()
@@ -210,67 +221,67 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
 		attachUfs();
 		
-		solver().assume(ctx.symTab()[3].lit);
+		solver().assume(ctx.symbolTable()[3].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
 		CPPUNIT_ASSERT(solver().numVars() == solver().numAssignedVars());
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[4].lit));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[5].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[4].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[5].lit));
 
 		solver().backtrack();
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
 		CPPUNIT_ASSERT(solver().numVars() == solver().numAssignedVars());
-		CPPUNIT_ASSERT_EQUAL(true, solver().isTrue(ctx.symTab()[4].lit));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[5].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isTrue(ctx.symbolTable()[4].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[5].lit));
 	}
 
 	void testDependentExtReason() {
-		// {z, q}.
+		// {z, q, r}.
 		// x :- not q.
 		// x :- 2 {x, y, z}.
 		// x :- y, z.
-		// y :- x.
-		builder.startProgram(ctx, ProgramBuilder::EqOptions().noEq())
-			.setAtomName(1, "x").setAtomName(2, "y").setAtomName(3, "z").setAtomName(4, "q")
-			.startRule(CHOICERULE).addHead(3).addHead(4).endRule()
+		// y :- x, r.
+		builder.start(ctx, LogicProgram::AspOptions().noEq())
+			.setAtomName(1, "x").setAtomName(2, "y").setAtomName(3, "z").setAtomName(4, "q").setAtomName(5, "r")
+			.startRule(CHOICERULE).addHead(3).addHead(4).addHead(5).endRule()
 			.startRule().addHead(1).addToBody(4,false).endRule()
 			.startRule(CONSTRAINTRULE, 2).addHead(1).addToBody(1, true).addToBody(2, true).addToBody(3, true).endRule()
 			.startRule().addHead(1).addToBody(2,true).addToBody(3, true).endRule()
-			.startRule().addHead(2).addToBody(1,true).endRule()
+			.startRule().addHead(2).addToBody(1,true).addToBody(5, true).endRule()
 		;
 		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
 		attachUfs();
 		
 		// assume ~B1, where B1 = 2 {x, y, z}
-		const SharedDependencyGraph::AtomNode& a = *ufs->graph()->getAtomByLit(ctx.symTab()[1].lit).node;
-		Literal x = ufs->graph()->getBodyNode(a.bodies()[1]).lit;
+		const SharedDependencyGraph::AtomNode& a = ufs->graph()->getAtom(builder.getAtom(1)->id());
+		Literal x = ufs->graph()->getBody(a.body(1)).lit;
 
 		solver().assume(~x);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()) && ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(value_free, solver().value(ctx.symTab()[1].lit.var()));
-		CPPUNIT_ASSERT_EQUAL(value_free, solver().value(ctx.symTab()[2].lit.var()));
-		// empty body + B1
-		CPPUNIT_ASSERT_EQUAL(2u, solver().numAssignedVars());
+		CPPUNIT_ASSERT_EQUAL(value_free, solver().value(ctx.symbolTable()[1].lit.var()));
+		CPPUNIT_ASSERT_EQUAL(value_free, solver().value(ctx.symbolTable()[2].lit.var()));
+		// B1
+		CPPUNIT_ASSERT_EQUAL(1u, solver().numAssignedVars());
 		
 		// assume q
-		solver().assume(ctx.symTab()[4].lit);
+		solver().assume(ctx.symbolTable()[4].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
-		// empty body + B1 + q + {not q}
-		CPPUNIT_ASSERT_EQUAL(4u, solver().numAssignedVars());
+		// B1 + q (hence {not q})
+		CPPUNIT_ASSERT_EQUAL(2u, solver().numAssignedVars());
 
 		// U = {x, y}.
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[1].lit));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[2].lit));
-		Literal extBody = ufs->graph()->getBodyNode(a.bodies()[0]).lit;
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[1].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[2].lit));
+		Literal extBody = ufs->graph()->getBody(a.body(0)).lit;
 		LitVec r;
-		solver().reason(~ctx.symTab()[1].lit, r);
+		solver().reason(~ctx.symbolTable()[1].lit, r);
 		CPPUNIT_ASSERT(r.size() == 1u && r[0] == ~extBody);
 	}
 
 	void testEqBodyDiffType() { 
-		builder.startProgram(ctx)
+		builder.start(ctx)
 			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c").setAtomName(4, "x").setAtomName(5,"y")
 			.startRule(CHOICERULE).addHead(1).addHead(4).addHead(5).endRule()
 			.startRule(CHOICERULE).addHead(2).addToBody(1,true).endRule()
@@ -281,15 +292,15 @@ public:
 		CPPUNIT_ASSERT(builder.stats.sccs == 1);
 		attachUfs();
 		
-		solver().assume(~ctx.symTab()[1].lit);
+		solver().assume(~ctx.symbolTable()[1].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[2].lit));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[3].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[3].lit));
 	}
 
 	void testChoiceCardInterplay() {  
-		builder.startProgram(ctx, ProgramBuilder::EqOptions().noEq())
+		builder.start(ctx, LogicProgram::AspOptions().noEq())
 			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c").setAtomName(4, "x")
 			.startRule(CHOICERULE).addHead(4).endRule() // {x}.
 			.startRule(CHOICERULE).addHead(1).addToBody(4, true).endRule()  // {a} :- x.
@@ -300,15 +311,15 @@ public:
 		CPPUNIT_ASSERT(builder.stats.sccs == 1);
 		attachUfs();
 		
-		solver().assume(~ctx.symTab()[1].lit);
+		solver().assume(~ctx.symbolTable()[1].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
 		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[2].lit));
-		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symTab()[3].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(true, solver().isFalse(ctx.symbolTable()[3].lit));
 	}
 
 	void testCardInterplayOnBT() {  
-		builder.startProgram(ctx, ProgramBuilder::EqOptions().noEq())
+		builder.start(ctx, LogicProgram::AspOptions().noEq())
 			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c").setAtomName(4, "d").setAtomName(5,"z")
 			.startRule(CHOICERULE).addHead(1).addHead(5).endRule()                                  // {a,z}.
 			.startRule(CONSTRAINTRULE,1).addHead(2).addToBody(1, true).addToBody(3, true).endRule() // b :- 1{a,c}
@@ -320,40 +331,55 @@ public:
 		CPPUNIT_ASSERT(builder.stats.sccs == 1);
 		attachUfs();
 		
-		solver().assume(~ctx.symTab()[1].lit);
+		solver().assume(~ctx.symbolTable()[1].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()) && ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symTab()[2].lit));
-		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symTab()[3].lit));
+		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symbolTable()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symbolTable()[3].lit));
 		solver().undoUntil(0);
 		
-		solver().assume(~ctx.symTab()[5].lit);
+		solver().assume(~ctx.symbolTable()[5].lit);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()) && ufs->propagate(solver()));
-		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symTab()[2].lit));
-		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symTab()[3].lit));
+		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symbolTable()[2].lit));
+		CPPUNIT_ASSERT_EQUAL(false, solver().isFalse(ctx.symbolTable()[3].lit));
 	}
 	
+	void testInitNoSource() {
+		builder.start(ctx, LogicProgram::AspOptions().noEq())
+			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "x")
+			.startRule(CHOICERULE).addHead(3).endRule()                                  // {x}.
+			.startRule(CHOICERULE).addHead(1).addHead(2).addToBody(3, true).endRule()    // {a,b} :- x.
+			.startRule(BASICRULE).addHead(1).addToBody(2, true).endRule()                // a :- b.
+			.startRule(BASICRULE).addHead(2).addToBody(1, true).endRule()                // b :- a.
+		.endProgram();
+		CPPUNIT_ASSERT(builder.stats.sccs == 1);
+		
+		solver().force(~ctx.symbolTable()[3].lit);
+		attachUfs();
+		CPPUNIT_ASSERT(solver().isFalse(ctx.symbolTable()[1].lit));
+	}
+
 	void testIncrementalUfs() {
-		builder.startProgram(ctx, ProgramBuilder::EqOptions().noEq());
+		builder.start(ctx, LogicProgram::AspOptions().noEq());
 		// I1:
 		// a :- not b.
 		// b :- not a.
 		// freeze(c).
-		builder.updateProgram()
-			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
+		builder.updateProgram();
+		builder.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c")
 			.startRule().addHead(1).addToBody(2, false).endRule()
 			.startRule().addHead(2).addToBody(1, false).endRule()
 			.freeze(3)
 		;
 		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram() && ctx.endInit());
-		CPPUNIT_ASSERT(0u == builder.dependencyGraph()->nodes());
+		CPPUNIT_ASSERT(ctx.sccGraph.get() == 0);
 		
 		// I2:
 		// {c, z}
 		// x :- not c.
 		// x :- y, z.
 		// y :- x, z.
-		builder.updateProgram()
-			.setAtomName(4, "x").setAtomName(5, "y").setAtomName(6, "z")
+		builder.updateProgram();
+		builder.setAtomName(4, "x").setAtomName(5, "y").setAtomName(6, "z")
 			.startRule(CHOICERULE).addHead(3).addHead(6).endRule()
 			.startRule().addHead(4).addToBody(3, false).endRule()
 			.startRule().addHead(4).addToBody(5, true).addToBody(6, true).endRule()
@@ -361,37 +387,121 @@ public:
 			.unfreeze(3)
 		;
 		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
-		CPPUNIT_ASSERT(5u == builder.dependencyGraph()->nodes());
+		CPPUNIT_ASSERT(6u == ctx.sccGraph.get()->nodes());
 		CPPUNIT_ASSERT(1 == builder.stats.sccs);
 		attachUfs();
-		CPPUNIT_ASSERT(5u == ufs->nodes());
+		CPPUNIT_ASSERT(6u == ufs->nodes());
 
 		// I3:
 		// a :- x, not r.
 		// r :- not a.
 		// a :- b.
 		// b :- a, not z.
-		builder.updateProgram()
-			.setAtomName(7, "a").setAtomName(8, "b").setAtomName(9, "r")
+		builder.updateProgram();
+		builder.setAtomName(7, "a").setAtomName(8, "b").setAtomName(9, "r")
 			.startRule().addHead(7).addToBody(4, true).addToBody(9, false).endRule()
 			.startRule().addHead(9).addToBody(7, false).endRule()
 			.startRule().addHead(7).addToBody(8, true).endRule()
 			.startRule().addHead(8).addToBody(7, true).addToBody(6, false).endRule()
 		;
 		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
-		CPPUNIT_ASSERT(10u == builder.dependencyGraph()->nodes());
+		CPPUNIT_ASSERT(11u == ctx.sccGraph.get()->nodes());
 		CPPUNIT_ASSERT(1 == builder.stats.sccs);
 		ctx.endInit();
-		CPPUNIT_ASSERT(10u == ufs->nodes());
+		CPPUNIT_ASSERT(11u == ufs->nodes());
 		CPPUNIT_ASSERT(builder.getAtom(7)->scc() != builder.getAtom(4)->scc());
 	}
+
+	void testInitialStopConflict() {
+		builder.start(ctx);
+		// I1:
+		// {x,y}.
+		// a :- {x, y}.
+		// a :- b,y.
+		// b :- a,y.
+		builder.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "x").setAtomName(4, "y")
+			.startRule(CHOICERULE).addHead(3).addHead(4).endRule()
+			.startRule().addHead(1).addToBody(3, true).addToBody(4, true).endRule()
+			.startRule().addHead(1).addToBody(2, true).addToBody(4, true).endRule()
+			.startRule().addHead(2).addToBody(1, true).addToBody(4, true).endRule()
+		;
+		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
+		CPPUNIT_ASSERT(builder.stats.sccs == 1);
+		struct M : PostPropagator {			
+			uint32 priority() const { return priority_reserved_msg; }
+			bool propagateFixpoint(Solver& s, PostPropagator*) {
+				s.setStopConflict();
+				return false;
+			}
+		} m;
+		solver().addPost(&m);
+		ctx.addUnary(~builder.getAtom(3)->literal());
+		attachUfs();
+		CPPUNIT_ASSERT(solver().hasStopConflict());
+		solver().removePost(&m);
+		solver().popRootLevel();
+		solver().propagate();
+		CPPUNIT_ASSERT(solver().isFalse(builder.getAtom(3)->literal()));
+		CPPUNIT_ASSERT(solver().isFalse(builder.getAtom(1)->literal()) && solver().isFalse(builder.getAtom(2)->literal()));
+	}
+
+	void testIncrementalLearnFact() {
+		builder.start(ctx);
+		builder.update();
+		// I1:
+		// {x,y}.
+		// a :- {x, y}.
+		// a :- b,y.
+		// b :- a,y.
+		builder.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "x").setAtomName(4, "y")
+			.startRule(CHOICERULE).addHead(3).addHead(4).endRule()
+			.startRule().addHead(1).addToBody(3, true).addToBody(4, true).endRule()
+			.startRule().addHead(1).addToBody(2, true).addToBody(4, true).endRule()
+			.startRule().addHead(2).addToBody(1, true).addToBody(4, true).endRule()
+		;
+		CPPUNIT_ASSERT_EQUAL(true, builder.endProgram());
+		attachUfs();
+		builder.update();
+		builder.endProgram();
+		ctx.addUnary(~builder.getAtom(3)->literal());
+		CPPUNIT_ASSERT(solver().propagate());
+		CPPUNIT_ASSERT(ctx.endInit());
+		CPPUNIT_ASSERT(solver().isFalse(builder.getAtom(3)->literal()));
+		CPPUNIT_ASSERT(solver().isFalse(builder.getAtom(1)->literal()) && solver().isFalse(builder.getAtom(2)->literal()));
+	}
+
+	void testApproxUfs() {
+		builder.start(ctx)
+			.setAtomName(1, "a").setAtomName(2, "b").setAtomName(3, "c").setAtomName(4, "d").setAtomName(5,"e")
+			.startRule(DISJUNCTIVERULE).addHead(1).addHead(2).endRule() // a | b.
+			.startRule(DISJUNCTIVERULE).addHead(3).addHead(4).endRule() // c | d.
+			.startRule(DISJUNCTIVERULE).addHead(3).addHead(5).addToBody(2, true).endRule() // c | e :- b.
+			.startRule(DISJUNCTIVERULE).addHead(2).addHead(4).addToBody(3, true).endRule() // b | d :- c.
+			.startRule().addHead(3).addToBody(4,true).addToBody(2, false).endRule() // c :- d, not b.
+			.startRule().addHead(3).addToBody(5,true).endRule()                     // c :- e.
+			.startRule().addHead(4).addToBody(5,true).addToBody(1, false).endRule() // d :- e, not a.
+			.startRule().addHead(5).addToBody(3,true).addToBody(4, true).endRule()  // e :- c, d.
+		.endProgram();
+		CPPUNIT_ASSERT(builder.stats.sccs == 1);
+		CPPUNIT_ASSERT(builder.getAtom(1)->scc() == PrgNode::noScc);
+		CPPUNIT_ASSERT(5u == ctx.sccGraph.get()->numAtoms());
+		CPPUNIT_ASSERT(8u == ctx.sccGraph.get()->numBodies());
+		attachUfs();
+		CPPUNIT_ASSERT_EQUAL(true, solver().assume(ctx.symbolTable()[2].lit) && solver().propagate());
+		solver().assume(ctx.symbolTable()[3].lit);
+		CPPUNIT_ASSERT_EQUAL(true, solver().propagateUntil(ufs.get()));
+		
+		CPPUNIT_ASSERT(solver().value(ctx.symbolTable()[4].lit.var()) == value_free);
+		CPPUNIT_ASSERT_EQUAL(true, ufs->propagate(solver()));
+		CPPUNIT_ASSERT_MESSAGE("TODO: Implement approx. ufs!", solver().isFalse(ctx.symbolTable()[4].lit));
+		
+	}
 private:
 	SharedContext ctx;
-	SingleOwnerPtr<DefaultUnfoundedCheck> ufs;
-	ProgramBuilder builder;
+	SingleOwnerPtr<WrapDefaultUnfoundedCheck> ufs;
+	LogicProgram builder;
 	void attachUfs() {
-		ufs->attachTo(solver(), builder.dependencyGraph());
-		ufs.release();
+		solver().addPost(ufs.release());
 		ctx.endInit();
 	}
 	void setupSimpleProgram() {
