@@ -49,10 +49,14 @@ class WeightConstraintTest : public CppUnit::TestFixture {
 	CPPUNIT_TEST(testWeightReasonAfterBackprop);
 	CPPUNIT_TEST(testOrderBug);
 	CPPUNIT_TEST(testBackwardAfterForward);
+
+	CPPUNIT_TEST(testOnlyBTB);
+	CPPUNIT_TEST(testOnlyBFB);
 	
 	CPPUNIT_TEST(testSimplify);
 	CPPUNIT_TEST(testSimplifyCardinality);
 	CPPUNIT_TEST(testSimplifyWeight);
+	CPPUNIT_TEST(testSimplifyWatches);
 
 	CPPUNIT_TEST(testAssertWeightTriviallySat);
 	CPPUNIT_TEST(testAssertWeightTriviallyUnSat);
@@ -66,6 +70,13 @@ class WeightConstraintTest : public CppUnit::TestFixture {
 
 	CPPUNIT_TEST(testCloneWeight);
 	CPPUNIT_TEST(testCloneWeightShared);
+
+	CPPUNIT_TEST(testAddOnLevel);
+	CPPUNIT_TEST(testCreateSat);
+	CPPUNIT_TEST(testCreateSatOnRoot);
+	CPPUNIT_TEST(testCreateSatOnRootNoProp);
+
+	CPPUNIT_TEST(testMergeNegativeWeight);
 	CPPUNIT_TEST_SUITE_END(); 
 public:
 	WeightConstraintTest() {
@@ -117,7 +128,7 @@ public:
 	void testTrivialBackpropTrue() {
 		WeightLitVec lits = makeWeightLits();
 		solver().force(body, 0);
-		CPPUNIT_ASSERT_EQUAL(true,  WeightConstraint::newWeightConstraint(ctx, body, lits, 7));
+		CPPUNIT_ASSERT_EQUAL(true,  newWeightConstraint(ctx, body, lits, 7));
 		CPPUNIT_ASSERT(ctx.numConstraints() == 1);
 		CPPUNIT_ASSERT(solver().isTrue(a));
 		CPPUNIT_ASSERT(solver().isTrue(~b));
@@ -140,7 +151,7 @@ public:
 	void testTrivialBackpropFalseWeight() {
 		WeightLitVec lits = makeWeightLits();
 		solver().force(~body, 0);
-		CPPUNIT_ASSERT_EQUAL(true,  WeightConstraint::newWeightConstraint(ctx, body, lits, 2));
+		CPPUNIT_ASSERT_EQUAL(true,  newWeightConstraint(ctx, body, lits, 2));
 		CPPUNIT_ASSERT(ctx.numConstraints() == 1);
 		CPPUNIT_ASSERT(solver().isFalse(a));
 		CPPUNIT_ASSERT(solver().isFalse(~b));
@@ -287,7 +298,7 @@ public:
 
 	void testWeightReasonAfterBackprop() {
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		solver().assume(~body) && solver().propagate();
 		CPPUNIT_ASSERT_EQUAL(true, solver().isTrue(~a));
 		solver().assume(d) && solver().propagate();
@@ -342,6 +353,53 @@ public:
 		CPPUNIT_ASSERT_EQUAL(true, solver().isTrue(body));
 	}
 
+	void testOnlyBTB() {
+		ctx.endInit(true);
+		Solver& s = *ctx.master();
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(a, 1));
+		lits.push_back(WeightLiteral(b, 1));
+		lits.push_back(WeightLiteral(c, 1));
+		lits.push_back(WeightLiteral(d, 1));
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, lits, 2, WeightConstraint::create_only_btb);
+		CPPUNIT_ASSERT(res.first());
+		s.assume(a) && s.propagate();
+		s.assume(b) && s.propagate();
+		// FTB_BFB not added
+		CPPUNIT_ASSERT(s.value(body.var()) == value_free);
+		s.undoUntil(0);
+		s.assume(~a) && s.propagate();
+		s.assume(~b) && s.propagate();
+		uint32 dl = s.decisionLevel();
+		s.assume(body) && s.propagate();
+		CPPUNIT_ASSERT(s.isTrue(c) && s.isTrue(d));
+		s.undoUntil(dl);
+		s.assume(~c) && s.propagate();
+		CPPUNIT_ASSERT(s.isFalse(body));
+	}
+	void testOnlyBFB() {
+		ctx.endInit(true);
+		Solver& s = *ctx.master();
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(a, 1));
+		lits.push_back(WeightLiteral(b, 1));
+		lits.push_back(WeightLiteral(c, 1));
+		lits.push_back(WeightLiteral(d, 1));
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, lits, 2, WeightConstraint::create_only_bfb);
+		CPPUNIT_ASSERT(res.first());
+		s.assume(a) && s.propagate();
+		uint32 dl = s.decisionLevel();
+		s.assume(b) && s.propagate();
+		CPPUNIT_ASSERT(s.isTrue(body));
+		s.undoUntil(dl);
+		s.assume(~body) && s.propagate();
+		CPPUNIT_ASSERT(s.isFalse(b) && s.isFalse(c) && s.isFalse(d));
+		s.undoUntil(0);
+		s.assume(~a) && s.propagate();
+		s.assume(~b) && s.propagate();
+		s.assume(~b) && s.propagate();
+		CPPUNIT_ASSERT(s.value(body.var()) == value_free);
+	}
 	void testSimplify() {
 		LitVec lits = makeLits();
 		CPPUNIT_ASSERT_EQUAL(true, newCardinalityConstraint(ctx, lits, 2));
@@ -375,7 +433,7 @@ public:
 		lits[2].second    = 2;
 		lits.push_back(WeightLiteral(e, 1));
 		lits.push_back(WeightLiteral(f, 1));
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 2));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 2));
 		ctx.addUnary(body);
 		ctx.addUnary(~a);
 		ctx.addUnary(~d);
@@ -391,23 +449,33 @@ public:
 		CPPUNIT_ASSERT(out.size() == 1 && out[0] == c);
 	}
 
+	void testSimplifyWatches() {
+		LitVec lits = makeLits();
+		CPPUNIT_ASSERT_EQUAL(true, newCardinalityConstraint(ctx, lits, 2));
+		uint32 nw3  = solver().numWatches(lits[3]) + solver().numWatches(~lits[3]);
+		solver().pushRoot(body);
+		CPPUNIT_ASSERT(nw3 >= solver().numWatches(lits[3]) + solver().numWatches(~lits[3]));
+		solver().popRootLevel(1);
+		CPPUNIT_ASSERT(nw3 == solver().numWatches(lits[3]) + solver().numWatches(~lits[3]));
+	}
+
 	void testAssertWeightTriviallySat() {
 		WeightLitVec lits;
 		lits.push_back(WeightLiteral(a, 2));
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 0));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 0));
 		CPPUNIT_ASSERT(solver().isTrue(body));
 	}
 	void testAssertWeightTriviallyUnSat() {
 		WeightLitVec lits;
 		lits.push_back(WeightLiteral(a, 2));
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		CPPUNIT_ASSERT(solver().isFalse(body));
 	}
 
 	void testAssertWeightNotSoTriviallySat() {
 		WeightLitVec lits = makeWeightLits();
 		solver().force(lits[1].first, 0);
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body,lits, 2));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body,lits, 2));
 		CPPUNIT_ASSERT(solver().isTrue(body));
 	}
 
@@ -415,7 +483,7 @@ public:
 		WeightLitVec lits = makeWeightLits();
 		solver().force(~lits[0].first, 0);
 		solver().force(~lits[2].first, 0);
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 4));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 4));
 		CPPUNIT_ASSERT(solver().isTrue(~body));
 	}
 
@@ -446,7 +514,7 @@ public:
 
 	void testWeightBackwardTrue() {
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		solver().assume(~a);
 		solver().force(body, 0);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagate());
@@ -477,7 +545,7 @@ public:
 
 	void testWeightBackwardFalse() {
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		solver().assume(~body);
 		CPPUNIT_ASSERT_EQUAL(true, solver().propagate());
 		CPPUNIT_ASSERT_EQUAL(true, solver().isTrue(~a));
@@ -502,7 +570,7 @@ public:
 
 	void testWeightConflict() {
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		LitVec assume;
 		assume.push_back(body);
 		assume.push_back(~a);
@@ -519,13 +587,13 @@ public:
 	}
 
 	void testCloneWeight() {
-		ctx.setSolvers(2);
+		ctx.setConcurrency(2);
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		solver().force(~a, 0);
-		ctx.endInit();
-		Solver solver2;
-		ctx.attach(solver2);
+		Solver& solver2 = ctx.addSolver();
+		ctx.endInit(true);
+		
 		CPPUNIT_ASSERT(solver2.numConstraints() == 1);
 		
 		CPPUNIT_ASSERT(solver2.numWatches(a) == 0 && solver2.numWatches(~a) == 0);
@@ -546,14 +614,14 @@ public:
 	}
 
 	void testCloneWeightShared() {
-		ctx.setSolvers(2);
-		ctx.physicalSharing(SharedContext::share_problem);
+		ctx.setConcurrency(2);
+		ctx.setShareMode(ContextParams::share_problem);
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		solver().force(~a, 0);
-		ctx.endInit();
-		Solver solver2;
-		ctx.attach(solver2);
+		Solver& solver2 = ctx.addSolver();
+		ctx.endInit(true);
+		
 		CPPUNIT_ASSERT(solver2.numConstraints() == 1);
 		
 		CPPUNIT_ASSERT(solver2.numWatches(a) == 0 && solver2.numWatches(~a) == 0);
@@ -572,6 +640,99 @@ public:
 		solver2.reason(~c, out);
 		CPPUNIT_ASSERT(std::find(out.begin(), out.end(), ~d) != out.end());
 	}
+
+	void testAddOnLevel() {
+		ctx.endInit(true);
+		WeightLitVec lits = makeWeightLits();
+		uint32 sz = lits.size() + 1;
+		Solver& s = *ctx.master();
+		s.pushRoot(f);
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, lits, 2, WeightConstraint::create_no_add);
+		CPPUNIT_ASSERT(res.ok() && res.first() != 0);
+		CPPUNIT_ASSERT(res.first()->size() == sz && lits.size() == sz-1);
+		s.force(body);
+		s.force(~lits[0].first);
+		s.force(~lits[1].first,0);
+		s.propagate();
+		CPPUNIT_ASSERT(s.isTrue(lits[2].first) && s.isTrue(lits[3].first));
+		res.first()->destroy(&s, true);
+	}
+	void testCreateSat() {
+		ctx.endInit(true);
+		WeightLitVec lits = makeWeightLits();
+		uint32 sz = lits.size() + 1;
+		Solver& s = *ctx.master();
+		
+		s.force(lits[0].first);
+		s.force(lits[1].first);
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, lits, 2, WeightConstraint::create_no_add|WeightConstraint::create_sat);
+		CPPUNIT_ASSERT(res.ok() && res.first() != 0);
+		CPPUNIT_ASSERT(s.isTrue(body));		
+		s.propagate();
+		CPPUNIT_ASSERT(s.isTrue(body));
+		while (!lits.empty()) {
+			CPPUNIT_ASSERT(s.force(~lits.back().first) && s.propagate());
+			lits.pop_back();
+		}
+		res.first()->destroy(&s, true);
+	}
+	void testCreateSatOnRoot() {
+		ctx.endInit(true);
+		Solver& s = *ctx.master();
+		s.pushRoot(f);
+		CPPUNIT_ASSERT(s.rootLevel() == 1);
+		s.force(a, 0);
+		s.force(b, 0);
+		s.propagate();
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(a, 1));
+		lits.push_back(WeightLiteral(b, 1));
+		lits.push_back(WeightLiteral(c, 1));
+		lits.push_back(WeightLiteral(d, 1));
+		WeightLitsRep rep           = WeightLitsRep::create(s, lits, 2);
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, rep, WeightConstraint::create_no_add|WeightConstraint::create_sat);
+		CPPUNIT_ASSERT(res.ok() && res.first());
+		CPPUNIT_ASSERT(s.isTrue(body));
+		CPPUNIT_ASSERT(s.reason(body) == res.first());
+		s.popRootLevel(1);
+		CPPUNIT_ASSERT(s.value(body.var()) == value_free);
+		res.first()->destroy(&s, true);
+	}
+	void testCreateSatOnRootNoProp() {
+		ctx.endInit(true);
+		Solver& s = *ctx.master();
+		s.pushRoot(f);
+		CPPUNIT_ASSERT(s.rootLevel() == 1);
+		s.force(a, 0);
+		s.force(b, 0);
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(a, 1));
+		lits.push_back(WeightLiteral(b, 1));
+		lits.push_back(WeightLiteral(c, 1));
+		lits.push_back(WeightLiteral(d, 1));
+		WeightLitsRep rep           = WeightLitsRep::create(s, lits, 2);
+		WeightConstraint::CPair res = WeightConstraint::create(s, body, rep, WeightConstraint::create_no_add|WeightConstraint::create_sat);
+		CPPUNIT_ASSERT(res.ok() && res.first());
+		CPPUNIT_ASSERT(!s.isTrue(body));
+		CPPUNIT_ASSERT(s.propagate());
+		CPPUNIT_ASSERT(s.isTrue(body));
+		CPPUNIT_ASSERT(s.reason(body) == res.first());
+		s.popRootLevel(1);
+		CPPUNIT_ASSERT(s.value(body.var()) == value_free);
+		res.first()->destroy(&s, true);
+	}
+	void testMergeNegativeWeight() {
+		WeightLitVec lits;
+		lits.push_back(WeightLiteral(a, -1));
+		lits.push_back(WeightLiteral(a, -1));
+		lits.push_back(WeightLiteral(b, -2));
+		WeightLitsRep rep = WeightLitsRep::create(*ctx.master(), lits, -2);
+		CPPUNIT_ASSERT(rep.size == 2);
+		CPPUNIT_ASSERT(rep.bound == 1);
+		CPPUNIT_ASSERT(rep.reach == 2);
+		CPPUNIT_ASSERT(rep.lits[0].second == 1);
+		CPPUNIT_ASSERT(rep.lits[1].second == 1);
+	}
 private:
 	static bool newCardinalityConstraint(SharedContext& ctx, const LitVec& lits, int bound) {
 		CPPUNIT_ASSERT(lits.size() >  1);
@@ -579,7 +740,10 @@ private:
 		for (LitVec::size_type i = 1; i < lits.size(); ++i) {
 			wlits.push_back(WeightLiteral(lits[i], 1));
 		}
-		return WeightConstraint::newWeightConstraint(ctx, lits[0], wlits, bound);
+		return newWeightConstraint(ctx, lits[0], wlits, bound);
+	}
+	static bool newWeightConstraint(SharedContext& ctx, Literal W, WeightLitVec& lits, weight_t bound) {
+		return WeightConstraint::create(*ctx.master(), W, lits, bound).ok();
 	}
 	void propCard(LitVec& assumptions, const LitVec& expected) {
 		LitVec lits = makeLits();
@@ -588,7 +752,7 @@ private:
 	}
 	void propWeight(LitVec& assume, LitVec& expect) {
 		WeightLitVec lits = makeWeightLits();
-		CPPUNIT_ASSERT_EQUAL(true, WeightConstraint::newWeightConstraint(ctx, body, lits, 3));
+		CPPUNIT_ASSERT_EQUAL(true, newWeightConstraint(ctx, body, lits, 3));
 		propImpl(assume, expect);
 	}
 	void propImpl(LitVec& assumptions, const LitVec& expected) {
